/*-----------------------------------------------------------------
rckt.cpp - hardware routines for rocket port board. OS independent.

1-10-96 PCI baudrate prescaler not setup, fix - kpb.
11-05-95 more work making it comprehendable - kpb.
10-09-95 strip out some more junk - kpb.
9-20-95 add PCI, clean up some misc things - kpb.
8-08-95 rewrite, strip out bogus typedefs, extra globals, etc. - kpb
 Copyright (c) 1995-97, Comtrol Corp.
|------------------------------------------------------------------*/
#include "rcktldef.h"

#define ROCKET_C
#include "rckt.h"

/* Master copy of AIOP microcode.  Organized as DWORDs.  The 1st word of each
   DWORD holds the microcode index, the second word holds the microcode
   data. */
BYTE RData[RDATASIZE] =
{
#ifdef FAST_ROCKETPLUS_SNAKEOIL
 /*  0*/ 0x00, 0x09, 0x08, 0x0a,  /* 0 - fast minimal microcode */
#else
 /*  0*/ 0x00, 0x09, 0xf6, 0x82, /* 0 - Stop Rx processor */
 /*  4*/ 0x02, 0x09, 0x86, 0xfb, /*  */
 /*  8*/ 0x04, 0x09, 0x00, 0x0a, /*  */
 /*  c*/ 0x06, 0x09, 0x01, 0x0a, /*  */
 /* 10*/ 0x08, 0x09, 0x8a, 0x13, /*  */
 /*  4*/ 0x0a, 0x09, 0xc5, 0x11, /*  */
 /*  8*/ 0x0c, 0x09, 0x86, 0x85, /*  */
 /*  c*/ 0x0e, 0x09, 0x20, 0x0a, /*  */
 /* 20*/ 0x10, 0x09, 0x21, 0x0a, /*  */
 /*  4*/ 0x12, 0x09, 0x41, 0xff, /* 12 Rx mask char */
 /*  8*/ 0x14, 0x09, 0x82, 0x00, /* 14 Compare/Ignore #0 */
 /*  c*/ 0x16, 0x09, 0x82, 0x7b, /* 16 Compare #1 */
 /* 30*/ 0x18, 0x09, 0x8a, 0x7d, /* 18 Compare #2 */
 /*  4*/ 0x1a, 0x09, 0x88, 0x81, /* 1a Interrupt #1 */
 /*  8*/ 0x1c, 0x09, 0x86, 0x7a, /* 1c Ignore/Replace #1 */
 /*  c*/ 0x1e, 0x09, 0x84, 0x81, /* 1e Interrupt #2 */
 /* 40*/ 0x20, 0x09, 0x82, 0x7c, /* 20 Ignore/Replace #2 */
 /*  4*/ 0x22, 0x09, 0x0a, 0x0a  /* 22 Rx FIFO Enable */
#endif
};

/* Per channel, Indirect Register addr-data */
BYTE RRegData[RREGDATASIZE]=
{
#ifdef FAST_ROCKETPLUS_SNAKEOIL
 /*  0*/ 0x00, 0x09, 0x08, 0x0a,  /* 0 - fast minimal microcode */
#else
 /*  0*/  0x00, 0x09, 0xf6, 0x82,  /*  0 Stop Rx processor */
 /*  4*/  0x08, 0x09, 0x8a, 0x13,  /*  8 Tx software flow control */
 /*  8*/  0x0a, 0x09, 0xc5, 0x11,  /*  a XON char */
 /*  c*/  0x0c, 0x09, 0x86, 0x85,  /*  c XANY */
 /* 10*/  0x12, 0x09, 0x41, 0xff,  /* 12 Rx mask char */
 /*  4*/  0x14, 0x09, 0x82, 0x00,  /* 14 Compare/Ignore #0 */
 /*  8*/  0x16, 0x09, 0x82, 0x7b,  /* 16 Compare #1 */
 /*  c*/  0x18, 0x09, 0x8a, 0x7d,  /* 18 Compare #2 */
 /* 20*/  0x1a, 0x09, 0x88, 0x81,  /* 1a Interrupt #1 */
 /*  4*/  0x1c, 0x09, 0x86, 0x7a,  /* 1c Ignore/Replace #1 */
 /*  8*/  0x1e, 0x09, 0x84, 0x81,  /* 1e Interrupt #2 */
 /*  c*/  0x20, 0x09, 0x82, 0x7c,  /* 20 Ignore/Replace #2 */
 /* 30*/  0x22, 0x09, 0x0a, 0x0a   /* 22 Rx FIFO Enable */
#endif
};
// this is a public global array of Controller structures used
// for each board

CONTROLLER_T sController[CTL_SIZE];

#ifdef COMMENT_OUT
/*-------------------------------------------------------------------
Function: sInitController
Purpose:  Initialization of controller global registers and controller
		  structure.
Call:     sInitController(CtlP,CtlNum,MudbacIO,AiopIOList,AiopIOListSize,
						  IRQNum,Frequency,PeriodicOnly)
		  CONTROLLER_T *CtlP; Ptr to controller structure
		  int CtlNum; Controller number
		  WORD MudbacIO; Mudbac base I/O address.
		  WORD *AiopIOList; List of I/O addresses for each AIOP.
			 This list must be in the order the AIOPs will be found on the
			 controller.  Once an AIOP in the list is not found, it is
			 assumed that there are no more AIOPs on the controller.
		  int AiopIOListSize; Number of addresses in AiopIOList
		  int IRQNum; Interrupt Request number.  Can be any of the following:
						 0: Disable global interrupts
						 3: IRQ 3
						 4: IRQ 4
						 5: IRQ 5
						 9: IRQ 9
						 10: IRQ 10
						 11: IRQ 11
						 12: IRQ 12
						 15: IRQ 15
		  BYTE Frequency: A flag identifying the frequency
				   of the periodic interrupt, can be any one of the following:
					  FREQ_DIS - periodic interrupt disabled
					  FREQ_137HZ - 137 Hertz
					  FREQ_69HZ - 69 Hertz
					  FREQ_34HZ - 34 Hertz
					  FREQ_17HZ - 17 Hertz
					  FREQ_9HZ - 9 Hertz
					  FREQ_4HZ - 4 Hertz
				   If IRQNum is set to 0 the Frequency parameter is
				   overidden, it is forced to a value of FREQ_DIS.
		  int PeriodicOnly: TRUE if all interrupts except the periodic
							   interrupt are to be blocked.
							FALSE is both the periodic interrupt and
							   other channel interrupts are allowed.
							If IRQNum is set to 0 the PeriodicOnly parameter is
							   overidden, it is forced to a value of FALSE.
Return:   int: Number of AIOPs on the controller, or CTLID_NULL if controller
			   initialization failed.

Comments:
		  If periodic interrupts are to be disabled but AIOP interrupts
		  are allowed, set Frequency to FREQ_DIS and PeriodicOnly to FALSE.

		  If interrupts are to be completely disabled set IRQNum to 0.

		  Setting Frequency to FREQ_DIS and PeriodicOnly to TRUE is an
		  invalid combination.

		  This function performs initialization of global interrupt modes,
		  but it does not actually enable global interrupts.  To enable
		  and disable global interrupts use functions sEnGlobalInt() and
		  sDisGlobalInt().  Enabling of global interrupts is normally not
		  done until all other initializations are complete.

		  Even if interrupts are globally enabled, they must also be
		  individually enabled for each channel that is to generate
		  interrupts.

Warnings: No range checking on any of the parameters is done.

		  No context switches are allowed while executing this function.

		  After this function all AIOPs on the controller are disabled,
		  they can be enabled with sEnAiop().
|--------------------------------------------------------------------*/
int sInitController(CONTROLLER_T *CtlP,
					int CtlNum,
					WORD MudbacIO,
					WORD *AiopIOList,
					int AiopIOListSize,
					int IRQNum,
					BYTE Frequency,
					int PeriodicOnly)
{
  int    i;
  WORD  io;
  WORD  Mreg3;  // mudback register 3, aiop en, freq sel.
  BYTE irq_bits,freq_bits;

   CtlP->MBaseIO = MudbacIO;
   CtlP->MReg1IO = MudbacIO + 1;
   CtlP->MReg2IO = MudbacIO + 2;

   switch (IRQNum)   // setup irq bits(MReg2)
   {
	 case  3: irq_bits = 0x10; break;
	 case  4: irq_bits = 0x20; break;
	 case  5: irq_bits = 0x30; break;
	 case  9: irq_bits = 0x40; break;
	 case 10: irq_bits = 0x50; break;
	 case 11: irq_bits = 0x60; break;
	 case 12: irq_bits = 0x70; break;
	 case 15: irq_bits = 0x80; break;
	 default: irq_bits = 0;    break;  /* 0=irq not used */
   }  
   if (irq_bits != 0) /* if valid IRQ used */
   {
	 freq_bits = Frequency;    // set freq.(hi nibble Reg3)
	 if (PeriodicOnly)         // periodic interrupt only
	   freq_bits |= PERIODIC_ONLY;  // 0x80
   }
   else          /* interrupts globally disabled */
	 freq_bits = 0;    // no periodic interrupts

   if (CtlP->BusType != BUS_PCI)  // mudback does not exist for PCI */
   {
	 sOutB(MudbacIO+2,irq_bits);  // mudback Interrupt Enable
	 CtlP->MReg2 = irq_bits;
	 sOutB(MudbacIO+3,freq_bits);  // mudback Aiopic En., Periodic Setup
	 if (irq_bits != 0) // if IRQ used
	   sControllerEOI(CtlP);               /* clear EOI if warm init */
   }
   else // PCI
   {
	 CtlP->Reg3AIO = AiopIOList[0] + _PCI_INT_FUNC;
	 CtlP->Reg3A = ((freq_bits >> 4) & 0xf);   // periodic & periodic_only
	 if (irq_bits != 0) // if IRQ used
	   sPCIControllerEOI(CtlP);               /* clear EOI if warm init */
   }

   /* Init AIOPs */
   CtlP->NumAiop = 0;
   for(i=0; i < AiopIOListSize; i++)  // 4
   {
	  io = AiopIOList[i];
	  CtlP->AiopIO[i] = (WORD)io;
	  CtlP->AiopIntChanIO[i] = io + _INT_CHAN;
	  if (CtlP->BusType != BUS_PCI)  // mudback does not exist for PCI */
	  {
		sOutB(MudbacIO+2, irq_bits | (i & 0x03)); /* AIOP index */
		sOutB(MudbacIO,(BYTE)(io >> 6));  /* set up AIOP I/O in MUDBAC */
		sEnAiop(CtlP,i);                            /* enable the AIOP */
	  }
//      else Mudback in PIC, AIOPIC address pre-set

	  if(sReadAiopID(io) == -1)       /* if AIOP does not exist */
	  {
//        sDisAiop(CtlP,i);                     /* disable AIOP */
		break;                                /* done looking for AIOPs */
	  }

	  CtlP->AiopNumChan[i] = sReadAiopNumChan((WORD)io); /* num channels in AIOP */

	  sOutW(io + _INDX_ADDR,_CLK_PRE);      /* clock prescaler */
	  sOutB(io + _INDX_DATA,CtlP->clk_prescaler);
	  CtlP->NumAiop++;                         /* bump count of AIOPs */
//      sDisAiop(CtlP,i);                        /* disable AIOP */
   }

   if(CtlP->NumAiop == 0)
	  return(-1);
   else
	  return(CtlP->NumAiop);
}
#endif

/*-------------------------------------------------------------------
 sInitCtlStruct - Setup a controller misc. configuration.
|--------------------------------------------------------------------*/
int sInitCtlStruct(CONTROLLER_T *Ctlp,
			 int BusType,
			 int CardNum,
			 int CardType)
{
 int j;


  Ctlp->PortsPerAiop = 8;
  Ctlp->CardNum = CardNum;
  Ctlp->BusType = BusType;

  if (BusType == BUS_PCI)
  {
	switch(CardType)
	{
		
		case 1: 
		{    
			j = 32; 
			break;
		}
		case 2: 
		{    
			j = 8;  
			break;
		}
		case 3: 
		{    
			j = 16; 
			break;
		}
		case 4: 
		{    
			j = 4; 
			break;
		}
		case 5: 
		{    
			j = 8; 
			break;
		}

		case 10:         // rplus 4 port(only uses 4-ports per aiop)
		{
			//puts("pci-A");
			j = 4;
			Ctlp->PortsPerAiop = 4;
			break;
		}

		case 11:         // rplus 8 port(only uses 4-ports per aiop)
		{
			//puts("pci-B");
			j = 8;
			Ctlp->PortsPerAiop = 4;
			break;
		}

		case 0x801: 
		{
			j = 32; 
			break;
		}
		case 0x802: 
		{
			j = 8;  
			break;
		}
	   
		case 0x804:   
		{
			j = 4;  
			break;
		}

		case 0x805: 
		{
			j = 8;  
			break;
		}
		
		case 0x803: 
		{
			j = 16; 
			break;
		}
	  
		default:
		{
				j = 8;
				puts("Bad PCI DevId");
				break;
		}
	  //#endif
	}
	Ctlp->NumPorts = j;
  }
  else
  {
  }
 return 0;
}

/*-------------------------------------------------------------------
 sInitIRQ - Setup a controller to generate interrupts.
  Still need to do a sEnGlobalInt() call to start interrupts
|--------------------------------------------------------------------*/
int sInitIRQ(CONTROLLER_T *CtlP,
			 int IRQNum,
			 BYTE Frequency,
			 int PeriodicOnly)
{
  int    i;
  WORD  io;
  WORD  Mreg3;  // mudback register 3, aiop en, freq sel.
  BYTE irq_bits,freq_bits;
  BYTE cur;

  switch (IRQNum)   // setup irq bits(MReg2)
  {
	case  3: irq_bits = 0x10; break;
	case  4: irq_bits = 0x20; break;
	case  5: irq_bits = 0x30; break;
	case  9: irq_bits = 0x40; break;
	case 10: irq_bits = 0x50; break;
	case 11: irq_bits = 0x60; break;
	case 12: irq_bits = 0x70; break;
	case 15: irq_bits = 0x80; break;
	default: irq_bits = 0;    break;  /* 0=irq not used */
  }  

  if (irq_bits == 0) /* not a valid IRQ */
  {
	return 1; // error
  }

  freq_bits = Frequency;    // set freq.(hi nibble Reg3)
  if (PeriodicOnly)         // periodic interrupt only
	freq_bits |= PERIODIC_ONLY;  // 0x80

  if (CtlP->BusType != BUS_PCI)  // mudback does not exist for PCI */
  {
	cur = sInB(CtlP->MBaseIO+2) & AIOP_EN_BITS; // retain AIOP_EN_BITS bits
	cur |= irq_bits;
	sOutB(CtlP->MBaseIO+2,cur);  // mudback Interrupt Enable
	CtlP->MReg2 = cur;           // retain current settings

	cur = sInB(CtlP->MBaseIO+3) & 0xf;   // retain aiop enable bits
	cur |= freq_bits;
	sOutB(CtlP->MBaseIO+3,cur); // mudback Aiopic En., Periodic Setup
	sControllerEOI(CtlP);               /* clear EOI if warm init */
  }
  else // PCI
  {
	CtlP->Reg3AIO = CtlP->AiopIO[0] + _PCI_INT_FUNC;
	CtlP->Reg3A = ((freq_bits >> 4) & 0xf);   // periodic & periodic_only
	sPCIControllerEOI(CtlP);     // clear EOI if warm init
  }
  return 0; // ok
}

/*-------------------------------------------------------------------
 sSetupPorts - Setup a range of ports on the board.
|--------------------------------------------------------------------*/
int sSetupPorts(CONTROLLER_T *CtlP,
				WORD aiop_base_io,
				int offset_port,
				int num_ports,
				int bus_type,
				int card_type)  // used to pass in PCI device-id if PCI,
								// for ISA, used to pass in index of isa card
{
  static WORD isa_first_io;
  WORD aiopio[AIOP_CTL_SIZE];
  int    i,pn, aiop, last_aiop;
  WORD MudbacIO;
  WORD  io;
  BYTE val;
  int isa_card_index;
  //  int ports_per_aiop = 8;
  
	
  CtlP->clk_prescaler = NORMAL_CLOCK_PRESC;  // default to normal rocketport(not rplus)
  CtlP->BusType = bus_type;
  CtlP->PortsPerAiop = 8;  // default to 8 ports per aiop chip

  if (bus_type == BUS_ISA)
  {
	isa_card_index = card_type;
	if (isa_card_index == 0) // first ISA card
	  isa_first_io = aiop_base_io;   // alias up from first card for mudback
	MudbacIO= (isa_first_io + 0x40) + (isa_card_index * 0x400);
	for (aiop=0; aiop<AIOP_CTL_SIZE; aiop++)      // alias up 400H each
	  aiopio[aiop]= aiop_base_io + (aiop * 0x400);
  }
  else  // Microchannel or PCI
  {
	MudbacIO= aiop_base_io + 0x100;  // BUS_MC has mudback at 100H above base_io
	for (aiop=0; aiop<AIOP_CTL_SIZE; aiop++)
	  aiopio[aiop]= aiop_base_io + (aiop * 0x40);

	if ((card_type == 0xa) ||  // 4 port rplus(only uses 4 first ports on aiop)
		(card_type == 0xb))    // 8 port rplus(only uses 8 first ports on aiop)
	{
	  CtlP->PortsPerAiop = 4;
	  CtlP->clk_prescaler = RPLUS_CLOCK_PRESC;
	}
  }
  // setup mudback io addresses for quick access later
  CtlP->MBaseIO = MudbacIO;
  CtlP->MReg1IO = MudbacIO + 1;  // 
  CtlP->MReg2IO = MudbacIO + 2;  // for aiop interrupt status.

  /* Init AIOPs */
  CtlP->NumAiop = 4;

					  // make sure AIOP's initialized and ready for use
  last_aiop = -1;
  for(i=0; i < num_ports; i++)
  {
	aiop = (i + offset_port) / CtlP->PortsPerAiop;
	if (last_aiop != aiop)  // new aiop
	{
	  last_aiop = aiop;
	  io = aiopio[aiop];
	  
	  CtlP->AiopIO[aiop] = (WORD)io; // retain copy of aiopic base address
	  CtlP->AiopIntChanIO[aiop] = io + _INT_CHAN;
	  if (CtlP->BusType != BUS_PCI)  // mudback does not exist for PCI */
	  {
		// tell mudback where to position the base-io of the aiopic
		// try not to upset any other driver which might be using this board
		val = sInB(MudbacIO+2);  // read in irq, aiop-io reg
		sOutB(MudbacIO+2, (val & 0xfc) | (aiop & 0x03)); /* AIOP index */
		sOutB(MudbacIO,(BYTE)(io >> 6));  /* set up AIOP I/O in MUDBAC */

		//----- Reset the AIOPIC. 3-23-96
		//sOutB(io + _CMD_REG,RESET_ALL);  /* reset channel AIOPIC */
		//sOutB(io + _CMD_REG,0x0);

		sEnAiop(CtlP,aiop);                   /* enable the AIOP */
	  }

	  //if (sReadAiopID(io) == -1)       /* if AIOP does not exist */
	  //  return 1;

	  sOutW(io + _INDX_ADDR,_CLK_PRE);      /* clock prescaler */
	  sOutB(io + _INDX_DATA, CtlP->clk_prescaler);
	  CtlP->AiopNumChan[aiop] = sReadAiopNumChan((WORD)io); /* num channels in AIOP */
	}
  }

  return 0;
}

/*-------------------------------------------------------------------
Function: sInitChan
Purpose:  Initialization of a channel and channel structure
Call:     sInitChan(CtlP,ChP,AiopNum,ChanNum)
		  CONTROLLER_T *CtlP; Ptr to controller structure
		  CHANNEL_T *ChP; Ptr to channel structure
		  int AiopNum; AIOP number within controller
		  int ChanNum; Channel number within AIOP
Return:   int: TRUE if initialization succeeded, FALSE if it fails because channel
			   number exceeds number of channels available in AIOP.
Comments: This function must be called before a channel can be used.
Warnings: No range checking on any of the parameters is done.

		  No context switches are allowed while executing this function.
|--------------------------------------------------------------------*/
int sInitChan(CONTROLLER_T *CtlP,
			  CHANNEL_T *ChP,
			  int AiopNum,
			  int ChanNum)
{
   int i;
   WORD AiopIO;
   WORD ChIOOff;
   BYTE *ChR;
   WORD ChOff;
   static BYTE R[4];

   if(ChanNum >= CtlP->AiopNumChan[AiopNum])
	  return 1;  // error: exceeds num chans in AIOP

   //------ Setup our Channel structure values
   ChP->CtlP = CtlP;        // ptr to parent controller(board) struct
   ChP->AiopNum = AiopNum;  // 0-3
   ChP->ChanNum = ChanNum;  // 0-7

   //------ Save the Global direct addresses for quick access later
   AiopIO = CtlP->AiopIO[AiopNum];
   ChP->Cmd       = AiopIO + _CMD_REG;
   ChP->IntChan   = AiopIO + _INT_CHAN;
   ChP->IntMask   = AiopIO + _INT_MASK;
   ChP->IndexAddr = AiopIO + _INDX_ADDR;
   ChP->IndexData = AiopIO + _INDX_DATA;

   ChP->IntID     = AiopIO + ChP->ChanNum + _INT_ID0;

   //---- Save the Channel direct addresses for quick access later
   ChIOOff = AiopIO + ChP->ChanNum * 2;
   ChP->TxRxData   = ChIOOff + _TD0;
   ChP->ChanStat   = ChIOOff + _CHN_STAT0;
   ChP->TxRxCount  = ChIOOff + _FIFO_CNT0;

   //------ Copy Parts of Micro code & Data to our Channel struct
   ChR = ChP->R;
   for(i=0; i < RREGDATASIZE; i+=4)
   {
	  ChR[i] = RRegData[i];
	  ChR[i+1] = RRegData[i+1] + 0x10 * ChanNum;
	  ChR[i+2] = RRegData[i+2];
	  ChR[i+3] = RRegData[i+3];
   }

   //----- Save a copy of the Channel Offset used with Indexed Registers
   ChOff = (WORD)ChanNum * 0x1000;
   ChP->ChOff = ChOff;

  // follow was put in to avoid an obscure bug were no Rx data was seen
  // with the Fossil driver under very obscure circumstances. V1.50.007
  // break-out of Rx program, (let overrun), re-start fossil, err: no Rx
  sDisRxFIFO(ChP);
  sStopRxProcessor(ChP);
  sChanCommand(ChP, RESTXFCNT);    // reset Tx FIFO count
  sChanOutWI(ChP, _TXF_OUTP,  0);  // clear Tx in/out ptrs
  sChanCommand(ChP, RESRXFCNT);    // reset Rx FIFO count
  sChanOutWI(ChP, _RXF_OUTP,  0);  // clear Rx out ptrs
  sChanOutWI(ChP, _RXF_OUTP+2,0);  // clear Rx out ptrs
  sChanOutBI(ChP, _TXP_CNT,  0);   // clear TxPrioCnt
  sChanOutBI(ChP, _TXP_PNTR, 0);   // clear TxPrioPtr

   //------ Upload the Micro code to the Aiopic Channel
   for(i=0; i < RDATASIZE; i+=4)
   {
	  R[0] = RData[i];
	  R[1] = RData[i+1] + 0x10 * ChanNum;
	  R[2] = RData[i+2];
	  R[3] = RData[i+3];
	  sOutDW(ChP->IndexAddr,*((DWORD *)&R[0]));
   }

		   //----- default some indirect registers
   sChanOutWI(ChP, _BAUD, BRD9600);
   sChanOutWI(ChP, _TXREP1B1, 0);  // TxReplace1[4]
   sChanOutWI(ChP, _TXREP2,   0);  // TxReplace2[4]

		   //------- default some of our indirect Reg. Arrays
		   //------- and dump into hardware memory
   ChP->RxControl[0] = (BYTE)(ChOff + _RX_CTRL);
   ChP->RxControl[1] = (BYTE)((ChOff + _RX_CTRL) >> 8);
   ChP->RxControl[2] = 0;
   ChP->RxControl[3] = 0;
   sOutDW(ChP->IndexAddr,*(DWORD *)&ChP->RxControl[0]);

   ChP->TxControl[0] = (BYTE)(ChOff + _TX_CTRL);
   ChP->TxControl[1] = (BYTE)((ChOff + _TX_CTRL) >> 8);
   ChP->TxControl[2] = 0;
   ChP->TxControl[3] = 0;
   sOutDW(ChP->IndexAddr,*(DWORD *)&ChP->TxControl[0]);

   ChP->TxEnables[0] = (BYTE)(ChOff + _TX_ENBLS);
   ChP->TxEnables[1] = (BYTE)((ChOff + _TX_ENBLS) >> 8);
   ChP->TxEnables[2] = 0;
   ChP->TxEnables[3] = 0;
   sOutDW(ChP->IndexAddr,*(DWORD *)&ChP->TxEnables[0]);

   ChP->TxCompare[0] = (BYTE)(ChOff + _TXCMP1);
   ChP->TxCompare[1] = (BYTE)((ChOff + _TXCMP1) >> 8);
   ChP->TxCompare[2] = 0;
   ChP->TxCompare[3] = 0;
   sOutDW(ChP->IndexAddr,*(DWORD *)&ChP->TxCompare[0]);

   sChanCommand(ChP, RESTXFCNT);    // reset Tx FIFO count
   sChanOutWI(ChP, _TXF_OUTP,  0);  // clear Tx in/out ptrs
   sChanCommand(ChP, RESRXFCNT);    // reset Rx FIFO count
   sChanOutWI(ChP, _RXF_OUTP,  0);  // clear Rx out ptrs
   sChanOutWI(ChP, _RXF_OUTP+2,0);  // clear Rx out ptrs
   sChanOutBI(ChP, _TXP_CNT,  0);   // clear TxPrioCnt
   sChanOutBI(ChP, _TXP_PNTR, 0);   // clear TxPrioPtr

   sEnRxProcessor(ChP);             // start the Rx processor

   return 0;  // ok
}

/*-------------------------------------------------------------------
Function: sReadAiopID
Purpose:  Read the AIOP idenfication number directly from an AIOP.
Call:     sReadAiopID(io)
		  WORD io: AIOP base I/O address
Return:   int: Flag AIOPID_XXXX if a valid AIOP is found, where X
				 is replace by an identifying number.
		  Flag -1 if no valid AIOP is found
Warnings: No context switches are allowed while executing this function.

|--------------------------------------------------------------------*/
int sReadAiopID(WORD io)
{
   BYTE AiopID;               /* ID byte from AIOP */

   sOutB(io + _CMD_REG,RESET_ALL);     /* reset AIOP */
   sOutB(io + _CMD_REG,0x0);
   AiopID = sInB(io + _CHN_STAT0) & 0x07;
   if(AiopID == 0x06)
	  return(1);   // aiop exists
   else
	  return(-1);  /* AIOP does not exist */
}

/*-------------------------------------------------------------------------
Function: sDisAiop
Purpose:  Disable I/O access to an AIOP
Call:     sDisAiop(CltP)
		  CONTROLLER_T *b; Ptr to controller structure
		  int aiop_num; Number of AIOP on controller
Returns:  0 on sucess, otherwise non-zero error
|-------------------------------------------------------------------------*/
int sDisAiop(CONTROLLER_T *b, int aiop_num)
{
 BYTE cur;
 BYTE bit = (1 << aiop_num);

  if (b->BusType != BUS_PCI)  // pci has no mudback and aiops always enabled
  {
	cur = sInB(b->MBaseIO+3);
	if ((cur & bit) != 0)     // aiop enabled
	{
	  cur &= ~bit;
	  sOutB(b->MBaseIO+3,cur);  // so disable it
	}
	cur = sInB(b->MBaseIO+3);
	if ((cur & bit) != 0)
	  return 1; // error
  }
  return 0;
}

/*-------------------------------------------------------------------------
Function: sEnAiop
Purpose:  Enable I/O access to an AIOP
Call:     sEnAiop(CltP)
		  CONTROLLER_T *b; Ptr to controller structure
		  int AiopNum; Number of AIOP on controller
Returns:  0 on sucess, otherwise non-zero error
|-------------------------------------------------------------------------*/
int sEnAiop(CONTROLLER_T *b, int aiop_num)
{
 BYTE cur;
 BYTE bit = (1 << aiop_num);

  if (b->BusType != BUS_PCI)  // pci has no mudback and aiops always enabled
  {
	cur = sInB(b->MBaseIO+3);
	if ((cur & bit) == 0)     // aiop disabled
	{
	  cur |= bit;
	  sOutB(b->MBaseIO+3,cur);  // so enable it
	}
	cur = sInB(b->MBaseIO+3);
	if ((cur & bit) == 0)
	  return 1; // error
  }
  return 0;
}

/*-------------------------------------------------------------------
Function: sReadAiopNumChan
Purpose:  Read the number of channels available in an AIOP directly from
		  an AIOP.
Call:     sReadAiopNumChan(io)
		  WORD io: AIOP base I/O address
Return:   int: The number of channels available
Comments: The number of channels is determined by write/reads from identical
		  offsets within the SRAM address spaces for channels 0 and 4.
		  If the channel 4 space is mirrored to channel 0 it is a 4 channel
		  AIOP, otherwise it is an 8 channel.
Warnings: No context switches are allowed while executing this function.
|--------------------------------------------------------------------*/
int sReadAiopNumChan(WORD io)
{
  WORD x;
  DWORD v;

  v= 0x12340000L;
  sOutDW(io + _INDX_ADDR,v); /* write to chan 0 SRAM */

   sOutW(io + _INDX_ADDR,0);       /* read from SRAM, chan 0 */
   x = sInW(io + _INDX_DATA);

   sOutW(io + _INDX_ADDR,0x4000);  /* read from SRAM, chan 4 */
   if(x != sInW(io + _INDX_DATA))  /* if different must be 8 chan */
	  return(8);
   else
	  return(4);
}

/*-------------------------------------------------------------------
Function: sGetRxErrStatus
Purpose:  Get a channel's receive error status
Call:     sGetRxErrStatus(ChP)
		  CHANNEL_T *ChP; Ptr to channel structure
Return:   BYTE: Receive error status, can be 0 if there are no
						 errors or any combination of the following flags:
							 STMBREAK:   BREAK
							 STMFRAME:   framing error
							 STMRCVROVR: receiver over run error
							 STMPARITY:  parity error
Warnings: The channel must be in Rx Status Mode (see sEnRxStatusMode())
		  before calling this function.

		  No context switches are allowed while executing this function.
|--------------------------------------------------------------------*/
BYTE sGetRxErrStatus(CHANNEL_T *ChP)
{
   WORD RxFIFOOut;             /* Rx FIFO out status ptr */
   BYTE stat;

   SET_LOCK;
					 /* get Rx FIFO out status ptr */
   RxFIFOOut = sChanInWI(ChP, _RXF_OUTP) * 2 + 1;

   sOutW(ChP->IndexAddr,
		 ChP->ChOff+_RX_FIFO + RxFIFOOut); /* return the status */

   stat = (sInB(ChP->IndexData) & (STMBREAK | STMFRAME | STMPARITY | STMRCVROVR));
   CLEAR_LOCK;
   return stat;
}

/*-------------------------------------------------------------------
Function: sSetParity
Purpose:  Set parity to none, odd, or even.
Call:     sSetParity(ChP,Parity)
		  CHANNEL_T *ChP; Ptr to channel structure
		  int Parity; Parity, can be one of the following:
					  0: no parity
					  1: odd parity
					  2: even parity
Return:   void
Comments: Function sSetParity() can be used in place of functions sEnParity(),
		  sDisParity(), sSetOddParity(), and sSetEvenParity().
|--------------------------------------------------------------------*/
void sSetParity(CHANNEL_T *ChP,int Parity)
{
   if(Parity == 0)
   {
	  ChP->TxControl[2] &= ~PARITY_EN;
   }
   else if(Parity == 1)
   {
	  ChP->TxControl[2] |= PARITY_EN;

	  ChP->TxControl[2]  &= ~EVEN_PAR;
   }
   else if(Parity == 2)
   {
	  ChP->TxControl[2] |= (PARITY_EN | EVEN_PAR);
   }
   sOutDW(ChP->IndexAddr,*(DWORD *)&ChP->TxControl[0]);
}

/*-------------------------------------------------------------------
Function: sStopRxProcessor
Purpose:  Stop the receive processor from processing a channel.
Call:     sStopRxProcessor(ChP)
		  CHANNEL_T *ChP; Ptr to channel structure

Comments: The receive processor can be started again with sStartRxProcessor().
		  This function causes the receive processor to skip over the
		  stopped channel.  It does not stop it from processing other channels.

Warnings: No context switches are allowed while executing this function.

		  Do not leave the receive processor stopped for more than one
		  character time.

		  After calling this function a delay of 4 uS is required to ensure
		  that the receive processor is no longer processing this channel.
|--------------------------------------------------------------------*/
void sStopRxProcessor(CHANNEL_T *ChP)
{
   BYTE R[4];

   R[0] = ChP->R[0];
   R[1] = ChP->R[1];
   R[2] = 0x0a;
   R[3] = ChP->R[3];
   sOutDW(ChP->IndexAddr,*(DWORD *)&R[0]);
}

/*-------------------------------------------------------------------
Function: sStopSWInFlowCtl
Purpose:  Stop the receive processor from processing a channel's
		  software input flow control routine.
Call:     sStopSWInFlowCtl(ChP)
		  CHANNEL_T *ChP; Ptr to channel structure

Warnings: No context switches are allowed while executing this function.

		  After calling this function a delay of 1 uS is required to ensure
		  that the receive processor is no longer processing software input
		  flow control for this channel.
|--------------------------------------------------------------------*/
void sStopSWInFlowCtl(CHANNEL_T *ChP)
{
   BYTE R[4];

   R[0] = ChP->R[0];
   R[1] = ChP->R[1];
   R[2] = ChP->R[2];
   R[3] = 0x0a;
   sOutDW(ChP->IndexAddr,*(DWORD *)&R[0]);
}

/*-------------------------------------------------------------------
Function: sFlushRxFIFO
Purpose:  Flush the Rx FIFO
Call:     sFlushRxFIFO(ChP)
		  CHANNEL_T *ChP; Ptr to channel structure
Return:   void
Comments: To prevent data from being enqueued or dequeued in the Tx FIFO
		  while it is being flushed the receive processor is stopped
		  and the transmitter is disabled.  After these operations a
		  4 uS delay is done before clearing the pointers to allow
		  the receive processor to stop.  These items are handled inside
		  this function.
Warnings: No context switches are allowed while executing this function.
|--------------------------------------------------------------------*/
void sFlushRxFIFO(CHANNEL_T *ChP)
{
 int i;
 BYTE Ch;                   /* channel number within AIOP */
 int RxFIFOEnabled;             /* TRUE if Rx FIFO enabled */

   if (sGetRxCnt(ChP) == 0)       /* Rx FIFO empty */
	  return;                     /* don't need to flush */

   RxFIFOEnabled = FALSE;
   if(ChP->R[0x32] == 0x08) /* Rx FIFO is enabled */
   {
	  RxFIFOEnabled = TRUE;
	  sDisRxFIFO(ChP);                 /* disable it */
	  for(i=0; i < 2000/200; i++)  /* delay 2 uS to allow proc to disable FIFO*/
		 sInB(ChP->IntChan);    /* depends on bus i/o timing */
   }
   SET_LOCK;
   sInW(ChP->ChanStat);    /* clear any pending Rx errors in chan stat */

   sChanCommand(ChP, RESRXFCNT);    // apply reset Rx FIFO count

   sChanOutWI(ChP, _RXF_OUTP,  0);    // clear Rx out ptrs
   sChanOutWI(ChP, _RXF_OUTP+2,  0);  // clear Rx out ptrs

   if(RxFIFOEnabled)
	  sEnRxFIFO(ChP);                  /* enable Rx FIFO */
   CLEAR_LOCK;

}

/*-------------------------------------------------------------------
Function: sFlushTxFIFO
Purpose:  Flush the Tx FIFO
Call:     sFlushTxFIFO(ChP)
		  CHANNEL_T *ChP; Ptr to channel structure
Return:   void
Comments: To prevent data from being enqueued or dequeued in the Tx FIFO
		  while it is being flushed the receive processor is stopped
		  and the transmitter is disabled.  After these operations a
		  4 uS delay is done before clearing the pointers to allow
		  the receive processor to stop.  These items are handled inside
		  this function.
Warnings: No context switches are allowed while executing this function.
|--------------------------------------------------------------------*/
void sFlushTxFIFO(CHANNEL_T *ChP)
{
   int i;
   BYTE Ch;                   /* channel number within AIOP */
   int TxEnabled;                      /* TRUE if transmitter enabled */

   if(sGetTxCnt(ChP) == 0)             /* Tx FIFO empty */
	  return;                          /* don't need to flush */

   TxEnabled = FALSE;
   if(ChP->TxControl[3] & TX_ENABLE)
   {
	  TxEnabled = TRUE;
	  sDisTransmit(ChP);               /* disable transmitter */
   }
   sStopRxProcessor(ChP);              /* stop Rx processor */
   for(i = 0; i < 4000/200; i++)         /* delay 4 uS to allow proc to stop */
	  sInB(ChP->IntChan);  /* depends on bus i/o timing */

   Ch = (BYTE)ChP->ChanNum;
   SET_LOCK;
   sOutB(ChP->Cmd,Ch | RESTXFCNT);     /* apply reset Tx FIFO count */
   sOutB(ChP->Cmd,Ch);                 /* remove reset Tx FIFO count */

   sChanOutWI(ChP, _TXF_OUTP,  0);  // clear Tx in/out ptrs

   CLEAR_LOCK;
   if(TxEnabled)
	  sEnTransmit(ChP);                /* enable transmitter */
   sStartRxProcessor(ChP);             /* restart Rx processor */
}

/*-------------------------------------------------------------------
Function: sFlushTxPriorityBuf
Purpose:  Flush the Tx priority buffer
Call:     sFlushTxPriorityBuf(ChP,BYTE *Data)
		  CHANNEL_T *ChP; Ptr to channel structure
		  BYTE *Data; Next data byte to be transmitted from the
			 Tx priority buffer before the flush occurred, if any.  If
			 the return value is TRUE a byte is returned in "Data," if
			 the return value is FALSE nothing is returned in "Data."
Return:   int: TRUE if there was data in the Tx priority buffer before
			   the flush occurs.  In this case the next byte that would
			   have been transmitted is returned in the "Data" parameter.
			   FALSE if there was no data in the Tx priority buffer before
			   the flush.
Comments: This flush returns the next byte in the priority buffer to
		  allow that byte to be sent via sOutB() after all
		  transmit flushing is complete.  This is done to allow pending
		  XON and XOFF bytes to be transmitted regardless of the flush.
Warnings: No context switches are allowed while executing this function.
|--------------------------------------------------------------------*/
int sFlushTxPriorityBuf(CHANNEL_T *ChP, BYTE *Data)
{
   WORD PrioState;       /* Tx prio buf status, count, and pointer */
   WORD BufOff;          /* Offset of next data byte in Tx prio buf */

   sDisTransmit(ChP);
   SET_LOCK;

   PrioState = sChanInWI(ChP, _TXP_CNT);    /* get priority buf status */

   if (PrioState & PRI_PEND)            /* data in Tx prio buf */
   {
	  BufOff = PrioState >> 8;   /* get offset of next data byte in buf */
 
	  *Data = sChanInBI(ChP, _TXP_BUF+BufOff); /* return next data byte */

	  CLEAR_LOCK;
	  sEnTransmit(ChP);
	  return(TRUE);
   }
   CLEAR_LOCK;
   sEnTransmit(ChP);                   /* no data in Tx prio buf */
   return(FALSE);
}

/*-------------------------------------------------------------------
Function: sGetTxPriorityCnt
Purpose:  Get the number of data bytes in the Tx priority buffer
Call:     sGetTxPriorityCnt(ChP)
		  CHANNEL_T *ChP; Ptr to channel structure
Return:   BYTE: The number of data bytes in the Tx FIFO.
Warnings: No context switches are allowed while executing this function.
|--------------------------------------------------------------------*/
BYTE sGetTxPriorityCnt(CHANNEL_T *ChP)
{
   BYTE Cnt;

   Cnt = sChanInBI(ChP, _TXP_CNT); /* get priority buf status */

   if(Cnt & PRI_PEND)
	  return(Cnt & 0x1f);              /* only lower 5 bits contain count */
   else
	  return(0);
}

/*-------------------------------------------------------------------
Function: sWriteTxPrioByte
Purpose:  Write a byte of priority transmit data to a channel
Call:     sWriteTxPrioByte(ChP,Data)
		  CHANNEL_T *ChP; Ptr to channel structure
		  BYTE Data; The transmit data byte

Return:   int: 1 if the bytes is successfully written, otherwise 0.

Comments: The priority byte is transmitted before any data in the Tx FIFO.

Warnings: No context switches are allowed while executing this function.
|--------------------------------------------------------------------*/
int sWriteTxPrioByte(CHANNEL_T *ChP, BYTE Data)
{
 BYTE Cnt;

  if (sGetTxCnt(ChP) == 0)
  {
	sOutB(ChP->TxRxData,Data);
	return 1; /* 1 byte sent */
  }

						  /* see if Tx priority buffer busy */

  Cnt = sChanInBI(ChP, _TXP_CNT); /* get priority buf status */
  if (Cnt & PRI_PEND) /* priority buffer busy */
	return 0;    /* nothing sent */
							
  sChanOutWI(ChP, _TXP_BUF, Data);  /* write it to Tx priority buffer */
  sChanOutWI(ChP, _TXP_CNT, PRI_PEND | 1); /* indicate 1 byte pending */

  return 1;   /* 1 byte sent */
}

#ifndef NO_INTERRUPT_PROCS
/*-------------------------------------------------------------------
Function: sEnInterrupts
Purpose:  Enable one or more interrupts for a channel
Call:     sEnInterrupts(ChP,Flags)
		  CHANNEL_T *ChP; Ptr to channel structure
		  WORD Flags: Interrupt enable flags, can be any combination
			 of the following flags:
				TXINT_EN:   Interrupt on Tx FIFO empty
				RXINT_EN:   Interrupt on Rx FIFO at trigger level (see
							sSetRxTrigger())
				SRCINT_EN:  Interrupt on SRC (Special Rx Condition)
				MCINT_EN:   Interrupt on modem input change
				CHANINT_EN: Allow channel interrupt signal to the AIOP's
							Interrupt Channel Register.
Return:   void
Comments: If an interrupt enable flag is set in Flags, that interrupt will be
		  enabled.  If an interrupt enable flag is not set in Flags, that
		  interrupt will not be changed.  Interrupts can be disabled with
		  function sDisInterrupts().

		  This function sets the appropriate bit for the channel in the AIOP's
		  Interrupt Mask Register if the CHANINT_EN flag is set.  This allows
		  this channel's bit to be set in the AIOP's Interrupt Channel Register.

		  Interrupts must also be globally enabled before channel interrupts
		  will be passed on to the host.  This is done with function
		  sEnGlobalInt().

		  In some cases it may be desirable to disable interrupts globally but
		  enable channel interrupts.  This would allow the global interrupt
		  status register to be used to determine which AIOPs need service.
|--------------------------------------------------------------------*/
void sEnInterrupts(CHANNEL_T *ChP,WORD Flags)
{
   BYTE Mask;                 /* Interrupt Mask Register */

   ChP->RxControl[2] |=
	  ((BYTE)Flags & (RXINT_EN | SRCINT_EN | MCINT_EN));

   sOutDW(ChP->IndexAddr,*(DWORD *)&ChP->RxControl[0]);

   ChP->TxControl[2] |= ((BYTE)Flags & TXINT_EN);

   sOutDW(ChP->IndexAddr,*(DWORD *)&ChP->TxControl[0]);

   if(Flags & CHANINT_EN)
   {
	  Mask = sInB(ChP->IntMask) | (1 << ChP->ChanNum);
	  sOutB(ChP->IntMask,Mask);
   }
}

/*-------------------------------------------------------------------
Function: sDisInterrupts
Purpose:  Disable one or more interrupts for a channel
Call:     sDisInterrupts(ChP,Flags)
		  CHANNEL_T *ChP; Ptr to channel structure
		  WORD Flags: Interrupt flags, can be any combination
			 of the following flags:
				TXINT_EN:   Interrupt on Tx FIFO empty
				RXINT_EN:   Interrupt on Rx FIFO at trigger level (see
							sSetRxTrigger())
				SRCINT_EN:  Interrupt on SRC (Special Rx Condition)
				MCINT_EN:   Interrupt on modem input change
				CHANINT_EN: Disable channel interrupt signal to the
							AIOP's Interrupt Channel Register.
Return:   void
Comments: If an interrupt flag is set in Flags, that interrupt will be
		  disabled.  If an interrupt flag is not set in Flags, that
		  interrupt will not be changed.  Interrupts can be enabled with
		  function sEnInterrupts().

		  This function clears the appropriate bit for the channel in the AIOP's
		  Interrupt Mask Register if the CHANINT_EN flag is set.  This blocks
		  this channel's bit from being set in the AIOP's Interrupt Channel
		  Register.
|--------------------------------------------------------------------*/
void sDisInterrupts(CHANNEL_T *ChP,WORD Flags)
{
   BYTE Mask;                 /* Interrupt Mask Register */

   ChP->RxControl[2] &=
		 ~((BYTE)Flags & (RXINT_EN | SRCINT_EN | MCINT_EN));
   sOutDW(ChP->IndexAddr,*(DWORD *)&ChP->RxControl[0]);
   ChP->TxControl[2] &= ~((BYTE)Flags & TXINT_EN);
   sOutDW(ChP->IndexAddr,*(DWORD *)&ChP->TxControl[0]);

   if(Flags & CHANINT_EN)
   {
	  Mask = sInB(ChP->IntMask) & ~(1 << ChP->ChanNum);
	  sOutB(ChP->IntMask,Mask);
   }

}

/*-------------------------------------------------------------------------
Function: sEnGlobalInt
Purpose:  Enable global interrupts for a controller
Call:     sEnGlobalInt(CtlP)
		  CONTROLLER_T *CtlP; Ptr to controller structure
|-------------------------------------------------------------------------*/
void sEnGlobalInt(CONTROLLER_T *CtlP)
{
  if (CtlP->BusType != BUS_PCI)
  {
	CtlP->MReg2 |= INTR_EN;
	sOutB(CtlP->MReg2IO, CtlP->MReg2);
  }
  else
  {
	CtlP->Reg3A |= PCI_INT_EN;
	sOutW(CtlP->Reg3AIO,CtlP->Reg3A);
  }
}

/*-------------------------------------------------------------------------
Function: sDisGlobalInt
Purpose:  Disable global interrupts for a controller
Call:     sDisGlobalInt(CtlP)
		  CONTROLLER_T *CtlP; Ptr to controller structure
|-------------------------------------------------------------------------*/
void sDisGlobalInt(CONTROLLER_T *CtlP)
{
  if (CtlP->BusType != BUS_PCI)
  {
	CtlP->MReg2 &= ~INTR_EN;
	sOutB(CtlP->MReg2IO,CtlP->MReg2);
  }
  else
  {
	CtlP->Reg3A &= ~PCI_INT_EN;
	sOutW(CtlP->Reg3AIO,CtlP->Reg3A);
  }
}
#endif

/*------------------------------------------------------------------
 sSetBaudRate - Set the desired baud rate.  Return non-zero on error.
|-------------------------------------------------------------------*/
int sSetBaudRate(CHANNEL_T *ChP,
				 DWORD desired_baud,
				 BOOL SetHardware)
{
 DWORD diff;
 DWORD act_baud;
 DWORD percent_error;
 DWORD div;
 DWORD base_clock_rate;

  switch (ChP->CtlP->clk_prescaler & 0xf)
  {
	case 2:  // assume rplus hardware
	  //base_clock_rate = ((44236800L/32) / ((ChP->CtlP->clk_prescaler & 0xf)+1));
	  base_clock_rate = ((44236800L/16) / ((ChP->CtlP->clk_prescaler & 0xf)+1));
	break;

	default:
	  base_clock_rate = ((36864000L/16) / ((ChP->CtlP->clk_prescaler & 0xf)+1));
	break;
  }
	 // calculate the divisor for our hardware register.
	 // this is really just div = clk/desired_baud -1.  but we do some
	 // work to minimize round-off error.
  if (desired_baud <= 0) desired_baud = 1;  // guard against div 0

  div =  ((base_clock_rate+(desired_baud>>1)) / desired_baud) - 1;
  if (div > 8191)  // overflow hardware divide register
	div = 8191;

  // this is really just (clk) / (div+1) but we do some
	 // work to minimize round-off error.
  act_baud = (base_clock_rate+((div+1)>>1)) / (div+1);

  if (desired_baud > act_baud)
	   diff = desired_baud - act_baud;
  else diff = act_baud - desired_baud;

  percent_error = (diff * 100) / desired_baud;
  if (percent_error > 5)
	return (int) percent_error;

  if (SetHardware)
  {
	sChanOutWI(ChP, _BAUD, (WORD) div);
  }
  return 0;
}

/*------------------------------------------------------------------
Function: sChanOutWI
Purpose:  Write an Indirect Register on the Rocket Port Board
Call:     sChanOutWI(CHANNEL_T *ChP, WORD RegNum,  WORD val)
		  CHANPTR_T ChP; Ptr to channel structure
		  WORD RegNum;   Indirect Register Number to Write
		  WORD val;      Value to Write.
Return:   void
Comments: The write DWORD is preferred since the whole operation occurs
		  in one processor shot, thus avoiding the need for contention
		  handling in multitasking environments.
|-------------------------------------------------------------------*/
void sChanOutWI(CHANNEL_T *ChP, WORD RegNum,  WORD val)
{
  WORD m[2];

   m[0] = ChP->ChOff + RegNum;
   m[1] = val;
   sOutDW(ChP->IndexAddr,*(DWORD *)&m[0]);
}

/*------------------------------------------------------------------
Function: sChanOutBI
Purpose:  Write an Indirect Register Byte on the Rocket Port Board
Call:     sChanOutBI(CHANNEL_T *ChP, WORD RegNum,  BYTE val)
		  CHANPTR_T ChP; Ptr to channel structure
		  WORD RegNum;   Indirect Register Number to Write
		  BYTE val;      Value to Write.
Return:   void
|-------------------------------------------------------------------*/
void sChanOutBI(CHANNEL_T *ChP, WORD RegNum,  BYTE val)
{
  sOutW(ChP->IndexAddr, ChP->ChOff + RegNum);
  sOutB(ChP->IndexData, val);
}

/*------------------------------------------------------------------
Function: sChanInBI
Purpose:  Read an Indirect Register Byte on the Rocket Port Board
Call:     sChanInBI(CHANNEL_T *ChP, WORD RegNum)
		  CHANPTR_T ChP; Ptr to channel structure
		  WORD RegNum;   Indirect Register Number to Read
Return:   BYTE
|-------------------------------------------------------------------*/
BYTE sChanInBI(CHANNEL_T *ChP, WORD RegNum)
{
 BYTE val;

  sOutW(ChP->IndexAddr, ChP->ChOff + RegNum);
  val = sInB(ChP->IndexData);

  return val;
}

/*------------------------------------------------------------------
Function: sChanInWI
Purpose:  Read an Indirect Register Byte on the Rocket Port Board
Call:     sChanInBI(CHANNEL_T *ChP, WORD RegNum)
		  CHANPTR_T ChP; Ptr to channel structure
		  WORD RegNum;   Indirect Register Number to Read
Return:   WORD
|-------------------------------------------------------------------*/
WORD sChanInWI(CHANNEL_T *ChP, WORD RegNum)
{
 WORD val;

  sOutW(ChP->IndexAddr, ChP->ChOff + RegNum);
  val = sInW(ChP->IndexData);

  return val;
}

/*------------------------------------------------------------------
Function: sChanCommand
Purpose:  Issue a command to the Aiopic Channel
Call:     sChanCommand(CHANNEL_T *ChP, BYTE command)
		  CHANPTR_T ChP; Ptr to channel structure
		  BYTE command;  Command to Write
Return:   void
|-------------------------------------------------------------------*/
void sChanCommand(CHANNEL_T *ChP, BYTE command)
{
 BYTE val;

  val = command | ChP->ChanNum;
  sOutB(ChP->Cmd,val); // apply command
  val = ChP->AiopNum;
  sOutB(ChP->Cmd,val); // remove command
}

#ifndef NO_BLK_RD_WR
/*-------------------------------------------------------------------
Function: sReadRxBlk
Purpose:  Read a block of receive data from a channel
Call:     sReadRxBlk(ChP,Buffer,Count)
		  CHANNEL_T *ChP; Ptr to channel structure
		  BYTE *Buffer; Ptr to buffer for receive data
		  int Count; Max number of bytes to read
Return:   int: Number of bytes actually read from the channel
Comments: Implemented as a macro.
Warnings: Buffer must be large enough to hold Count characters.

		  This function must not be called when in Rx Status Mode.
|--------------------------------------------------------------------*/
int sReadRxBlk(CHANNEL_T *ChP,BYTE *Buffer,int Count)
{
   int RetCount;
   int WordCount;
   WORD *WordP;
   WORD io;

   RetCount = sGetRxCnt(ChP);          /* number bytes in Rx FIFO */

   /* are there pending chars? */
   if(RetCount <= 0)                   /* no data available */
	  return(0x0);
   if(RetCount > Count)                /* only dequeue as much as requested */
	  RetCount = Count;

   WordCount = RetCount >> 1;          /* compute count as words */
   WordP = (WORD *)Buffer;     /* word ptr to buffer */

   io = ChP->TxRxData;
   sInStrW(io,WordP,WordCount);
   if(RetCount & 1)                    /* odd count */
	  Buffer[RetCount - 1] = sInB(io); /* read last byte */

   return(RetCount);
}

/*-------------------------------------------------------------------
Function: sWriteTxBlk
Purpose:  Write a block of transmit data to a channel
Call:     sWriteTxBlk(ChP,Buffer,Count)
		  CHANNEL_T *ChP; Ptr to channel structure
		  BYTE *Buffer; Ptr to buffer containing data to transmit
		  int Count; Size of buffer in bytes
Return:   int: Number of bytes actually written to the channel
|--------------------------------------------------------------------*/
int sWriteTxBlk(CHANNEL_T *ChP,BYTE *Buffer,int Count)
{
   int RetCount;
   int WordCount;
   WORD *WordP;

#define MAXTX_SIZE 250

   RetCount = MAXTX_SIZE - (int)sGetTxCnt(ChP); /* open space in Tx FIFO*/
   if(RetCount <= 0)                   /* no space available */
	  return(0x0);
   if(RetCount > Count)
	  RetCount = Count;                /* only enqueue as much as requested */

   WordCount = RetCount >> 1;          /* compute count as words */
   WordP = (WORD *)Buffer;    /* word ptr to buffer */

   /* Write the data */
   sOutStrW(ChP->TxRxData,WordP,WordCount);
   if(RetCount & 1)                    /* odd count */
	  sOutB(ChP->TxRxData,(BYTE)*WordP); /* send last byte */

   return(RetCount);
}
#endif
