/*------------------------------------------------------------------
 rckt.h - rocket port hardware routines, code is os independent.

11-05-95 more work making it comprehendable - kpb.
10-09-95 strip out some more junk - kpb.
9-20-95 add PCI, clean up some misc things - kpb.
8-08-95 rewrite, strip out bogus typedefs, extra globals, etc. - kpb
LAST UPDATE: 9-20-95  0:16 A.M.
 Copyright (c) 1995 Comtrol Corporation
|---------------------------------------------------------------------*/
#ifdef __cplusplus
extern "C" {            /* Assume C declarations for C++ */
#endif  /* __cplusplus */

#define RPLUS_SUPPORT

#define ROCKET_H

#define AIOP_CTL_SIZE 4
#define CHAN_AIOP_SIZE 8
#define MAX_PORTS_PER_AIOP 8
#define MAX_AIOPS_PER_BOARD 4
#define MAX_PORTS_PER_BOARD 32

#define CTLID_NULL -1
/*----------------------------------------------------------------------
The following are io-register address definitions for the "AIOPIC" chip.
A single AIOPIC chip can handle 8 ports.  And a board(controller) can have
up to 4 AIOPIC chips thus 32 ports(channels).  The AIOPIC chip takes up about
40H bytes of IO-ADDRESS SPACE as follows:

 0- FH - Transm/Recv Data Registers for Ports 0-7(Word Length)
10-1FH - Transm/Recv FIFO Count Registers for Ports 0-7(Word Length)
20-2FH - Channel Status Registers for Ports 0-7(Word Length)
30-37H - Channel Interrupt Identification Registers for Ports 0-7(Word Length)
   38H - Command Register
   39H - Interrupt Channel Identification
   3AH - Interrupt Mask  \ __ PCI(first AIOP a WORD read/write used
   3BH - Reserved        /        for board level functions.)
   3CH - Index Register Address
   3EH - Index Register Data

All other AIOPIC chips(on our same board or an aditional board) are
addressed by aliasing using the normally unused high address bits for
ISA BUS cards, other buses use consequative 40H blocks.

The AIOPIC Indexed Registers(using 3CH/3EH registers) are as follows:
 XFF0H - Transmit Control(Word)
 XFF2H - Receive Control(Byte)
 XFF3H - Unused(Byte)
 XFF4H - Baud Rate(Word)
 XFF6H - Clock Prescaler(Byte)
 XFF7H - Unused(Byte)

The MUDBACK chip(for ISA & MCA buses) handles overall board configuration
and status and takes up 4 bytes of register space per AIOPIC chip.
In an ISA bus the MUDBACK is addressed immediately after the 40H byte
register space of the first AIOPIC, and remaining MUDBACKS are aliased
in 400H byte increments.  It is the MUDBACK chip address which is actually
set by the rocker switches on the ISA version.  All AIOPIC addressing is
setup via the first register in the mudback(for ISA bus).
 0H    - mudback AIOPIC Register Base Address, Read/Write
 1H    - mudback Aiopic Interrupt Status Register, Read
 2H    - mudback Interrupt Enable Register, Read/Write
 3H    - mudback Aiopic Enable Register, Read/Write

---------------------------------------------------------------------------*/
/*--- Channel Register Offsets for 1st channel in AIOP - Direct Access */
#define _TD0       0x00  /* Transmit Data               16   Write */
#define _RD0       0x00  /* Receive Data                16   Read */
#define _FIFO_CNT0 0x10  /* Transmit/Receive FIFO Count 16   Read */
#define _CHN_STAT0 0x20  /* Channel Status              8/16 Read / Write */
#define _INT_ID0   0x30  /* Interrupt Identification    8    Read */

/*--- Global Register Offsets - Direct Access - Fixed values */
#define _CMD_REG   0x38   /* Command Register            8    Write */
#define _INT_CHAN  0x39   /* Interrupt Channel Register  8    Read */
#define _INT_MASK  0x3A   /* BYTE ACCESS, 
                              Interrupt Mask Register     8    Read / Write */

#define _PCI_INT_FUNC 0x3A  /* AIOPIC Status & Interrupt Strobe Read/Write,
                               WORD access(PCI only) ISA different */

#define _UNUSED    0x3B     /* Unused                      8 */

#define _INDX_ADDR 0x3C   /* Index Register Address      16   Write */
#define _INDX_DATA 0x3E   /* Index Register Data         8/16 Read / Write */


/*--- Tx Control Register Offsets - Indexed - External - Fixed */
#define _TX_ENBLS  0x980    /* Tx Processor Enables Register 8 Read / Write */
#define _TXCMP1    0x988    /* Transmit Compare Value #1     8 Read / Write */
#define _TXCMP2    0x989    /* Transmit Compare Value #2     8 Read / Write */
#define _TXREP1B1  0x98A    /* Tx Replace Value #1 - Byte 1  8 Read / Write */
#define _TXREP1B2  0x98B    /* Tx Replace Value #1 - Byte 2  8 Read / Write */
#define _TXREP2    0x98C    /* Transmit Replace Value #2     8 Read / Write */

/*--- Memory Controller Register Offsets - Indexed - External - Fixed */
#define _RX_FIFO    0x000    /* Rx FIFO */
#define _TX_FIFO    0x800    /* Tx FIFO */
#define _RXF_OUTP   0x990    /* Rx FIFO OUT pointer        16 Read / Write */
#define _RXF_INP    0x992    /* Rx FIFO IN pointer         16 Read / Write */
#define _TXF_OUTP   0x994    /* Tx FIFO OUT pointer        8  Read / Write */
#define _TXF_INP    0x995    /* Tx FIFO IN pointer         8  Read / Write */
#define _TXP_CNT    0x996    /* Tx Priority Count          8  Read / Write */
#define _TXP_PNTR   0x997    /* Tx Priority Pointer        8  Read / Write */

#define PRI_PEND    0x80     /* Priority data pending (bit7, Tx pri cnt) */
#define TXFIFO_SIZE 255      /* size of Tx FIFO */
#define RXFIFO_SIZE 1023     /* size of Rx FIFO */

/*--- Tx Priority Buffer - Indexed - External - Fixed */
#define _TXP_BUF    0x9C0    /* Tx Priority Buffer  32  Bytes   Read / Write */
#define TXP_SIZE    0x20     /* 32 bytes */


/*---- Channel Register Offsets - Indexed - Internal - Fixed ----*/
#define _TX_CTRL    0xFF0    /* Transmit Control               16  Write */
#define _RX_CTRL    0xFF2    /* Receive Control                 8  Write */
#define _BAUD       0xFF4    /* Baud Rate                      16  Write */
#define _CLK_PRE    0xFF6    /* Clock Prescaler                 8  Write */

//----------------------------- DIRECT REGISTERS
/*---- 0-FH Rec. Data & Status --*/
#define STMBREAKH   0x800      /* BREAK */
#define STMFRAMEH   0x400      /* framing error */
#define STMRCVROVRH 0x200      /* receiver over run error */
#define STMPARITYH  0x100      /* parity error */
#define STMERRORH   (STMBREAKH | STMFRAMEH | STMPARITYH)
/* same but you need to shift it down(data rec is in low byte) */
#define STMBREAK   0x08        /* BREAK */
#define STMFRAME   0x04        /* framing error */
#define STMRCVROVR 0x02        /* receiver over run error */
#define STMPARITY  0x01        /* parity error */
#define STMERROR   (STMBREAK | STMFRAME | STMPARITY)

/*---- 20-2FH Direct - Channel Status Reg. */
#define CTS_ACT   0x20        /* CTS input asserted */
#define DSR_ACT   0x10        /* DSR input asserted */
#define CD_ACT    0x08        /* CD input asserted */
#define TXFIFOMT  0x04        /* Tx FIFO is empty */
#define TXSHRMT   0x02        /* Tx shift register is empty */
#define RDA       0x01        /* Rx data available */
#define DRAINED (TXFIFOMT | TXSHRMT)  /* indicates Tx is drained */
#define STATMODE  0x8000      /* status mode enable bit */
#define RXFOVERFL 0x2000      /* receive FIFO overflow */
#define RX2MATCH  0x1000      /* receive compare byte 2 match */
#define RX1MATCH  0x0800      /* receive compare byte 1 match */
#define RXBREAK   0x0400      /* received BREAK */
#define RXFRAME   0x0200      /* received framing error */
#define RXPARITY  0x0100      /* received parity error */
#define STATERROR (RXBREAK | RXFRAME | RXPARITY)

/*---- 30-37H Direct - Interrupt Identification Register */
#define RXF_TRIG  0x20        /* Rx FIFO trigger level interrupt */
#define TXFIFO_MT 0x10        /* Tx FIFO empty interrupt */
#define SRC_INT   0x08        /* special receive condition interrupt */
#define DELTA_CD  0x04        /* CD change interrupt */
#define DELTA_CTS 0x02        /* CTS change interrupt */
#define DELTA_DSR 0x01        /* DSR change interrupt */

/*----- 38H Direct Command Register */
#define RESET_ALL 0x80        /* reset AIOP (all channels) */
#define TXOVERIDE 0x40        /* Transmit software off override */
#define RESETUART 0x20        /* reset channel's UART */
#define RESTXFCNT 0x10        /* reset channel's Tx FIFO count register */
#define RESRXFCNT 0x08        /* reset channel's Rx FIFO count register */
/* bits 1,2,4 select channel for action bits */

/*----- 39H - Interrupt Channel Identification */
// 1H - bit when set indicates channel 0 has interrupt pending
// 2H - bit when set indicates channel 1 has interrupt pending
// .... up to channel 7

/*----- 3AH PCI, WORD access, Interrupt Aiopic Status and Strobe Command Register */
#define PCI_PERIODIC_FREQ     0x0007    // setup periodic
#define PCI_PERIODIC_ONLY     0x0008    // disable AIOPIC interrupts
#define PCI_INT_EN            0x0010    // enable global interrupts
#define PCI_AIOPIC_INT_STATUS 0x0f00    // 1bit=Aiop1, 2bit=Aiop2,etc
#define PCI_OUTPUT_STATE      0x1000    // interrupt status from board
#define PCI_STROBE            0x2000    // clear interrupt from board

// This bit used in sEnInterrupts(), sDisInterrupts().  No direct hardware
// bit, is translated for use with Interrupt Mask Register 0x3A register.
#define CHANINT_EN 0x0100           /* flags to enable/disable channel ints */


//----------------------------- INDIRECT REGISTERS
/*----- X980H - Transmit Processor Enable(Byte), WRITE */
// TxEnables[2]
#define REP1W2_EN 0x10        /* replace byte 1 with 2 bytes enable */
#define IGN2_EN   0x08        /* ignore byte 2 enable */
#define IGN1_EN   0x04        /* ignore byte 1 enable */
#define COMP2_EN  0x02        /* compare byte 2 enable */
#define COMP1_EN  0x01        /* compare byte 1 enable */

/*----- XFF0H - Indirect, Transmit Control(Word), Low Byte, WRITE */
// TxControl[2]
#define CTSFC_EN  0x80        /* CTS flow control enable bit */
#define RTSTOG_EN 0x40        /* RTS toggle enable bit */

#define TXINT_EN  0x10        /* transmit interrupt enable */
#define STOP2     0x08        /* enable 2 stop bits (0 = 1 stop) */
#define PARITY_EN 0x04        /* enable parity (0 = no parity) */
#define EVEN_PAR  0x02        /* even parity (0 = odd parity) */
#define DATA8BIT  0x01        /* 8 bit data (0 = 7 bit data) */

/*----- XFF1H - Indirect, Transmit Control(Word), High Byte, WRITE */
// TxControl[3]
#define TX_SETBREAK  0x10        /* send break condition (must clear) */
#define LOCALLOOP 0x08        /* local loopback set for test */
#define SET_DTR   0x04        /* assert DTR */
#define SET_RTS   0x02        /* assert RTS */
#define TX_ENABLE 0x01        /* enable transmitter */

/*----- XFF2H - Indirect, Receive Control(Byte), WRITE */
// RxControl[2]
#define RTSFC_EN  0x40        /* RTS flow control enable */
#define RXPROC_EN 0x20        /* receive processor enable */
#define TRIG_7_8  0x18        /* trigger level 7/8 */
#define TRIG_MASK 0x18        /* trigger level mask */
#define TRIG_1_2  0x10        /* trigger level 1/2 */
#define TRIG_1    0x08        /* trigger level 1 char */
#define TRIG_NO   0x00        /* Rx FIFO trigger level 0 (no trigger) */
#define SRCINT_EN 0x04        /* special Rx condition interrupt enable */
#define RXINT_EN  0x02        /* Rx interrupt enable */
#define MCINT_EN  0x01        /* modem change interrupt enable */

/*---- XFF6H - Clock Prescaler Register, Global Indexed(1 per board) */
// RocketPort base clock freq: 36.864 Mhz, this value is divided down
//     by 16 prior to getting to the AIOP chip.
// RocketPortPlus base clock freq: 44.2368 Mhz, this value is divided down
//     by 32 prior to getting to the AIOP chip.

// Bits 0-3 define baud rate clock prescaler:
//          where 0 = div 1, 1= div 2, 2= div 3, 3= div 4(value+1)

// bits 4-7 are sysclkout, mux rate, 0=nowork, 1=18mhz, 2=12.5mhz, 3=9mhz
//          this sets the mux-rate out to the external interface unit

// following values used for modified PCI clock rates and tests:
#ifdef RPLUS_SUPPORT
// clock prescaler for normal RocketPort crystal is 5, this allows 460800.
// clock prescaler for Rplus  RocketPort crystal is 3, this allows 921600.
#define RCKT_NORMAL_CLK_RATE  ((36864000L/16) / ((clk_prescaler & 0xf)+1))
#define RCKT_RPLUS_CLK_RATE   ((44236800L/32) / ((clk_prescaler & 0xf)+1))
#define RPLUS_CLOCK_PRESC  0x12 /* div by 3 baud prescale, 921600, crystal:44.2368Mhz */
#define NORMAL_CLOCK_PRESC 0x14 /* div 5 prescale, max 460800 baud(NO 50baud!) */
#else  // normal rocketport
// standard :
#define CLOCK_PRESC 0x14 /* div 5 prescale, max 460800 baud(NO 50baud!) */
#define RCKT_CLK_RATE   (2304000L / ((clk_prescaler & 0xf)+1))
#endif

#define OLD_CLOCK_PRESC 0x19 /* div 10 prescale, max 230400 baud */
#define BRD9600           47

//-------------------- ISA/MCA MUDBACK IO Registers
/* 0H - mudback AIOPIC Register Base Address, Read/Write
     // this byte sets the AIOPIC chip 40H address space base address
     // where 1 = 40H, 2=80H, 3=C0H, 4=100H, 5=140H, 6=180H.....

//----- 1H - mudback Aiopic Interrupt Status Register, Read
#define INTSTAT0  0x01        /* AIOP 0 interrupt status */
#define INTSTAT1  0x02        /* AIOP 1 interrupt status */
#define INTSTAT2  0x04        /* AIOP 2 interrupt status */
#define INTSTAT3  0x08        /* AIOP 3 interrupt status */

//----- 2H - mudback Interrupt Enable Register, Read/Write
// 4H bit always reads back as 0, others read & write
#define IRQ_BITS  0xF0    // irq setup bits, where */
                          // 0=none,1=IRQ3,2=IRQ4,3=IRQ5,4=IRQ9, 5=IRQ10
                          // 6=IRQ11, 7=IRQ12, 8=IRQ15
#define INTR_EN   0x08    // 1=Enable global interrupts from board
#define INT_STROB 0x04    // 1=strobe and clear interrupt line (EOI)
#define AIOP_EN_BITS 0x03 // 0=aiop0, 1=aiop1, 2="", 3=""

//----- 3H - mudback Aiopic Enable Register, Read/Write
#define CHAN3_EN  0x08        /* enable AIOP 3 */
#define CHAN2_EN  0x04        /* enable AIOP 2 */
#define CHAN1_EN  0x02        /* enable AIOP 1 */
#define CHAN0_EN  0x01        /* enable AIOP 0 */
#define FREQ_DIS  0x00
#define FREQ_560HZ 0x70
#define FREQ_274HZ 0x60
#define FREQ_137HZ 0x50
#define FREQ_69HZ  0x40
#define FREQ_34HZ  0x30
#define FREQ_17HZ  0x20
#define FREQ_9HZ   0x10
#define PERIODIC_ONLY 0x80    /* only PERIODIC interrupt */

//--------------- MicroCode Space Sizes,  Where RDATASIZE is size
// of single complete array of microcode.  RREGDATASIZE is size of
// select register/index [4] byte arrays used on a per-channel basis
#define RDATASIZE 72
#define RREGDATASIZE 52

//------------------ Controller(Board) level structure
typedef struct
{
  BYTE BusType;   // 0=BUS_ISA, 1=BUS_MCA, 2=BUS_PCI
  BYTE NumPorts;  // number of ports on ctlr board

                  // following mudback regs are not on PCI version of card
  WORD MBaseIO;  // addr of 1st mudback register
  WORD MReg1IO;  // addr of 2nd mudback register
  WORD MReg2IO;  // addr of 3rd mudback register
  BYTE MReg2;    // copy of MUDBAC register 2, ISA
  BYTE CardNum;    // 0,1,2 - card count per BUS-type

  WORD Reg3A;      // PCI AIOPIC per. int. & int. status, contents of write
  WORD Reg3AIO; // PCI AIOPIC per. int. & int. status, address(fast)
  WORD NumAiop;                       // number of aiopics on ctlr board
  WORD AiopIO[AIOP_CTL_SIZE];         // list of aiopic regs
  WORD AiopIntChanIO[AIOP_CTL_SIZE];  // addr of aiopic interrupt status
  WORD AiopNumChan[AIOP_CTL_SIZE];    // number of channels on each aiopic
  BYTE PortsPerAiop;   // normally 8, but rplus this is 4
  BYTE clk_prescaler;  // 4 for normal, 2 for rplus
} CONTROLLER_T;

//------------------ Channel(Port) level structure
typedef struct
{
  CONTROLLER_T  *CtlP; // ptr to parent controller struct
  WORD AiopNum;        // AIOP number on the controller
  WORD ChanNum;        // channel within AIOP

                 // AIOP's global direct register addresses
  WORD Cmd;              // AIOP's Command register, +38H
  WORD IntChan;          // AIOP's Interrupt channel register, +39H
  WORD IntMask;          // AIOP's Interrupt mask register, +3AH
  WORD IndexAddr;        // AIOP's Index Register, +3CH
  WORD IndexData;        // AIOP's Index Register Data +3EH

                 // Channel's direct register addresses
  WORD TxRxData;     // Transmit and Receive register address, +0H+(ch*4)
  WORD ChanStat;     // Channel Status register address, +20H+(ch*2)
  WORD TxRxCount;    // Tx and Rx FIFO count register address, +10H+(ch*2)
  WORD IntID;        // Interrupt ID register address, +30H+(ch*2)

                 // Channel indirect register addresses base
  WORD ChOff;        // (ChanNum * 0x1000) Indirect Address

  BYTE R[RREGDATASIZE];  // holds indexed registers address and state

  BYTE TxControl[4];     // transmit control register vals, XFF0H
  BYTE RxControl[4];     // receive control register vals
  BYTE TxEnables[4];     // transmit processor enable vals
  BYTE TxCompare[4];     // transmit compare values 1 & 2
} CHANNEL_T;

typedef CHANNEL_T * CHANPTR_T;

/* Controller::BusType defines */
#define BUS_ISA 0
#define BUS_MC  1
#define BUS_PCI 2

/*-------------------------------------------------------------------------
Function: sClrBreak
Purpose:  Stop sending a transmit BREAK signal
Call:     sClrBreak(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sClrBreak(ChP) \
{ \
   (ChP)->TxControl[3] &= ~TX_SETBREAK; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sClrDTR
Purpose:  Clr the DTR output
Call:     sClrDTR(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sClrDTR(ChP) \
{ \
   (ChP)->TxControl[3] &= ~SET_DTR; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sClrRTS
Purpose:  Clr the RTS output
Call:     sClrRTS(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sClrRTS(ChP) \
{ \
   (ChP)->TxControl[3] &= ~SET_RTS; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sClrTxXOFF
Purpose:  Clear any existing transmit software flow control off condition
Call:     sClrTxXOFF(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sClrTxXOFF(ChP) \
{ \
   sOutB((ChP)->Cmd,TXOVERIDE | (BYTE)(ChP)->ChanNum); \
   sOutB((ChP)->Cmd,(BYTE)(ChP)->ChanNum); \
}

/*-------------------------------------------------------------------------
Function: sControllerEOI
Purpose:  Strobe the MUDBAC's End Of Interrupt bit.
Call:     sControllerEOI(CtlP)
          CONTROLLER_T *CtlP; Ptr to controller structure
Comments: For ISA & MCA cards only.
 instead.
|-------------------------------------------------------------------------*/
#define sControllerEOI(CTLP) sOutB((CTLP)->MReg2IO,(CTLP)->MReg2 | INT_STROB)

/*-------------------------------------------------------------------------
Function: sPCIControllerEOI
Purpose:  Strobe the boards End Of Interrupt bit.
Call:     sControllerEOI(CtlP)
          CONTROLLER_T *CtlP; Ptr to controller structure
Comments: For PCI cards only.
  WORD Access must be used, BYTE access has other data.
|-------------------------------------------------------------------------*/
#define sPCIControllerEOI(CTLP) sOutW((CTLP)->Reg3AIO, \
                                      (CTLP)->Reg3A | PCI_STROBE)

/*-------------------------------------------------------------------------
Function: sDisCTSFlowCtl
Purpose:  Disable output flow control using CTS
Call:     sDisCTSFlowCtl(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sDisCTSFlowCtl(ChP) \
{ \
   (ChP)->TxControl[2] &= ~CTSFC_EN; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sDisIXANY
Purpose:  Disable IXANY Software Flow Control
Call:     sDisIXANY(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sDisIXANY(ChP) \
{ \
   (ChP)->R[0x0e] = 0x86; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x0c]); \
}

/*-------------------------------------------------------------------------
Function: sDisLocalLoopback
Purpose:  Disable local loopback of transmit to receive
Call:     sDisLocalLoopback(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sDisLocalLoopback(ChP) \
{ \
   (ChP)->TxControl[3] &= ~LOCALLOOP; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: DisParity
Purpose:  Disable parity
Call:     sDisParity(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Comments: Function sSetParity() can be used in place of functions sEnParity(),
          sDisParity(), sSetOddParity(), and sSetEvenParity().
|-------------------------------------------------------------------------*/
#define sDisParity(ChP) \
{ \
   (ChP)->TxControl[2] &= ~PARITY_EN; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sDisRTSFlowCtl
Purpose:  Disable input flow control using RTS
Call:     sDisRTSFlowCtl(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sDisRTSFlowCtl(ChP) \
{ \
   (ChP)->RxControl[2] &= ~RTSFC_EN; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->RxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sDisRTSToggle
Purpose:  Disable RTS toggle
Call:     sDisRTSToggle(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sDisRTSToggle(ChP) \
{ \
   (ChP)->TxControl[2] &= ~RTSTOG_EN; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sDisRxCompare1
Purpose:  Disable Rx compare byte 1
Call:     sDisRxCompare1(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Comments: This function is used to disable Rx replace 1, Rx ignore 1,
          and Rx compare and interrupt 1.
|-------------------------------------------------------------------------*/
#define sDisRxCompare1(ChP) \
{ \
   (ChP)->R[0x1a] = 0x82; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x18]); \
   (ChP)->R[0x26] = 0x86; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x24]); \
   (ChP)->R[0x23] = 0x81; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x20]); \
}

/*-------------------------------------------------------------------------
Function: sDisRxCompare0
Purpose:  Disable Rx compare byte 0
Call:     sDisRxCompare0(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Comments: This function is used to disable Rx ignore 0,
|-------------------------------------------------------------------------*/
#define sDisRxCompare0(ChP) \
{ \
   (ChP)->R[0x16] = 0x82; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x14]); \
}

/*-------------------------------------------------------------------------
Function: sDisRxCompare2
Purpose:  Disable Rx compare byte 2
Call:     sDisRxCompare2(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Comments: This function is used to disable Rx replace 2, Rx ignore 2,
          and Rx compare and interrupt 2.
|-------------------------------------------------------------------------*/
#define sDisRxCompare2(ChP) \
{ \
   (ChP)->R[0x1e] = 0x8a; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x1c]); \
   (ChP)->R[0x2e] = 0x82; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x2c]); \
   (ChP)->R[0x2b] = 0x81; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x28]); \
}

/*-------------------------------------------------------------------------
Function: sDisRxFIFO
Purpose:  Disable Rx FIFO
Call:     sDisRxFIFO(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sDisRxFIFO(ChP) \
{ \
   (ChP)->R[0x32] = 0x0a; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x30]); \
}

/*-------------------------------------------------------------------------
Function: sDisRxStatusMode
Purpose:  Disable the Rx status mode
Call:     sDisRxStatusMode(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Comments: This takes the channel out of the receive status mode.  All
          subsequent reads of receive data using sReadRxWord() will return
          two data bytes.
|-------------------------------------------------------------------------*/
#define sDisRxStatusMode(ChP) sOutW((ChP)->ChanStat,0)

/*-------------------------------------------------------------------------
Function: sDisTransmit
Purpose:  Disable transmit
Call:     sDisTransmit(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
          This disables movement of Tx data from the Tx FIFO into the 1 byte
          Tx buffer.  Therefore there could be up to a 2 byte latency
          between the time sDisTransmit() is called and the transmit buffer
          and transmit shift register going completely empty.
|-------------------------------------------------------------------------*/
#define sDisTransmit(ChP) \
{ \
   (ChP)->TxControl[3] &= ~TX_ENABLE; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sDisTxCompare1
Purpose:  Disable Tx compare byte 1
Call:     sDisTxCompare1(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Comments: This function is used to disable Tx replace 1 with 1, Tx replace
          1 with 2, and Tx ignore 1.
|-------------------------------------------------------------------------*/
#define sDisTxCompare1(ChP) \
{ \
   (ChP)->TxEnables[2] &= ~COMP1_EN; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxEnables[0]); \
}

/*-------------------------------------------------------------------------
Function: sDisTxCompare2
Purpose:  Disable Tx compare byte 2
Call:     sDisTxCompare2(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Comments: This function is used to disable Tx replace 2 with 1 and Tx ignore 2.
|-------------------------------------------------------------------------*/
#define sDisTxCompare2(ChP) \
{ \
   (ChP)->TxEnables[2] &= ~COMP2_EN; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxEnables[0]); \
}

/*-------------------------------------------------------------------------
Function: sDisTxSoftFlowCtl
Purpose:  Disable Tx Software Flow Control
Call:     sDisTxSoftFlowCtl(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sDisTxSoftFlowCtl(ChP) \
{ \
   (ChP)->R[0x06] = 0x8a; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x04]); \
}

/*-------------------------------------------------------------------------
Function: sEnCTSFlowCtl
Purpose:  Enable output flow control using CTS
Call:     sEnCTSFlowCtl(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sEnCTSFlowCtl(ChP) \
{ \
   (ChP)->TxControl[2] |= CTSFC_EN; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sEnIXANY
Purpose:  Enable IXANY Software Flow Control
Call:     sEnIXANY(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sEnIXANY(ChP) \
{ \
   (ChP)->R[0x0e] = 0x21; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x0c]); \
}

/*-------------------------------------------------------------------------
Function: sEnLocalLoopback
Purpose:  Enable local loopback of transmit to receive
Call:     sEnLocalLoopback(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sEnLocalLoopback(ChP) \
{ \
   (ChP)->TxControl[3] |= LOCALLOOP; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: EnParity
Purpose:  Enable parity
Call:     sEnParity(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Comments: Function sSetParity() can be used in place of functions sEnParity(),
          sDisParity(), sSetOddParity(), and sSetEvenParity().

Warnings: Before enabling parity odd or even parity should be chosen using
          functions sSetOddParity() or sSetEvenParity().
|-------------------------------------------------------------------------*/
#define sEnParity(ChP) \
{ \
   (ChP)->TxControl[2] |= PARITY_EN; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sEnRTSFlowCtl
Purpose:  Enable input flow control using RTS
Call:     sEnRTSFlowCtl(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Warnings: RTS toggle and the RTS output will be both be cleared by this
          function.  The original state of RTS toggle and RTS output will
          not be preserved after a subsequent call to sDisRTSFlowCtl().
|-------------------------------------------------------------------------*/
#define sEnRTSFlowCtl(ChP) \
{ \
   (ChP)->TxControl[2] &= ~RTSTOG_EN; \
   (ChP)->TxControl[3] &= ~SET_RTS; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
   (ChP)->RxControl[2] |= RTSFC_EN; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->RxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sEnRTSToggle
Purpose:  Enable RTS toggle
Call:     sEnRTSToggle(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Comments: This function will disable RTS flow control and clear the RTS
          line to allow operation of RTS toggle.
|-------------------------------------------------------------------------*/
#define sEnRTSToggle(ChP) \
{ \
   (ChP)->RxControl[2] &= ~RTSFC_EN; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->RxControl[0]); \
   (ChP)->TxControl[2] |= RTSTOG_EN; \
   (ChP)->TxControl[3] &= ~SET_RTS; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sEnRxFIFO
Purpose:  Enable Rx FIFO
Call:     sEnRxFIFO(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sEnRxFIFO(ChP) \
{ \
   (ChP)->R[0x32] = 0x08; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x30]); \
}

/*-------------------------------------------------------------------------
Function: sEnRxIgnore0
Purpose:  Enable compare of Rx data with compare byte #1 and ignore (discard)
          that byte if a match is found.
Call:     sEnRxIgnore0(ChP,CmpByte)
          CHANNEL_T *ChP; Ptr to channel structure
          BYTE CmpByte; Byte to compare Rx data byte with
Comments: This function enables comparison of the receive data byte
          with CmpByte.  The comparison is done after the receive data
          byte has been masked (see sSetRxMask()).  If a match is found the
          receive data byte is ignored (discarded).

          Rx ignore 0 can be disabled with sDisRxCompare0().
|-------------------------------------------------------------------------*/
#define sEnRxIgnore0(ChP,CMPBYTE) \
{ \
   (ChP)->R[0x17] = (CMPBYTE); \
   (ChP)->R[0x16] = 0xce; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x14]); \
}

/*-------------------------------------------------------------------------
Function: sEnRxIgnore1
Purpose:  Enable compare of Rx data with compare byte #1 and ignore (discard)
          that byte if a match is found.
Call:     sEnRxIgnore1(ChP,CmpByte)
          CHANNEL_T *ChP; Ptr to channel structure
          BYTE CmpByte; Byte to compare Rx data byte with
Comments: This function enables comparison of the receive data byte
          with CmpByte.  The comparison is done after the receive data
          byte has been masked (see sSetRxMask()).  If a match is found the
          receive data byte is ignored (discarded).

          Rx ignore 1 can be disabled with sDisRxCompare1().
|-------------------------------------------------------------------------*/
#define sEnRxIgnore1(ChP,CMPBYTE) \
{ \
   (ChP)->R[0x1b] = (CMPBYTE); \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x18]); \
   (ChP)->R[0x1a] = 0xc4; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x18]); \
   (ChP)->R[0x23] = 0x81; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x20]); \
   (ChP)->R[0x26] = 0x0a; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x24]); \
}

/*-------------------------------------------------------------------------
Function: sEnRxIgnore2
Purpose:  Enable compare of Rx data with compare byte #2 and ignore (discard)
          that byte if a match is found.
Call:     sEnRxIgnore2(ChP,CmpByte)
          CHANNEL_T *ChP; Ptr to channel structure
          BYTE CmpByte; Byte to compare Rx data byte with
Comments: This function enables comparison of the receive data byte
          with CmpByte.  The comparison is done after the receive data
          byte has been masked (see sSetRxMask()).  If a match is found the
          receive data byte is ignored (discarded).

          Rx ignore 2 can be disabled with sDisRxCompare2().
|-------------------------------------------------------------------------*/
#define sEnRxIgnore2(ChP,CMPBYTE) \
{ \
   (ChP)->R[0x1f] = (CMPBYTE); \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x1c]); \
   (ChP)->R[0x1e] = 0xc6; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x1c]); \
   (ChP)->R[0x2b] = 0x81; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x28]); \
   (ChP)->R[0x2e] = 0x0a; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x2c]); \
}

/*-------------------------------------------------------------------------
Function: sEnRxIntCompare1
Purpose:  Enable compare of Rx data with compare byte #1 and interrupt
          if a match is found.
Call:     sEnRxIntCompare1(ChP,CmpByte)
          CHANNEL_T *ChP; Ptr to channel structure
          BYTE CmpByte; Byte to compare Rx data byte with
Comments: This function enables comparison of the receive data byte
          with CmpByte.  The comparison is done after the receive data
          byte has been masked (see sSetRxMask()).  If a match is found an
          interrupt is generated after adding the data byte to the receive
          FIFO.

          Rx compare interrupt 1 can be disabled with sDisRxCompare1().

Warnings: Before an interrupt will be generated SRC interrupts must be
          enabled (see sEnInterrupts()).
|-------------------------------------------------------------------------*/
#define sEnRxIntCompare1(ChP,CMPBYTE) \
{ \
   (ChP)->R[0x1b] = (CMPBYTE); \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x18]); \
   (ChP)->R[0x1a] = 0xc4; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x18]); \
   (ChP)->R[0x26] = 0x86; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x24]); \
   (ChP)->R[0x23] = 0x11; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x20]); \
}

/*-------------------------------------------------------------------------
Function: sEnRxIntCompare2
Purpose:  Enable compare of Rx data with compare byte #2 and interrupt
          if a match is found.
Call:     sEnRxIntCompare2(ChP,CmpByte)
          CHANNEL_T *ChP; Ptr to channel structure
          BYTE CmpByte; Byte to compare Rx data byte with
Comments: This function enables comparison of the receive data byte
          with CmpByte.  The comparison is done after the receive data
          byte has been masked (see sSetRxMask()).  If a match is found an
          interrupt is generated after adding the data byte to the receive
          FIFO.

          Rx compare interrupt 2 can be disabled with sDisRxCompare2().

Warnings: Before an interrupt will be generated SRC interrupts must be
          enabled (see sEnInterrupts()).
|-------------------------------------------------------------------------*/
#define sEnRxIntCompare2(ChP,CMPBYTE) \
{ \
   (ChP)->R[0x1f] = (CMPBYTE); \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x1c]); \
   (ChP)->R[0x1e] = 0xc6; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x1c]); \
   (ChP)->R[0x2e] = 0x82; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x2c]); \
   (ChP)->R[0x2b] = 0x10; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x28]); \
}

/*-------------------------------------------------------------------------
Function: sEnRxCmpIgn1
Purpose:  Enable compare of Rx data with compare byte #1 and interrupt
          and then discard that byte if a match is found.
Call:     sEnRxCmpIgn1(ChP,CmpByte)
          CHANNEL_T *ChP; Ptr to channel structure
          BYTE CmpByte; Byte to compare Rx data byte with
Comments: This function enables comparison of the receive data byte
          with CmpByte.  The comparison is done after the receive data
          byte has been masked (see sSetRxMask()).  If a match is found the
          receive data byte is ignored (discarded).

          Rx ignore 1 can be disabled with sDisRxCompare1().
|-------------------------------------------------------------------------*/
#define sEnRxCmpIgn1(ChP,CMPBYTE) \
{ \
   (ChP)->R[0x1b] = (CMPBYTE); \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x18]); \
   (ChP)->R[0x1a] = 0xc4; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x18]); \
   (ChP)->R[0x23] = 0x11; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x20]); \
   (ChP)->R[0x26] = 0x0a; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x24]); \
}

/*-------------------------------------------------------------------------
Function: sEnRxCmpIgn2
Purpose:  Enable compare of Rx data with compare byte #2 and interrupt
          and then discard that byte if a match is found.
Call:     sEnRxCmpIgn2(ChP,CmpByte)
          CHANNEL_T *ChP; Ptr to channel structure
          BYTE CmpByte; Byte to compare Rx data byte with
Comments: This function enables comparison of the receive data byte
          with CmpByte.  The comparison is done after the receive data
          byte has been masked (see sSetRxMask()).  If a match is found the
          receive data byte is ignored (discarded).

          Rx ignore 2 can be disabled with sDisRxCompare2().
|-------------------------------------------------------------------------*/
#define sEnRxCmpIgn2(ChP,CMPBYTE) \
{ \
   (ChP)->R[0x1f] = (CMPBYTE); \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x1c]); \
   (ChP)->R[0x1e] = 0xc6; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x1c]); \
   (ChP)->R[0x2b] = 0x10; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x28]); \
   (ChP)->R[0x2e] = 0x0a; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x2c]); \
}

/*-------------------------------------------------------------------------
Function: sEnRxProcessor
Purpose:  Enable the receive processor
Call:     sEnRxProcessor(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Comments: This function is used to start the receive processor.  When
          the channel is in the reset state the receive processor is not
          running.  This is done to prevent the receive processor from
          executing invalid microcode instructions prior to the
          downloading of the microcode.

Warnings: This function must be called after valid microcode has been
          downloaded to the AIOP, and it must not be called before the
          microcode has been downloaded.
|-------------------------------------------------------------------------*/
#define sEnRxProcessor(ChP) \
{ \
   (ChP)->RxControl[2] |= RXPROC_EN; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->RxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sEnRxReplace1
Purpose:  Enable compare of Rx data with compare byte #1 and replacement
          with a single byte if a match is found.
Call:     sEnRxReplace1(ChP,CmpByte,ReplByte)
          CHANNEL_T *ChP; Ptr to channel structure
          BYTE CmpByte; Byte to compare Rx data byte with
          BYTE ReplByte; Byte to replace Rx data byte with if
                                  a match is found on the compare.
Comments: This function enables comparison of the receive data byte
          with CmpByte.  The comparison is done after the receive data
          byte has been masked (see sSetRxMask()).  If a match is found
          the receive data byte is replaced by ReplByte.

          Rx replace 1 can be disabled with sDisRxCompare1().

|-------------------------------------------------------------------------*/
#define sEnRxReplace1(ChP,CMPBYTE,REPLBYTE) \
{ \
   (ChP)->R[0x1b] = (CMPBYTE); \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x18]); \
   (ChP)->R[0x27] = (REPLBYTE); \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x24]); \
   (ChP)->R[0x1a] = 0xc4; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x18]); \
   (ChP)->R[0x23] = 0x81; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x20]); \
   (ChP)->R[0x26] = 0x40; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x24]); \
}

/*-------------------------------------------------------------------------
Function: sEnRxReplace2
Purpose:  Enable compare of Rx data with compare byte #2 and replacement
          with a single byte if a match is found.
Call:     sEnRxReplace2(ChP,CmpByte,ReplByte)
          CHANNEL_T *ChP; Ptr to channel structure
          BYTE CmpByte; Byte to compare Rx data byte with
          BYTE ReplByte; Byte to replace Rx data byte with if
                                  a match is found on the compare.
Comments: This function enables comparison of the receive data byte
          with CmpByte.  The comparison is done after the receive data
          byte has been masked (see sSetRxMask()).  If a match is found
          the receive data byte is replaced by ReplByte.

          Rx replace 2 can be disabled with sDisRxCompare2().

|-------------------------------------------------------------------------*/
#define sEnRxReplace2(ChP,CMPBYTE,REPLBYTE) \
{ \
   (ChP)->R[0x1f] = (CMPBYTE); \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x1c]); \
   (ChP)->R[0x2f] = (REPLBYTE); \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x2c]); \
   (ChP)->R[0x1e] = 0xc6; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x1c]); \
   (ChP)->R[0x2b] = 0x81; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x28]); \
   (ChP)->R[0x2e] = 0x40; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x2c]); \
}

/*-------------------------------------------------------------------------
Function: sEnRxStatusMode
Purpose:  Enable the Rx status mode
Call:     sEnRxStatusMode(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Comments: This places the channel in the receive status mode.  All subsequent
          reads of receive data using sReadRxWord() will return a data byte
          in the low word and a status byte in the high word.

|-------------------------------------------------------------------------*/
#define sEnRxStatusMode(ChP) sOutW((ChP)->ChanStat,STATMODE)

/*-------------------------------------------------------------------------
Function: sEnTransmit
Purpose:  Enable transmit
Call:     sEnTransmit(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sEnTransmit(ChP) \
{ \
   (ChP)->TxControl[3] |= TX_ENABLE; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sEnTxIgnore1
Purpose:  Enable compare of Tx data with compare byte #1 and ignore (do not
          transmit) that byte if a match is found.
Call:     sEnTxIgnore1(ChP,CmpByte)
          CHANNEL_T *ChP; Ptr to channel structure
          BYTE CmpByte; Byte to compare Tx data byte with
Comments: This function enables comparison of the transmit data byte
          with CmpByte.  If a match is found the transmit data byte
          is ignored, that is, it is not transmitted.

          Tx ignore 1 can be disabled with sDisTxCompare1().

|-------------------------------------------------------------------------*/
#define sEnTxIgnore1(ChP,CMPBYTE) \
{ \
   (ChP)->TxCompare[2] = (CMPBYTE); \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxCompare[0]); \
   (ChP)->TxEnables[2] &= ~REP1W2_EN; \
   (ChP)->TxEnables[2] |= (COMP1_EN | IGN1_EN); \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxEnables[0]); \
}

/*-------------------------------------------------------------------------
Function: sEnTxIgnore2
Purpose:  Enable compare of Tx data with compare byte #2 and ignore (do not
          transmit) that byte if a match is found.
Call:     sEnTxIgnore2(ChP,CmpByte)
          CHANNEL_T *ChP; Ptr to channel structure
          BYTE CmpByte; Byte to compare Tx data byte with
Comments: This function enables comparison of the transmit data byte
          with CmpByte.  If a match is found the transmit data byte
          is ignored, that is, it is not transmitted.

          Tx ignore 2 can be disabled with sDisTxCompare2().

|-------------------------------------------------------------------------*/
#define sEnTxIgnore2(ChP,CMPBYTE) \
{ \
   (ChP)->TxCompare[3] = (CMPBYTE); \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxCompare[0]); \
   (ChP)->TxEnables[2] |= (COMP2_EN | IGN2_EN); \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxEnables[0]); \
}

/*-------------------------------------------------------------------------
Function: sEnTxSoftFlowCtl
Purpose:  Enable Tx Software Flow Control
Call:     sEnTxSoftFlowCtl(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sEnTxSoftFlowCtl(ChP) \
{ \
   (ChP)->R[0x06] = 0xc5; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x04]); \
}

/*-------------------------------------------------------------------------
Function: sGetAiopIntStatus
Purpose:  Get the AIOP interrupt status
Call:     sGetAiopIntStatus(CtlP,AiopNum)
          CONTROLLER_T *CtlP; Ptr to controller structure
          int AiopNum; AIOP number
Return:   BYTE: The AIOP interrupt status.  Bits 0 through 7
                         represent channels 0 through 7 respectively.  If a
                         bit is set that channel is interrupting.
|-------------------------------------------------------------------------*/
#define sGetAiopIntStatus(CTLP,AIOPNUM) sInB((CTLP)->AiopIntChanIO[AIOPNUM])

/*-------------------------------------------------------------------------
Function: sGetChanIntID
Purpose:  Get a channel's interrupt identification byte
Call:     sGetChanIntID(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Return:   BYTE: The channel interrupt ID.  Can be any
             combination of the following flags:
                RXF_TRIG:     Rx FIFO trigger level interrupt
                TXFIFO_MT:    Tx FIFO empty interrupt
                SRC_INT:      Special receive condition interrupt
                DELTA_CD:     CD change interrupt
                DELTA_CTS:    CTS change interrupt
                DELTA_DSR:    DSR change interrupt
|-------------------------------------------------------------------------*/
#define sGetChanIntID(ChP) (sInB((ChP)->IntID) & (RXF_TRIG | TXFIFO_MT | SRC_INT | DELTA_CD | DELTA_CTS | DELTA_DSR))

/*-------------------------------------------------------------------------
Function: sGetRPlusRingIntID
Purpose:  Get the RPlus delta change status.
Call:     sGetRPlusRingIntID(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Return:   BYTE: The channel interrupt ID.  Can be any
             combination of the following flags:
                DELTA_DSR:    Ring-Indicator change interrupt
|-------------------------------------------------------------------------*/
#define sGetRPlusRingIntID(ChP) (sInB((ChP)->IntID+4) & (DELTA_DSR))

/*-------------------------------------------------------------------------
Function: sGetControllerIntStatus
Purpose:  Get the controller interrupt status
Call:     sGetControllerIntStatus(CtlP)
          CONTROLLER_T *CtlP; Ptr to controller structure
Return:   BYTE: The controller interrupt status in the lower 4
                         bits.  Bits 0 through 3 represent AIOP's 0
                         through 3 respectively.  If a bit is set that
                         AIOP is interrupting.  Bits 4 through 7 will
                         always be cleared.
Comment: This is only valid for ISA/MC bus cards.
|-------------------------------------------------------------------------*/
#define sGetControllerIntStatus(CTLP) (sInB((CTLP)->MReg1IO) & 0x0f)

/*-------------------------------------------------------------------------
Function: sPCIGetControllerIntStatus
Purpose:  Get the controller interrupt status
Call:     sPCIGetControllerIntStatus(CtlP)
          CONTROLLER_T *CtlP; Ptr to controller structure
Return:   BYTE: The controller interrupt status in the lower 4
                         bits.  Bits 0 through 3 represent AIOP's 0
                         through 3 respectively.  If a bit is set that
                         AIOP is interrupting.  Bits 4 through 7 will
                         always be cleared.
Comment: This is only valid for PCI bus cards.
  WORD Access must be used, BYTE access has other data.
|-------------------------------------------------------------------------*/
#define sPCIGetControllerIntStatus(CTLP) \
       ((sInW((CTLP)->Reg3AIO) >> 8) & 0xf)

/*-------------------------------------------------------------------------
Function: sGetModemStatus
Purpose:  Get the modem status
Call:     sGetModemStatus(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Return:   BYTE: Modem status using flags CD_ACT, DSR_ACT, and CTS_ACT.
|-------------------------------------------------------------------------*/
#define sGetModemStatus(ChP) (BYTE)(sInB((WORD)(ChP)->ChanStat) & (CD_ACT | DSR_ACT | CTS_ACT))

/*-------------------------------------------------------------------
Function: sGetRPlusModemRI
Purpose:  Get the modem status(DSR on upper unused ports have RI signal)
          CHANPTR_T ChP; Ptr to channel structure
Return:   unsigned char: Modem status using flags DSR_ACT signifying
          RI signal.
---------------------------------------------------------------------*/
#define sGetRPlusModemRI(ChP) (BYTE)(sInB((WORD)(ChP)->ChanStat+8) & (DSR_ACT))

/*-------------------------------------------------------------------------
Function: sGetRxCnt
Purpose:  Get the number of data bytes in the Rx FIFO
Call:     sGetRxCnt(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Return:   int: The number of data bytes in the Rx FIFO.
Comments: Byte read of count register is required to obtain Rx count.
|-------------------------------------------------------------------------*/
#define sGetRxCnt(ChP) sInW((ChP)->TxRxCount)

/*-------------------------------------------------------------------------
Function: sGetTxCnt
Purpose:  Get the number of data bytes in the Tx FIFO
Call:     sGetTxCnt(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Return:   BYTE: The number of data bytes in the Tx FIFO.
Comments: Byte read of count register is required to obtain Tx count.
|-------------------------------------------------------------------------*/
#define sGetTxCnt(ChP) sInB((WORD)(ChP)->TxRxCount)

/*-------------------------------------------------------------------------
Function: sResetAiopByNum
Purpose:  Reset the AIOP by number
Call:     sResetAiopByNum(CTLP,AIOPNUM)
  CONTROLLER_T CTLP; Ptr to controller structure
  AIOPNUM; AIOP index 
|-------------------------------------------------------------------------*/
#define sResetAiopByNum(CTLP,AIOPNUM) \
{ \
   sOutB((CTLP)->AiopIO[(AIOPNUM)]+_CMD_REG,RESET_ALL); \
   sOutB((CTLP)->AiopIO[(AIOPNUM)]+_CMD_REG,0x0); \
}

/*-------------------------------------------------------------------------
Function: sSendBreak
Purpose:  Send a transmit BREAK signal
Call:     sSendBreak(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sSendBreak(ChP) \
{ \
   (ChP)->TxControl[3] |= TX_SETBREAK; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sSetData7
Purpose:  Set data bits to 7
Call:     sSetData7(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sSetData7(ChP) \
{ \
   (ChP)->TxControl[2] &= ~DATA8BIT; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sSetData8
Purpose:  Set data bits to 8
Call:     sSetData8(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sSetData8(ChP) \
{ \
   (ChP)->TxControl[2] |= DATA8BIT; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sSetDTR
Purpose:  Set the DTR output
Call:     sSetDTR(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sSetDTR(ChP) \
{ \
   (ChP)->TxControl[3] |= SET_DTR; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sSetEvenParity
Purpose:  Set even parity
Call:     sSetEvenParity(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Comments: Function sSetParity() can be used in place of functions sEnParity(),
          sDisParity(), sSetOddParity(), and sSetEvenParity().

Warnings: This function has no effect unless parity is enabled with function
          sEnParity().
|-------------------------------------------------------------------------*/
#define sSetEvenParity(ChP) \
{ \
   (ChP)->TxControl[2] |= EVEN_PAR; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sSetOddParity
Purpose:  Set odd parity
Call:     sSetOddParity(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Comments: Function sSetParity() can be used in place of functions sEnParity(),
          sDisParity(), sSetOddParity(), and sSetEvenParity().

Warnings: This function has no effect unless parity is enabled with function
          sEnParity().
|-------------------------------------------------------------------------*/
#define sSetOddParity(ChP) \
{ \
   (ChP)->TxControl[2] &= ~EVEN_PAR; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sSetRTS
Purpose:  Set the RTS output
Call:     sSetRTS(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sSetRTS(ChP) \
{ \
   (ChP)->TxControl[3] |= SET_RTS; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sSetRxMask
Purpose:  Set the Rx mask value
Call:     sSetRxMask(ChP,Ch)
          CHANNEL_T *ChP; Ptr to channel structure
          BYTE Ch; The value to set the Rx mask to
|-------------------------------------------------------------------------*/
#define sSetRxMask(ChP,CH) \
{ \
   (ChP)->R[0x13] = (CH); \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x10]); \
}

/*-------------------------------------------------------------------------
Function: sSetRxTrigger
Purpose:  Set the Rx FIFO trigger level
Call:     sSetRxProcessor(ChP,Level)
          CHANNEL_T *ChP; Ptr to channel structure
          BYTE Level; Number of characters in Rx FIFO at which the
             interrupt will be generated.  Can be any of the following flags:

             TRIG_NO:   no trigger
             TRIG_1:    1 character in FIFO
             TRIG_1_2:  FIFO 1/2 full
             TRIG_7_8:  FIFO 7/8 full
Comments: An interrupt will be generated when the trigger level is reached
          only if function sEnInterrupt() has been called with flag
          RXINT_EN set.  The RXF_TRIG flag in the Interrupt Idenfification
          register will be set whenever the trigger level is reached
          regardless of the setting of RXINT_EN.
|-------------------------------------------------------------------------*/
#define sSetRxTrigger(ChP,LEVEL) \
{ \
   (ChP)->RxControl[2] &= ~TRIG_MASK; \
   (ChP)->RxControl[2] |= LEVEL; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->RxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sSetStop1
Purpose:  Set stop bits to 1
Call:     sSetStop1(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sSetStop1(ChP) \
{ \
   (ChP)->TxControl[2] &= ~STOP2; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sSetStop2
Purpose:  Set stop bits to 2
Call:     sSetStop2(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
|-------------------------------------------------------------------------*/
#define sSetStop2(ChP) \
{ \
   (ChP)->TxControl[2] |= STOP2; \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->TxControl[0]); \
}

/*-------------------------------------------------------------------------
Function: sSetTxXOFFChar
Purpose:  Set the Tx XOFF flow control character
Call:     sSetTxXOFFChar(ChP,Ch)
          CHANNEL_T *ChP; Ptr to channel structure
          BYTE Ch; The value to set the Tx XOFF character to
|-------------------------------------------------------------------------*/
#define sSetTxXOFFChar(ChP,CH) \
{ \
   (ChP)->R[0x07] = (CH); \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x04]); \
}

/*-------------------------------------------------------------------------
Function: sSetTxXONChar
Purpose:  Set the Tx XON flow control character
Call:     sSetTxXONChar(ChP,Ch)
          CHANNEL_T *ChP; Ptr to channel structure
          BYTE Ch; The value to set the Tx XON character to
|-------------------------------------------------------------------------*/
#define sSetTxXONChar(ChP,CH) \
{ \
   (ChP)->R[0x0b] = (CH); \
   sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0x08]); \
}

/*-------------------------------------------------------------------------
Function: sStartRxProcessor
Purpose:  Start a channel's receive processor
Call:     sStartRxProcessor(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Comments: This function is used to start a Rx processor after it was
          stopped with sStopRxProcessor() or sStopSWInFlowCtl().  It
          will restart both the Rx processor and software input flow control.
|-------------------------------------------------------------------------*/
#define sStartRxProcessor(ChP) sOutDW((ChP)->IndexAddr,*(DWORD *)&(ChP)->R[0])

// unused macros
// #define sPCIGetControllerIntStatus(CTLP)
// #define sGetModemStatus(ChP) (BYTE)(sInB((WORD)(ChP)->ChanStat) & (CD_ACT | DSR_ACT | CTS_ACT))
// #define sGetRxCnt(ChP) sInW((ChP)->TxRxCount)
// #define sGetRxStatus(ChP) (sInB((WORD)(ChP)->ChanStat) & RDA)
// #define sGetTxRxDataIO(ChP) (ChP)->TxRxData

/*-------------*/
int sInitIRQ(CONTROLLER_T *CtlP,
             int IRQNum,
             BYTE Frequency,
             int PeriodicOnly);

int sSetupPorts(CONTROLLER_T *CtlP,
                WORD aiop_base_io,
                int offset_port,
                int num_ports,
                int bus_type,
                int card_type);

int sInitCtlStruct(CONTROLLER_T *Ctlp,
             int BusType,
             int CardNum,
             int CardType);

int sDisAiop(CONTROLLER_T *b, int aiop_num);
int sEnAiop(CONTROLLER_T *b, int aiop_num);
int sReadAiopID(WORD io);
int sReadAiopNumChan(WORD io);
int sInitChan(CONTROLLER_T *CtlP,
              CHANNEL_T *ChP,
              int AiopNum,
              int ChanNum);
BYTE sGetRxErrStatus(CHANNEL_T *ChP);
void sSetParity(CHANNEL_T *ChP,int Parity);
void sStopRxProcessor(CHANNEL_T *ChP);
void sStopSWInFlowCtl(CHANNEL_T *ChP);
void sFlushRxFIFO(CHANNEL_T *ChP);
void sFlushTxFIFO(CHANNEL_T *ChP);
int sFlushTxPriorityBuf(CHANNEL_T *ChP, BYTE *Data);
BYTE sGetTxPriorityCnt(CHANNEL_T *ChP);
int sReadRxBlk(CHANNEL_T *ChP,BYTE *Buffer,int Count);
int sWriteTxBlk(CHANNEL_T *ChP,BYTE *Buffer,int Count);
// int sWriteTxPrioBlk(CHANNEL_T *ChP,BYTE *Buffer,int Count);
int sWriteTxPrioByte(CHANNEL_T *ChP, BYTE Data);
void sEnInterrupts(CHANNEL_T *ChP,WORD Flags);
void sDisInterrupts(CHANNEL_T *ChP,WORD Flags);
int sSetBaudRate(CHANNEL_T *ChP, DWORD desired_baud, BOOL SetHardware);
void sEnGlobalInt(CONTROLLER_T *CtlP);
void sDisGlobalInt(CONTROLLER_T *CtlP);
void sChanOutWI(CHANNEL_T *ChP, WORD RegNum,  WORD val);
void sChanOutBI(CHANNEL_T *ChP, WORD RegNum,  BYTE val);
BYTE sChanInBI(CHANNEL_T *ChP, WORD RegNum);
WORD sChanInWI(CHANNEL_T *ChP, WORD RegNum);
void sChanCommand(CHANNEL_T *ChP, BYTE command);

/*-------------*/

#define CTL_SIZE 4
extern CONTROLLER_T sController[CTL_SIZE];

#ifdef __cplusplus
}
#endif  /* __cplusplus */
