/*--------------------------------------------------------------------------
 rcktfoss.c - FOSSIL driver for Comtrol RocketPort board
11-5-2004 V1.73 release    RLH - Comtrol Europe Ltd
1. Fix problem that would allow a command line interrupt override
   (eg -ui<n> where n > 0 ) of the software timer interrupts.
   This problem manifested itself as an "Interrupts not working" 
   error message

7-19-2004 V1.72 release    RLH - Comtrol Europe Ltd
1. Add device id support for RocketPort uPCi cards (NOT RocketModem III)
2. Force option -ui0 to s/w timer base interrupts if these cards
   are detected.

9-26-97 V1.71 release.
9-26-97 V1.7.02B correct major bugs in rfmulti.com which caused load
  parameters to be read wrong(-p, -k).  Also correct minor error
  in -uk unload option processing which would cause uneeded
  "Unload with -uk" messages.
4-21-97 V1.7.01B add Rplus support, take out os/2 compatibility switch.
2-28-97 V1.7 recompile as V1.7 for release.
1-30-97 V1.60.003 add ability to unload.
12-20-96 V1.60.002 add BIOS05 function if BIOS mode specified.
8-22-96 V1.60.001 add rs485 control.
6-06-96 V1.60 release, update version.
4-24-96 V1.50.010 BIOS mode(-j) not running correctly due to chgs in .008v
4-20-96 V1.50.009 rfmulti reading addresses in dec not hex(fix) - kpb
4-20-96 V1.50.008 create a version which will do multiple boards.

3-12-96 V1.50.007 add events_option to test 230kbaud rate framing errors.
  
3-07-96  add -g option to allow running with OS/2 driver.

1-29-96 V1.50.003 added -ta# option to allow setting timeout period
  of BIOS calls with a possible wait condition.

1-18-96 V1.50.002
  Problem: When windows driver loaded, it would disable the
  receive hardware, if RCKTFOSS.EXE loaded before windows,
  you would have to run a port from Windows first before receive
  would work.  FIX: Re-enable some additional hardware on
  configuration calls(recieve hardware).
  Change ASM signature so it does not confuse itself with RF.EXE.

1-10-96 V1.50.001, PCI baudrate prescaler not setup, fix.  Also
  change how version/id string work - kpb.

11-12-95 Add extended INT14 functions(-j).
11-05-95 Allow peaceful coexistance with other RocketPort Drivers
10-22-95 Implement non-comm functions.
  Allow peaceful coexistance with previously loaded fossil driver(for com1-4).
  Allow IRQ=0 to be a valid selection.  - kpb.

V1.1 to Present by Karl Bongers
V1.0 by C. Elmquist

 Copyright 1996 Comtrol Corporation.  All rights reserved.  Proprietary
 information not permitted for development or use with non-Comtrol products.
|---------------------------------------------------------------------------*/
#define UNI

#ifdef SOFT_BUFS
  char *driver_str = {"Comtrol Corp., RocketPort FOSSIL V1.73, Buffered"};
#endif
#ifdef ALLOW_FOUR_BOARDS
  char *driver_str = {"Comtrol Corp., RocketPort FOSSIL V1.73, Multicard"};
#endif
#ifdef SMALL
  char *driver_str = {"Comtrol Corp., RocketPort FOSSIL V1.73, Small"};
#endif

//#include <stdio.h>

#include <dos.h>


//#ifdef BC_16
//#include <mem.h>
//#endif

#define TRUE 1
#define FALSE 0
#define NULL 0

#define DELTA_PROCESSING 1
// #define TEST_FREQ 1
// #define WIERD_BIOS_MODE 1
// #define EVENT_PROCESS 1

//#define DTRACE(s) puts(s)
#define DTRACE(s)

#include "rcktldef.h"
#include "rckt.h"
#include "rcktfoss.h"
#include "pcibios.h"

#define TICK_TIMEOUT tick_timeout

int tick_timeout = (18*3);

int debug_level = 0;

#define QOUT Port->QOutBase

#ifdef ALLOW_FOUR_BOARDS
#define MAX_NUM_PORTS (MAX_PORTS_PER_BOARD*4)
#else
#define MAX_NUM_PORTS (MAX_PORTS_PER_BOARD)
#endif
RCKTPort_t RcktPort[MAX_NUM_PORTS];

typedef struct {    // a struct to hold IRQ config
  BYTE Mask;    // information
  BYTE IntNum;
} IRQData_t;

static IRQData_t IRQData[16]= {
  {0,0},    /* IRQ0 */
  {0,0},    /* IRQ1 */
  {0,0},    /* IRQ2 */
  {0xf7,0xb},  /* IRQ3 */
  {0xef,0xc},  /* IRQ4 */
  {0xdf,0xd},  /* IRQ5 */
  {0,0},    /* IRQ6 */
  {0,0},    /* IRQ7 */
  {0,0},    /* IRQ8 */
  {0xfd,0x71},  /* IRQ9 */
  {0xfb,0x72},  /* IRQ10 */
  {0xf7,0x73},  /* IRQ11 */
  {0xef,0x74},  /* IRQ12 */
  {0,0},     /* IRQ13 */
  {0,0},    /* IRQ14 */
  {0x7f,0x77}  /* IRQ15 */
};

CONTROLLER_T  *ctlp;

DWORD OldIsr14;  // old interrupt vector
DWORD OldIsr8;   // old interrupt vector
DWORD OldIsr;    // old interrupt vector
DWORD AppHook = 0L;   // application irq hook

void RCKTInt(void);

int shared = 1;  // sharing of hardware? Yes then, don't mess with whats not yours.
				 // this is for running more than one driver.

WORD Boardio = 0x180;  // base address of board
int irq = 0;        // irq number used(0=none)
int comFirst  = 0;  // first port used(COM1)
int comLast   = 0;  // used in assembly to check range
int comOffset = 0;  // offset
int nPorts = 0;     // num ports used
int use_timer = 1;
int num_boards = 0;
volatile int poll_ticks = 0;
int wait_option = 0;
int events_option = 0;
int default_flow_ctrl=1;  // rts & dtr autoflow default
int test_flags = 0;

long locked_baud = 0;
unsigned int loaded_stat;  // tells if other fossil driver present and used.
int only_one_instance = 1;  // don't allow more than one of our drivers
int pass_thru_flag = 0;  // pass all non-RocketPort calls down to previously
						 // loaded fossil
int bus_style = BUS_ISA;  // assume isa-bus card

int driver_mode = 0;  // fossil=0, int14=1

#ifdef SOFT_BUFS
void PutTxChars(RCKTPort_t *Port);
#define SB_TOTAL_SIZE 0x7000
  WORD QOutSize;
unsigned char soft_buf[SB_TOTAL_SIZE];
#endif

int do_cmd_line(char *line, int loaded);
void puts(char *str);
void memset(void *ptr, unsigned char v, int size);
void memcpy(void far *dest, void far *src, unsigned int size);
int strlen(char *str);

int deinstall(int free_flags);
static int hook_irq(int irq);
static void debv_outs(char *str, int level);
static void intoa(char *str, WORD v, int places);
static void hextoa(char *str, WORD v, int places);

#ifdef DO_DOS_NAMES
struct dos_name_dev dos_dev[MAX_PORTS_PER_BOARD];
static int hook_name(void);
#endif
struct dos_req_header far *glob_req;
void strat_c(struct dos_req_header far *req);
void intr_c(void);
DWORD get_vect(int vector);
void set_vect(int vector, DWORD proc);
int toupper(int c);

//--------------- following added for PCI support
typedef struct  DevInfo {
	WORD   DevId ;          /* DeviceId */
	WORD   Base_Addr;       /* base io address for the board */
	BYTE   Irq;             /* int line assigned to board */
	BYTE   Bus_num;         /* pci bus num where board is located */
	BYTE   Dev_num;         /* boards device number on pci bus */
	BYTE   Fctn_num;        /* boards function number */
	WORD   DeviceId;        
} DEVINFO;

char cmd_line[160];
int send_cmd_line(char *line);

#define MAX_BRDS  4
DEVINFO pci_config[MAX_BRDS];
int num_pci_s = 0;

int GetDevInfo(DEVINFO *pci_config);
int FindDevs(int vendor_id, int max_device_id);
#define MAX_DEV_ID    0x0805    //0x10 // our 4-port=1, 8-port=2, 16=3, 32-4.
#define COMTROL_ID    0x11FE    // Comtrols ROCKF.VXD assigned vendor id

#ifdef ALLOW_FOUR_BOARDS
WORD isa_addr[4] = {0,0,0,0};
BYTE isa_sizes[4] = {32,32,32,32};
#endif

//                                   012345678901234567890123456789
static char *chan_init_error_str = {"brd# port# failed."};


/*----------------------------------------------------------------------------
| main -
|----------------------------------------------------------------------------*/
int main(char *argv)
{
  WORD mbio;
  int aiop, ch, pn, ctlid;
  RCKTPort_t *Port;
  CHANNEL_T *cp;
  int i,j,k,stat, port_num, ports_left, num_ports;
  char *buf;
  WORD pci_io;

  puts(driver_str);
  puts("Copyright 1995-2004 all rights reserved.");

  memset(&sController[0], 0, sizeof(sController));   // our boards
  memset(RcktPort, 0, sizeof(RcktPort));  // all of our Port structs

  loaded_stat = fossil_loaded();  // check if existing fossil already loaded

  j= *argv++;   // size of tail
  for (i=0; i<j; i++)
   cmd_line[i] = *argv++;

  cmd_line[i] = 0;

  switch (loaded_stat)
  {
	case 0:  // non-loaded
	  puts("loading..");
	break;
	case 1:  // other guys loaded, pass comm-port request to original
	  puts("Existing Fossil used.");
	break;
	case 2:  // ours is loaded, so don't load again
	  if (only_one_instance)
	  {
		send_cmd_line(cmd_line);
		puts("OK.");
		return 1;
	  }
	  puts("ld.");
	  loaded_stat = 1;
	break;
  }

  stat = do_cmd_line(cmd_line, 0);
  if (stat)
	return 1;


#ifdef ALLOW_FOUR_BOARDS
  ports_left = nPorts;  // count this down as we initialize ports
#endif

  if (bus_style == BUS_ISA)
  {
	if (Boardio < 0x100 || Boardio > 0x3ff)
	{
	  puts("Bad -p#.");
	  return 1;
	}
  }
  else if (bus_style == BUS_PCI)
  {
	 stat = pci_init();
	 if (stat == 0)  // ok
	 {      
		 stat = FindDevs(COMTROL_ID, MAX_DEV_ID);
	 }

	 if (num_pci_s == 0)
	   stat = 1;  // err

	 if (stat == 0)  // ok
	 {
#ifdef ALLOW_FOUR_BOARDS
	   for (i=0; i<num_pci_s; i++)
	   {
		 GetDevInfo(&pci_config[i]);  // get irq & addr for 1st board
		 pci_io = pci_config[i].Base_Addr;

		#ifdef TRACEON
		 printf("Base=%x, IRQ=%x",pci_io, pci_config[i].Irq);
		#endif
		 
		 if ((pci_config[i].Irq > 0) && (pci_config[i].Irq < 16))
		 {
		   if (irq > 0)  // make them select interrupts
			 irq = pci_config[i].Irq;
		 }

		 /*if uPCi card then must force irq = 0 to use s/w timer interrupt*/
		 if( pci_config[i].DevId > 0x0800 && pci_config[i].DevId < 0x806 )
		 {
			irq = 0;
		 }

		 sInitCtlStruct(&sController[num_boards],
			 BUS_PCI,       // bustype
			 i,             // cardnum
			 pci_config[i].DeviceId); // pci-id for PCI type card

		 num_ports = sController[num_boards].NumPorts;
		 ports_left -= num_ports;
		 sSetupPorts(&sController[num_boards],
					 pci_config[i].Base_Addr,
					 0,  // first port on board
					 num_ports,  // number of ports
					 BUS_PCI,
					 pci_config[i].DeviceId);  // card type(handles rplus)
		 num_boards++;
	   }
#else
	   GetDevInfo(&pci_config[0]);  // get irq & addr for 1st board
	   ctlp= &sController[0];

	   pci_io = pci_config[0].Base_Addr;

#ifdef TRACE_ON
	   printf("Base=%x, IRQ=%x",pci_io, pci_config[0].Irq);
#endif
	   
	   sInitCtlStruct(&sController[0],
			 BUS_PCI,       // bustype
			 0,             // cardnum
			 pci_config[0].DeviceId); // pci-id for PCI type card

	   //if (ctlp->PortsPerAiop == 4)
	   //  puts("ok1");
	   if ((pci_config[0].Irq > 0) && (pci_config[0].Irq < 16))
	   {
		 if (irq > 0)  // make them select interrupts
		   irq = pci_config[0].Irq;
	   }

	   /*
		 if uPCi card then must force irq = 0 to use s/w timer interrupt
	   */
		if( pci_config[0].DevId > 0x0800 && pci_config[0].DevId < 0x806 )
		{
		   irq = 0;
		}


	   nPorts = sController[num_boards].NumPorts;
	   sSetupPorts(ctlp, pci_io, comOffset, nPorts, bus_style,
				   pci_config[0].DeviceId);

	   //if ((ctlp->clk_prescaler & 0xf) == 2)
	   //  puts("ok2");
#endif
	 }
	 else
	 {
	   puts("No PCI Board.");
	   return 1;
	 }
  }  // end PCI case

  if ((irq != 0) && (IRQData[irq].IntNum == 0))
  {
	puts("Bad -ui#.");
	return 1;
  }

  if (pass_thru_flag)
  {
	if (loaded_stat != 0)  // existing fossil present
	{
	  // puts("Non-RocketPort calls passed to existing fossil");
	  loaded_stat |= 4;  // signal int. routine to skip
	}
	//else puts("No-other fossil loaded to pass to, -A switch ignored");
  }

  //----------- Init the board and determine how many ports it has
  ctlp= &sController[0];
  if (nPorts <= 0) nPorts = 1;
  if (comFirst < 0)
  {
	puts("No First.");
	return 1;
  }
  comLast = comFirst + (nPorts-1);  // used by asm end to check range

  //------ setup ISA board hardware and board structure
#ifdef ALLOW_FOUR_BOARDS
  j = 0;  // isa-board num
  while ((ports_left > 0) && (num_boards < 4))
  {
	sInitCtlStruct(&sController[num_boards],
			 BUS_ISA,       // bustype
			 num_boards,    // cardnum
			 0); // pci-id for PCI type card
	if (ports_left >= isa_sizes[j]) k = isa_sizes[j];
	else k = ports_left;
	sController[num_boards].NumPorts = k;  // number of ports on ctlr board
	if (isa_addr[j] == 0)
	  isa_addr[j] = (j*0x1000) + Boardio;

	//if (isa_addr[j] == 0x180)
	//  puts("ok2b");
	//if (k == 8)
	//  puts("ok2c");

	sSetupPorts(&sController[num_boards],
				isa_addr[j],  // io-base for AIOPIC chips
				comOffset,
				k,  // num ports
				BUS_ISA,
				j);  // pass in ISA board index
	//printf("sSetup%d, off:%d, nump:%d addr:%x\n",
	//        num_boards+1,comOffset, k, isa_addr[j]);

	ports_left -= isa_sizes[j];
	++num_boards;
	++j;
  }
#else

  if (bus_style == BUS_ISA)
  {
	sInitCtlStruct(&sController[0],
			  BUS_ISA, // bustype
			  0,       // cardnum
			  0);      // pci-id for PCI type card
	sController[0].NumPorts = nPorts;  // number of ports on ctlr board
	sSetupPorts(ctlp, Boardio, comOffset, nPorts, bus_style, 0);
  }

  num_boards = 1;
#endif

  //------ setup each port hardware and board structure
#ifdef ALLOW_FOUR_BOARDS

#ifdef COMMENT_OUT
  if (num_boards == 2)
	puts("ok3");
  if (sController[0].NumPorts == 8)
	puts("ok4");
  if (sController[0].PortsPerAiop == 4)
	puts("ok5");
  if (sController[1].PortsPerAiop == 8)
	puts("ok6");
  if ((sController[0].clk_prescaler & 0xf) == 2)
	puts("ok7");
  if ((sController[1].clk_prescaler & 0xf) == 4)
	puts("ok8");
#endif

  port_num = 0;
  for (i=0; i<num_boards; i++)
  {
	ctlp = &sController[i];
	for (j=0; j< ctlp->NumPorts; j++)
	{
	  Port= &RcktPort[port_num];
	  cp= &Port->ch;
	  
	  sEnAiop(ctlp, j/ctlp->PortsPerAiop);  // enable Aiopic Access
	  if (sInitChan(ctlp, cp, j/ctlp->PortsPerAiop, j%ctlp->PortsPerAiop) != 0)
	  {
		//printf("ctrl%d, port:%d, pb:%d addr:%x\n",
		//       i+1,port_num, j, ctlp->AiopIO[0]);
		
//                                   012345678901234567890123456789
static char *chan_init_error_str = {"brd# port# failed."};

	  chan_init_error_str[3] = i+'1';
	  chan_init_error_str[9] = j+'1';
	  puts(chan_init_error_str);
	  //puts("Channel init failed.");
#ifdef COMMENT_OUT
		if (i == 0) puts("CTL1");
		if (i == 1) puts("CTL2");
		if (i == 2) puts("CTL3");
		if (i == 3) puts("CTL4");
#endif
		return 1;
	  }
	  ++port_num;
	}
  }
#else
  ctlp = &sController[0];

  for (j=0; j<nPorts; j++)
  {
	Port= &RcktPort[j];
	cp= &Port->ch;
	pn = j + comOffset;
	sEnAiop(ctlp, pn/ctlp->PortsPerAiop);  // enable Aiopic Access
	if (sInitChan(ctlp,cp,pn/ctlp->PortsPerAiop, pn%ctlp->PortsPerAiop) != 0)
	{

		//printf("ctrl%d, port:%d, pb:%d addr:%x\n",
		//       i+1,port_num, j, ctlp->AiopIO[0]);
//                                   012345678901234567890123456789
static char *chan_init_error_str = {"brd# port# failed."};

	  chan_init_error_str[3] = '1';
	  chan_init_error_str[9] = j+'1';
	  puts(chan_init_error_str);
	  //puts("Channel init failed.");
	  return 1;
	}
  }
#endif
  ctlp = &sController[0];  // must point to first board(which does IRQ)

#ifdef SOFT_BUFS
  QOutSize = (SB_TOTAL_SIZE / nPorts) & 0xfffe;
#endif

  //----- init the RcktPort structure
  for (pn=0; pn<nPorts; pn++)
  {
	Port= &RcktPort[pn];
	cp= &Port->ch;

#ifdef SOFT_BUFS
	Port->QOutBase = &soft_buf[QOutSize * pn];  // set base of buffer
#endif
	init_port_software(Port);
	init_port_hardware(Port);
  }

////////// Setup the PIC and vectors for the choosen IRQ and enable interrupts //////////
  if (irq != 0)
  {
#ifdef EVENT_PROCESS
	switch (events_option)
	{
	  case 0:  stat = sInitIRQ(ctlp, irq, FREQ_137HZ, TRUE);  break;
	  case 1:  stat = sInitIRQ(ctlp, irq, FREQ_137HZ, FALSE); break;
	  case 2:  stat = sInitIRQ(ctlp, irq, FREQ_DIS, FALSE);   break;
	  case 3:  stat = sInitIRQ(ctlp, irq, FREQ_DIS, FALSE);   break;
	}
#else

#ifdef TEST_FREQ
	if (bus_style == BUS_PCI)
	  stat = sInitIRQ(ctlp, irq, FREQ_560HZ, TRUE);  // only PCI capable
	else
	  stat = sInitIRQ(ctlp, irq, FREQ_274HZ, TRUE);
#else
	stat = sInitIRQ(ctlp, irq, FREQ_137HZ, TRUE);
#endif

#endif

	if (stat)
	{
	  puts("IRQ Settings Bad");
	  return 1;
	}

	hook_irq(irq);      // hook irq and setup PIC

	sEnGlobalInt(ctlp); // enable INTs from the board
#ifdef EVENT_PROCESS
	if (events_option)
	{
	  sDisInterrupts(cp, RXINT_EN | SRCINT_EN | MCINT_EN | TXINT_EN);
	  sEnInterrupts(cp, SRCINT_EN | MCINT_EN | CHANINT_EN);

	  //sEnInterrupts(cp, SRCINT_EN | CHANINT_EN);
	  //sEnInterrupts(cp, RXINT_EN | SRCINT_EN | MCINT_EN | TXINT_EN | CHANINT_EN);
	  //sSetRxTrigger(ChP,TRIG_1_2);
	}
#endif

	if (events_option != 2)
	{
	  use_timer = 0;      // then don't use timer
	}

	if (events_option < 2)
	{
	  tick_delay(2);
	  poll_ticks = 0;
	  tick_delay(6);
	  if (poll_ticks == 0) 
		tick_delay(8);
	  if (poll_ticks == 0)
	  {
		sDisGlobalInt(ctlp); // enable INTs from the board
		puts("Interrupts Not Working");
		return 1;
	  }
	}
	puts("IRQ OK");
  }

  sDisable();
  //---------- Patch us in to replace INT14 (COMM BIOS)
  OldIsr14= get_vect(0x14);
  {
	WORD our_cs;
	 ASM {mov our_cs, CS}
	 set_vect(0x14,(DWORD)MK_FP(our_cs, RCKTIsr14));
	 DTRACE("i14 s");
  }

  if (use_timer)
  {
	WORD our_cs;
	OldIsr8= get_vect(0x8);
	//---------- Patch us in to replace INT8 (TIMER TICK)
	ASM {mov our_cs, CS}
	 set_vect(0x8,(DWORD)MK_FP(our_cs, RCKTIsr8));
	 DTRACE("i8 s");
  }
  sEnable();

#ifdef DO_DOS_NAMES
  hook_name();
#endif

  //-------------- Closing statements
  puts("Initialized");

  return 0;  // ok
}

/*----------------------------------------------------------------------------
| deinstall - unhook and unload TSR.
|----------------------------------------------------------------------------*/
int deinstall(int free_flags)
{
  puts("Unload");

  disable();
  // restore original int14 vector
  set_vect(0x14, OldIsr14);
  DTRACE("i14 c");

  if (use_timer)
  {
	 set_vect(0x8, OldIsr8);
	 DTRACE("i8 c");
  }

  if (irq != 0)
  {
	 sDisGlobalInt(ctlp);
	 set_vect(IRQData[irq].IntNum, OldIsr);
	 DTRACE("ix c");
  }
  enable();

  if (free_flags & 1)
  {
	ASM {
	  MOV     ES,CS:[2CH]; // deallocate environment string ram
	  MOV     AH,49H
	  INT     21H
  
	  MOV     AX,CS  ; //deallocate psp program ram
	  MOV     ES,AX
	  MOV     AH,49H ; //free allocated memory
	  INT     21H
	}
  }
  return 0; // ok
}

/*----------------------------------------------------------------------------
| do_cmd_line - process command line parameters.
|----------------------------------------------------------------------------*/
int do_cmd_line(char *line, int loaded)
{
 int i,k;
 static int j;
 int help_flag = 0;
 char *buf;
// int index;
 int name_index;
 int argc = 1;
 

  buf = line;
  while (*buf != 0)
  {
	while (*buf == ' ')
	  ++buf;
	//puts("....");
	//puts(buf);

	if ((*buf == '-') || (*buf == '/'))
	{
	  ++buf;
	  ++argc;

	  // index = -1;             // look for port number index -l38400:1
	  j = 0;
	  while ((buf[j] != 0) && (buf[j] != ' '))
	  {
		if (buf[j] == ':')
		{
#ifdef COMMENT_OUT
		  index = getint(&buf[j+1], NULL);
		  for (i=0; i<MAX_NUM_PORTS; i++)
		  {
			if (RcktPort[i].dev == index)
			  break;
		  }
		  if (i >= MAX_NUM_PORTS)
			   index = -1;
		  else index = i;
#endif
		}
		++j;
	  }

	  j = toupper(*buf++);
	  if (loaded)
	  {
		if (j != 'U')
		{
		  puts("Unload with -uk");
		}
	  }

	  switch(toupper(j))
	  {
		case 'A': pass_thru_flag = 1; break;
		case 'B': bus_style = BUS_PCI; break;
		case 'C': comFirst = getint(buf, NULL) - 1; break;
		case 'D': only_one_instance = 0; break;
		case 'E': bus_style = BUS_MC; break;
		case 'F': 
		  default_flow_ctrl=getint(buf, NULL);  // 0=none, 1=rts/dtr, 2=xon/xoff, 3=rs485
		break;
		case 'H': help_flag = 1; break;
//        case 'G': clk_prescaler = OLD_CLOCK_PRESC; break;  // os2 compatibility
// take this out, too complicated with rplus support, if needed then update
// os driver to do 460K.
		case 'J': driver_mode |= BIOS_MODE; break;  // bios mode
#ifdef ALLOW_FOUR_BOARDS
		case 'K': 
		  k = 0;
		  while ((*buf >= '0') && (*buf <= '9') && (k< 4))
		  {
			isa_sizes[k++] = getint(buf, &j);
			//puts("Size:");  puts(buf);
			buf += j;
			if (*buf == ',') ++buf;
		  }
		break;
#endif
		case 'M':
		  switch(upcase(*buf++))
		  {
			case 'N': nPorts = getint(buf, NULL); break;
			case 'M': shared = 0; break;
		  }
		break;
#ifndef ALLOW_FOUR_BOARDS
		case 'O':
		  comOffset = getint(buf, NULL);
		break;
#endif
		case 'T':
		  switch(upcase(*buf++))
		  {
			case 'A': tick_timeout = getint(buf, NULL); break;
			case 'W': wait_option = 1; break;
			case 'N': use_timer   = 0; break;
			case 'T': test_flags  = 1; break;
			case 'E': events_option = getint(buf, NULL); break;
		  }
		break;

		case 'P':
		  Boardio = gethint(buf, NULL);
#ifdef ALLOW_FOUR_BOARDS
		  k = 0;
		  while ((*buf >= '0') && (*buf <= '9') && (k< 4))
		  {
			isa_addr[k++] = gethint(buf, &j);
			//puts("Addr:");  puts(buf);
			buf += j;
			if (*buf == ',') ++buf;
		  }
#endif
		break;

		case 'U':
		  switch(upcase(*buf++))
		  {
			case 'I':
			  if (!loaded)
				irq = getint(buf, NULL);
			break;

			case 'K':
			  if (loaded)
				deinstall(1);
			break;
		  }
		break;

		case 'L':
		  locked_baud = getlint(buf, NULL);
		  // printf("locked baud = %ld\n",locked_baud);
		break;
	
		case 'Z':
		  debug_level = getint(buf, NULL);
		break;

		default:
		  puts("Bad option");
		  help_flag = 1;
		break;
	  }
	}
	else
	{
	  puts("Expecting Option -##");
	}
	while ((*buf != ' ') && (*buf != 0))
	  ++buf;
  }

  if ((help_flag) || (argc == 1))
  {
#ifdef SOFT_BUFS
	puts("Use>rcktfoss -options");
#endif
#ifdef ALLOW_FOUR_BOARDS
	puts("Use>rfmulti -options");
#endif
#ifdef SMALL
	puts("Use>rfsmall -options");
#endif
	puts("Where options:");
	puts("-c#  COM# num to start assigning(def:-c1=COM1).");
	puts("-l#  locked baud to use on all ports.");
#ifdef ALLOW_FOUR_BOARDS
	puts("-k#,#,#,#  Where # are the number of ports per board.");
#else
	puts("-o#  Offset, first port to use on RocketPort, 0 based(default 0).");
#endif
#ifdef ALLOW_FOUR_BOARDS
	puts("-p#,#,#,#  Base address per board(hex).");
#else
	puts("-p#  RocketPort base address in HEX(default 180H).");
#endif
	puts("-mn# Number of ports(default 1).");
	puts("");
	puts("(-b=PCI -e=MCA -j=bios -ui#=intr, -g=OS/2, -j=Int14)");
	puts("-h   This help.");
//    puts("-tw  wait for room in queue on outputs.");
	puts("-z#  Set Debug Level.");
	puts("SEE README.TXT file for other options");
#ifdef SOFT_BUFS
	puts("Example>rcktfoss -p180 -c3 -l38400");
#endif
#ifdef ALLOW_FOUR_BOARDS
	puts("Example>rfmulti  -p180,100 -k32,8 -mn40");
#endif
#ifdef SMALL
	puts("Example>rfsmall  -p180 -c3 -l38400");
#endif
	return 1;
  }

  comLast = comFirst + (nPorts-1);  // used by asm end to check range

  return 0;
}

/*----------------------------------------------------------------------------
| send_cmd_line - send command line parameters to resident TSR.
|  uses the Bios function 0 to pass the info.
|----------------------------------------------------------------------------*/
int send_cmd_line(char *line)
{
 WORD our_si;
 WORD our_cx;

  // ASM {int 1}
  //our_si = FP_SEG(line);

  ASM {mov our_si, cs}
  our_cx = FP_OFF(line);

  ASM {
	mov si, our_si;  // ptr
	mov cx, our_cx;
	mov ax, 16ffH ;  // func=16, parms=ff(key).
	mov bx, 5555H ;  // key indicating our special function
	int 14H;
  }
 return 0;
}

/*----------------------------------------------------------------------
| GetDevInfo - Given Bus, Dev, Func find out the allocated irq and io-
|   address from the PCI config space.
|--------------------------------------------------------------------*/
int GetDevInfo(DEVINFO *pci_config)
{
 int Bus, Dev, Func;
// int result = 0;
 static WORD data_word;
 static BYTE data_byte;
 static WORD off_base_addr;

  Bus  = pci_config->Bus_num;
  Dev  = pci_config->Dev_num;
  Func = pci_config->Fctn_num;
	
  /*
   if <= than 0x000F then original RocketPort PCi card  
  */
  if( pci_config->DevId <= 0x000F )
  {
	 off_base_addr = OFF_BASE_ADDR;
  }  
  else /*assume RP uPCi fitted PLX9030 bridge*/
  {
	 off_base_addr = PLX9030_OFF_BASE_ADDR;
  }
  
  if (!pci_read_word(Bus, Dev, Func, off_base_addr, (WORD far *) &data_word))
	  pci_config->Base_Addr = data_word-1;
  else
  {
	return 1;
   // printf("ERROR reading config word \n\n");
  }

  if(!pci_read_byte(Bus, Dev, Func, OFF_INT_LINE, (BYTE far *) &data_byte))
	 pci_config->Irq = data_byte;
  else
  {
	return 1;
   // printf("ERROR reading config byte \n\n");
  }

  if(!pci_read_word(Bus, Dev, Func, OFF_DEVICE_ID, (WORD far *) &data_word))
	pci_config->DeviceId = data_word;

  
  return 0;
}

/*-----------------------------------------------------------------------
 FindDevs - This function finds all the PCI devices with the specified
  vendor &device id.
|------------------------------------------------------------------------*/
int FindDevs(int vendor_id, int max_device_id)
{
  int found = 0;
  int stat;
  int ctlindex;
  int devindex;
  int devid;
  static int bus_number, function_number, device_number;

	 // loop thru all PCI slots
  for ( ctlindex=0; ctlindex<MAX_BRDS; ctlindex++)
  {
	for (devindex=max_device_id; devindex>0; devindex--)
	{
	  
	  devid = devindex;

	  if( devid == 0x800 )
	  {
		/*Done all uPCi cards, drop down to original RP PCi cards
		  don't want to look for non-comtrol PCi cards */
		devindex = devid = 0x000F;
	  }
	  stat = pci_find_device(devid,       //devindex,    // device ID
							 vendor_id,   // vendor ID
							 ctlindex,    // next index
							 (int *) &bus_number,
							 (int *) &function_number,
							 (int *) &device_number);
	  if (stat == 0) // ok
	  {        /* found a device */
//        if (verbose) printf("Found Board\n");
		pci_config[num_pci_s].DevId = devindex;
		pci_config[num_pci_s].Bus_num = bus_number;
		pci_config[num_pci_s].Dev_num = device_number;
		pci_config[num_pci_s].Fctn_num = function_number;
		if (num_pci_s < (MAX_BRDS-1))
		  num_pci_s++;
		else
		  return 1;  // err: too many boards found.
		found = 1;
	  }       /* found a device */
	}           /* device id loop */
  }       /* controller for loop */
  
  if (!found)
  {
	return 2; // err: none found;
  }

  return 0; // ok
}

#ifdef DO_DOS_NAMES
/*----------------------------------------------------------------------------
| hook_name - hook into driver list as named dos device.
|----------------------------------------------------------------------------*/
static int hook_name(void)
{
 char far *buf;
 struct dos_name_dev far *dos_dv;
 struct dos_name_dev *odv;
 int i,j,k;
 char tmpstr[10];

  for (i=0; i<nPorts; i++)
  {
	odv = (struct dos_name_dev *) &named_dev;  // space in asm
	odv->next = (struct dos_name_dev far *) -1L;
	
	for (j=0; j<8; j++)
	  odv->name[j] = ' ';

	odv->name[0] = 'C';
	odv->name[1] = 'O';
	odv->name[2] = 'M';
	j =comFirst+i+1;
	if (j < 10)  // COM
		 k = 1; // places
	else k = 2;
	intoa(&odv->name[3], j, k);

	odv->attr = 0x8000;
//    odv->strat = ((WORD) RCKT_Strat) - ((WORD) odv);
//    odv->inter = ((WORD) RCKT_Intr) - ((WORD) odv);
	odv->strat = (WORD) RCKT_Strat;
	odv->inter = (WORD) RCKT_Intr;
  }

  buf = (char far *)0x400L;  // start searching somewhere in low mem
  dos_dv = NULL;
						 // search for AUX dos devices
  for (i=0; i<10000; i++)
  {
	if ((buf[0] == 'A') && (buf[1] == 'U') &&
		(buf[2] == 'X') && (buf[3] == ' '))
	{
	  dos_dv = (struct dos_name_dev far *) (buf - 10);
	  break;
	}
	++buf;
  }

  if (dos_dv == NULL)
  {
	puts("Could not find device chain\n");
	return 1;  // err
  }

  i = 0;
  while ((dos_dv->next != -1L) && (i<200))
  {
	dos_dv = dos_dv->next;
	++i;
  }
  if (i > 150)
  {
	puts("Bad chain end\n");
	return 1;
  }

  dos_dv->next = (struct dos_name_dev far *) odv;

//  dos_dev[0].next = (struct dos_name_dev far *) -1L;

  puts("Linked");
  for (i=0; i<8; i++)
  {
	tmpstr[i] = odv->name[i];
  }
  tmpstr[5] = 0;
  puts(tmpstr);

  return 0;
}

/*----------------------------------------------------------------------------
| strat_c - Dos Named device, es:bx points to request_struct.
   The strategy routine is called prior to the interrupt routine(intr_c)
   and queues up the io request packet.
|----------------------------------------------------------------------------*/
void strat_c(struct dos_req_header far *req)
{
//  struct dos_req_header far *req;
//static char tmpstr[8];
static char *tmpstr = {"0123456"};

//  DEB_outs("Strategy",1);
  glob_req = req;

  if (debug_level & 4)
  {
	tmpstr[4] = 'x';
	tmpstr[5] = 'y';
	hextoa(&tmpstr[4], req->command, 2);
	tmpstr[3] = 'C';
	tmpstr[6] = 0;
	DEBV_outs(&tmpstr[3],1);
  }

#ifdef COMMENT_OUT
  switch(req->command)
  {
	case 3:  // ioctl input
	case 4:  // input
	case 5:  // input, no-wait, non-destructive
	case 6:  // input status
	case 7:  // input flush
	case 8:  // output
	case 9:  // output with verify
	case 10:  // output status
	case 11:  // output flush
	case 12:  // ioctl output
  }
#endif
  req->status = 0x100; // done flag
}

/*----------------------------------------------------------------------------
| intr_c - part of DOS named device driver, called from assembly.
|----------------------------------------------------------------------------*/
void intr_c(void)
{
//static char tmpstr[8];
static char *tmpstr = {"0123456"};

  if (debug_level & 4)
  {
	tmpstr[4] = 'x';
	tmpstr[5] = 'y';
	hextoa(&tmpstr[4], 0x123, 2);
//    hextoa(&tmpstr[4], glob_req->command, 2);
	tmpstr[3] = 'I';
	tmpstr[6] = 0;
	DEBV_outs(&tmpstr[3],1);
  }

//  req = MK_FP(m_es, m_bx);
//  DEB_outs("Interrupt",1);
}
#endif

/*----------------------------------------------------------------------------
| hook_irq - hook into IRQ table, and setup PIC.
|----------------------------------------------------------------------------*/
static int hook_irq(int irq)
{
  WORD PICio;
  BYTE mask;

  sDisable();

  OldIsr = get_vect(IRQData[irq].IntNum);
  {
	WORD our_cs;
	 ASM {mov our_cs, CS}
	 set_vect(IRQData[irq].IntNum, (DWORD)MK_FP(our_cs,RCKTIsr));
	 DTRACE("ix s");
  }

  PICio= (irq > 7) ? 0xa1 : 0x21;    // setup the 8259
  mask= inp(PICio);
  mask &= IRQData[irq].Mask;
  sOutB(PICio, mask);
  sEnable();
  return 0;
}

/*----------------------------------------------------------------------------
| fossil_loaded - Determine if this driver is already loaded.
|----------------------------------------------------------------------------*/
int fossil_loaded(void)
{
  unsigned char far *ptr = (unsigned char far *) get_vect(0x14);
  ptr += 6; // offset 6 has special marker
  if (*ptr++ != 0x54) return 0; // no-fossil loaded
  if (*ptr++ != 0x19) return 0;
  ++ptr;  // skip max Foss supported.
  if (*ptr++ != 'R') return 1;  // fossil loaded, not ours
  if (*ptr++ != 'o') return 1;  // fossil loaded, not ours
  if (*ptr++ != 'c') return 1;  // fossil loaded, not ours
  if (*ptr++ != 'k') return 1;  // fossil loaded, not ours
  if (*ptr++ != 'e') return 1;  // fossil loaded, not ours
  return 2;  // our fossil loaded.
}

/*----------------------------------------------------------------------------
| tick_delay - simple 18ticks per sec delay.
|----------------------------------------------------------------------------*/
void tick_delay(WORD ticks)
{
 WORD st, ct;

  st = getdosticks();
  ct = getdosticks();

  while ((ct-st) < ticks)
  {
	kill_time(1);
	ct = getdosticks();
  }
}

/*----------------------------------------------------------------------------
| kill_time - allow other tasks to run, kill some time, check for timeout
  condition
|----------------------------------------------------------------------------*/
int kill_time(WORD base_time)
{
 WORD ct;

  sEnable();  // make sure interrupts enabled, otherwise clock never
			 // gets bumped.

  DEB_twidle(TW_KILL_TIME);

  ct = getdosticks();
  if ((ct-base_time) > TICK_TIMEOUT) // timeout
  {
	return 1;  // err: timeout
  }

  return 0;
}

/*----------------------------------------------------------------------------
| Reboot - Watchdog timeout, reboot machine!
|----------------------------------------------------------------------------*/
void Reboot(void)
{
  WORD PICio;
  BYTE mask;

  if (irq != 0)
  {
	sDisable();
	PICio= (irq > 7) ? 0xa1 : 0x21;    // restore the 8259
	mask= inp(PICio);
	mask |= IRQData[irq].Mask;
	sOutB(PICio, mask);
	sDisGlobalInt(ctlp);    // disable INTs from the board
	sEnable();
  }

  ASM {
	mov ax,0x62
	cli
	push ax
	mov ax,0xf000
	push ax
	mov ax,0xfff0
	push ax
	xor ax,ax
	iret
  }
}

/*----------------------------------------------------------------------------
| init_port_software - initialize Port software to default settings.
|----------------------------------------------------------------------------*/
static void init_port_software(RCKTPort_t *Port)
{

//  memset (Port, 0, sizeof(*Port));   // our port struct
  
#ifdef SOFT_BUFS
  Port->QOutGet = 0;
  Port->QOutPut = 0;
  Port->QOutCount = 0;
#endif
  Port->Status = 0;  // flags
  Port->allow_isr_access = 0;  // ISR Tx Access
  Port->Status = 0;
  Port->bitRate= 38400;    // default 38400,N,8,1
  Port->dataBits= 1;  // 8
  Port->stopBits= 0;  // 1
  Port->parityMode= 0;
  Port->baud_no_access = 0;
}

/*----------------------------------------------------------------------------
| init_port_hardware - initialize Port Hardware to default settings.
   This is ran on driver initialization time only.
|----------------------------------------------------------------------------*/
static void init_port_hardware(RCKTPort_t *Port)
{
  CHANNEL_T *cp;
  cp= &Port->ch;

  sClrBreak(cp);
  sDisRTSFlowCtl(cp);
  sDisCTSFlowCtl(cp);
  sDisRxStatusMode(cp);
  sDisTxSoftFlowCtl(cp);
  sDisIXANY(cp);
  sClrTxXOFF(cp);
  sFlushTxFIFO(cp);
  sFlushRxFIFO(cp);

  if (events_option == 0)
  {
	sEnRxIntCompare1(cp,ASCII_CTRLC);
	sEnRxIntCompare2(cp,ASCII_CTRLK);
	sSetRxTrigger(cp, TRIG_1);
  }
  Port->Status = 0;

  sInW(cp->ChanStat);

  sSetTxXONChar(cp, ASCII_XON);
  sSetTxXOFFChar(cp, ASCII_XOFF);

  RPortConfigure(Port);

  sEnRxFIFO(cp);
  sEnTransmit(cp);

  set_rts_auto(Port, 0);
  set_cts_auto(Port, 0);

  sSetRTS(cp);
  sSetDTR(cp);
  Port->MCRVal |= 3;  // RTS | DTR;

  switch (default_flow_ctrl)
  {
	case 0:  // None
	break;
	case 1:  // RTS & CTS auto flow control
	  set_rts_auto(Port, 1);
	  set_cts_auto(Port, 1);
	break;
	case 2:  // XON & XOFF software flow control
	  sEnTxSoftFlowCtl(cp);
	  Port->Status |=  (S_TX_SOFT_CHK | S_RX_SOFT_CHK);
	break;
	case 3:  // RS485
	  set_rts_auto(Port, 2);  // rs485
	break;
  }
}

/*----------------------------------------------------------------------------
| RPortConfigure -
|----------------------------------------------------------------------------*/
void RPortConfigure(RCKTPort_t *Port)
{
  CHANNEL_T *cp;
  cp= &Port->ch;

	// do this here incase Windows driver reset them on us.
  sEnRxProcessor(cp);             // start the Rx processor
  if (events_option == 0)
  {
	sEnRxIntCompare1(cp,ASCII_CTRLC);
	sEnRxIntCompare2(cp,ASCII_CTRLK);
	sSetRxTrigger(cp, TRIG_1);
  }

  sEnRxFIFO(cp);
  sEnTransmit(cp);

  if (locked_baud > 0)
  {
	sSetBaudRate(cp, locked_baud, TRUE);
	sSetStop1(cp);
	sDisParity(cp);
	sSetData8(cp);
	sSetRxMask(cp,0xff);
	return;
  }

  sSetBaudRate(cp, Port->bitRate, TRUE);

  switch(Port->parityMode)      // set parity
  {
	case 0:          // no parity
	  sDisParity(cp);
	  break;
	case 1:          // odd parity
	  sEnParity(cp);
	  sSetOddParity(cp);
	  break;
	case 2:          // even parity
	  sEnParity(cp);
	  sSetEvenParity(cp);
	  break;
  }

  switch(Port->dataBits)
  {
	case 0:
	  sSetData7(cp);
	  sSetRxMask(cp,0x7f);
	  break;
	case 1:
	  sSetData8(cp);
	  sSetRxMask(cp,0xff);
	  break;
  }

  switch(Port->stopBits)      // set stop bits
  {
	case 0:
	  sSetStop1(cp);
	break;
	case 1:
	  DEB_outs("STP-2! ",1);
	  sSetStop2(cp);
	break;
  }
}
	
/*----------------------------------------------------------------------------
| RPortTxFlush -
|----------------------------------------------------------------------------*/
void RPortTxFlush(RCKTPort_t *Port)
{
  CHANNEL_T *cp;
  static BYTE junk;

  cp= &Port->ch;

#ifdef SOFT_BUFS
  Port->allow_isr_access = 0;  // ISR Tx Access Off
  Port->QOutCount = 0;
  Port->QOutPut = 0;
  Port->QOutGet = 0;
  sFlushTxFIFO(cp);
  sFlushTxPriorityBuf(cp, &junk);
#endif

  sFlushTxFIFO(cp);
  sFlushTxPriorityBuf(cp, &junk);

  Port->allow_isr_access = 1;  // ISR Tx Access On
}

/*----------------------------------------------------------------------------
| RPortRxFlush -
|----------------------------------------------------------------------------*/
void RPortRxFlush(RCKTPort_t *Port)
{
  CHANNEL_T *cp;

  cp= &Port->ch;


  sFlushRxFIFO(cp);
}

/*----------------------------------------------------------------------------
| peek_char - Return 0xffff if no chars in input queue to peek at, otherwise
|   return peeked character.
|----------------------------------------------------------------------------*/
static WORD peek_char(CHANNEL_T *cp)
{
  WORD RxFIFOOut;
  BYTE RxByte;

  if (sGetRxCnt(cp) == 0)    // no chars waiting, so return 0xffff
  {
	return 0xffff;
  }
  else
  {  // dork directly with the ram on a port to peek at
	// the top of the RxFIFO
	sDisable();
	RxFIFOOut = sChanInWI(cp, _RXF_OUTP) * 2;
	RxByte = sChanInBI(cp, _RX_FIFO + RxFIFOOut);
	sEnable();

	return RxByte;
  }
}

/*----------------------------------------------------------------------------
| set_break - set or remove the break condition.
|----------------------------------------------------------------------------*/
static void set_break(RCKTPort_t *Port, int state)
{
  CHANNEL_T *cp = &Port->ch;

  if (state)        // send break
  {
	if (!(Port->Status & S_BREAK_ON))
	{
	  sDisTransmit(cp);  // stop the transmitter
	  while ((sInW(cp->ChanStat) & TXSHRMT)==0)   // wait for
		kill_time(1);
				// last char
				// to go out
	  Port->Status |= S_BREAK_ON;
	  sSendBreak(cp);
	}
  }
  else
  {
	if (Port->Status & S_BREAK_ON)
	{
	  sClrBreak(cp);
	  sEnTransmit(cp);
	  Port->Status &= ~S_BREAK_ON;
	}
  }
}

/*----------------------------------------------------------------------------
| set_rts - set or clear RTS
|----------------------------------------------------------------------------*/
static void set_rts(RCKTPort_t *Port, WORD state)
{
  CHANNEL_T *cp = &Port->ch;

  switch (state)
  {
	case 0:   // turn off RTS output pin 4(stop sending data!)
	  Port->MCRVal &= ~2;  // RTS;

	  if ((cp->RxControl[2] & RTSFC_EN))  // automatic
	  {
		 // then turn off automatic(or ClrRTS has no effect)
		sDisRTSFlowCtl(cp);
		sDisRTSToggle(cp);
	  }
	  sClrRTS(cp);  // De-assert the RTS signal
	  // want to be able to override RTS automatic control.
	break;

	case 1:
	  Port->MCRVal |= 2;  // RTS;
	  if (!(cp->RxControl[2] & (RTSFC_EN | RTSTOG_EN)))  // not automatic
	  {
		if (Port->Status & S_RTS_FLOW)  // should be on automatic
		  {sEnRTSFlowCtl(cp);}   // so set to automatic
		else if (Port->Status & S_RS485_FLOW)  // should be on automatic
		  {sEnRTSToggle(cp);}
		else
		  {sSetRTS(cp);}              // then just set it
	  }
	  // never do a sSetRTS() and expect hardware RTS flowctrl to work!
	break;

  }
}

/*----------------------------------------------------------------------------
| set_rts_auto -
|----------------------------------------------------------------------------*/
static void set_rts_auto(RCKTPort_t *Port, WORD state)
{
  CHANNEL_T *cp = &Port->ch;

  switch (state)
  {
	case 0:  // turn off auto-flow control
	  Port->Status &= ~S_RTS_FLOW;  // automatic
	  sDisRTSFlowCtl(cp);
	  sSetRTS(cp);  // add this, kpb
	break;

	case 1:  // turn on auto-flow control
	  Port->MCRVal |= 2;  // RTS;
	  Port->Status |= S_RTS_FLOW;  // automatic
	  sEnRTSFlowCtl(cp);
	break;

	case 2:
	  Port->Status |= S_RS485_FLOW;  // automatic
	  {sEnRTSToggle(cp);}
	break;

  }
}

/*----------------------------------------------------------------------------
| set_cts_auto -
|----------------------------------------------------------------------------*/
static void set_cts_auto(RCKTPort_t *Port, WORD state)
{
  CHANNEL_T *cp = &Port->ch;

  switch (state)
  {
	case 0:  // turn off auto-flow control
	  Port->Status &= ~S_CTS_FLOW;  // automatic
	  sDisCTSFlowCtl(cp);
	break;

	case 1:  // turn on auto-flow control
	  Port->Status |= S_CTS_FLOW;  // automatic
	  sEnCTSFlowCtl(cp);
	break;
  }
}

/*----------------------------------------------------------------------------
| set_dtr -
|----------------------------------------------------------------------------*/
static void set_dtr(RCKTPort_t *Port, WORD state)
{
  CHANNEL_T *cp = &Port->ch;

  switch (state)
  {
	case 0:   // turn off DTR output pin (stop sending data!)
	  Port->MCRVal &= ~1;  // DTR;
	  sClrDTR(cp);
	break;

	case 1:   // turn on DTR
	case 2:   // DTR auto-flow control off
	case 3:   // DTR auto-flow control
	  Port->MCRVal |= 1;  // DTR;
	  sSetDTR(cp);
	break;
  }
}

/*----------------------------------------------------------------------------
| rp_block_read -
|----------------------------------------------------------------------------*/
void rp_block_read(Caller far *caller)
{
  RCKTPort_t *Port = caller->Port;
  CHANNEL_T *cp;
  int ToMove, wToMove, RxCnt, stat;
  WORD RegES;
  WORD RegDI;
  WORD RegDX;

  cp= &Port->ch;

  RegES = caller->es;
  if (driver_mode & BIOS_MODE)
	   RegDI = caller->bx;  // bios mode
  else RegDI = caller->di;  // fossil mode

  ToMove= caller->cx;
  RxCnt= sGetRxCnt(cp);

  if (driver_mode & BIOS_MODE)   // bios does error checking
  {
	stat = sInW(cp->ChanStat);
	if (stat & (RXBREAK | RXFRAME | RXPARITY | RXFOVERFL))
	{
	  DEB_outs("ERR-R1 ",1);
	  caller->ax = 0;  // zero bytes transferred
	  caller->dx = 0xff00;  // error condition
	  caller->flags |= 0x0040; // ZF=1
	  return;
	}
  }

  if (ToMove > RxCnt)
  {
	if (driver_mode & BIOS_MODE)  // bios mode can't do partial write
	{
	  DEB_outs("ERR-R2 ",1);
	  caller->ax = 0;  // zero bytes transferred
	  caller->dx = 0xff00;  // error condition
	  caller->flags |= 0x0040; // ZF=1
	  return;
	}
	DEB_outs("ERR-R2 ",1);
	ToMove= RxCnt;
  }

  wToMove= ToMove >> 1;
  RegDX= cp->TxRxData;

  ASM {
	mov DI, RegDI
	mov DX, RegDX
	cld   // dir up
  }

  if (wToMove)
  {
	ASM {
	  mov es,RegES
	  mov cx,wToMove
	  rep insw
	}
  }

  if (ToMove & 1)    // pick up the straggling byte
  {  
	ASM {
	  mov es,RegES
	  insb
	}
  }

  caller->ax= ToMove;

  if (driver_mode & BIOS_MODE)
  {
	caller->dx= 0;  // DH  FFh  Error indication.
	caller->flags &= ~0x0040; // ZF=0, no error
  }

}

/*----------------------------------------------------------------------------
| rp_block_write -
|----------------------------------------------------------------------------*/
void rp_block_write(Caller far *caller)
{
  RCKTPort_t *Port = caller->Port;
  CHANNEL_T *cp;
  int ToMove, wToMove, TxFree;
  WORD RegDI;
  WORD RegES;
  WORD RegDX;
  WORD Put1, Put2;
  unsigned char far *buf;
  WORD time;

#ifdef SOFT_BUFS
  if (driver_mode & BIOS_MODE)
	   buf = MK_FP(caller->es,caller->bx);  // bios mode
  else buf = MK_FP(caller->es,caller->di);  // fossil mode

//  cp= &Port->ch;
  ToMove= caller->cx;
  TxFree = QOutSize - Port->QOutCount; // more data than space
  if ((wait_option) || (driver_mode & BIOS_MODE))
  {
	time = getdosticks();
	while (TxFree < ToMove)  // wait for space
	{
	  if (kill_time(time)) // timeout
		break;
	  TxFree = QOutSize - Port->QOutCount; // more data than space
	}
  }
													// Copy data into Q buffer
  if (ToMove > TxFree)
  {
	if (driver_mode & BIOS_MODE)
	{
	  DEB_outs("ERR-T1 ",1);
	  caller->ax = 0;  // zero bytes transferred
	  caller->dx = 0xff00;  // error condition
	  caller->flags |= 0x0040; // ZF=1, error
	  return;
	}
	ToMove= TxFree;
  }

  Put1 = QOutSize - Port->QOutPut; // num bytes to end of buffer

  if (Put1 < ToMove)            // Q wraps
	Put2 = ToMove - Put1;
  else                             // no Q wrap
  {
	Put2 = 0;
	Put1 = ToMove;
  }

  memcpy(&QOUT[Port->QOutPut], buf, Put1);                // copy the data

  Port->QOutPut = (Port->QOutPut + Put1) % QOutSize; // update index
  if (Put2 > 0)                                     // if Q wraps
  {
	memcpy(&QOUT[0], &buf[Put1],Put2); // copy balance of data
	Port->QOutPut = Put2;                         // update Q put index
  }

  Port->allow_isr_access = 0;  // ISR Tx Access Off

  Port->QOutCount += ToMove;                           // update Q count

				 // notice that PutTxChars() is called at app time here
				 // as well as ISR time.  This complicates life.
  PutTxChars(Port);  //*** start sending chars from our buf to rocket immediately

  Port->allow_isr_access = 1;  // ISR Tx Access On

  caller->ax= ToMove;
  return;
#else

 //------------------ NO SOFTWARE BUFFER PROCEDURE
  RegES = caller->es;
  if (driver_mode & BIOS_MODE)
	   RegDI = caller->bx;  // bios mode
  else RegDI = caller->di;  // fossil mode

  cp= &Port->ch;
  ToMove= caller->cx;
  TxFree= TXFIFO_SIZE - sGetTxCnt(cp);

  if ((wait_option) || (driver_mode & BIOS_MODE))
  {
	time = getdosticks();
	while (TxFree < ToMove)  // wait for space
	{
	  if (kill_time(time)) // timeout
		break;
	  TxFree= TXFIFO_SIZE - sGetTxCnt(cp);
	}
  }

  if (ToMove > TxFree)
  {
	if (driver_mode & BIOS_MODE)  // bios mode can't do partial write
	{
	  DEB_outs("ERR-T2 ",1);
	  caller->ax = 0;  // zero bytes transferred
	  caller->dx = 0xff00;  // error condition
	  caller->flags |= 0x0040; // ZF=1, error
	  return;
	}
	ToMove= TxFree;
  }

  wToMove= ToMove >> 1;

  //  _SI= RegDI;
  //  _DX= cp->TxRxData;

  RegDX = cp->TxRxData;

  ASM {
	cld   // dir up
	mov si, RegDI
	mov dx, RegDX
   }

  if (wToMove)
  {
	ASM {
	  mov cx,wToMove
	  push ds
	  mov ds,RegES
	  rep outsw
	  pop ds
	}
  }

  if (ToMove & 1)
  {
	ASM {
	  push ds
	  mov ds,RegES
	  outsb
	  pop ds
	}
  }

  caller->ax= ToMove;

  if (driver_mode & BIOS_MODE)
  {
	caller->dx= 0;  // DH  FFh  Error indication.
	caller->flags &= ~0x0040; // ZF=0, no error
  }
#endif
}

/*-----------------------------------------------*/
/*--------- COMMON BIOS & FOSSIL ----------------*/
/*-----------------------------------------------*/

/*----------------------------------------------------------------------------
| Bios04 - Extended Port Initialization.  Same as Int14 Bios func 04 on PS/2s.

  AL  xx   Break setting:
		   0 = No break
		   1 = Set Break
  BH  xx   Parity:
		   0 = No parity
		   1 = Odd parity
		   2 = Even parity
		   3 = Stick parity odd (not supported)
		   4 = Stick parity even (not supported)
  BL  xx   Stop bits:
		   0 = 1 stop
		   1 = 2 stop
  CH  xx   Character length:
		   0 = 5 bits
		   1 = 6 bits
		   2 = 7 bits
		   3 = 8 bits
  CL  xx   Baud rate:
		   00h   110 baud     
		   01h   150 baud     
		   02h   300 baud     
		   03h   600 baud     
		   04h   1200 baud    
		   05h   2400 baud    
		   06h   4800 baud    
		   07h   9600 baud    
		   08h   19200 baud   
		  09h   38400 baud
		  0Ah   57600 baud
		  0Bh   76800 baud
		  0Ch   115200 baud
		  0dh   230400 baud
		  0eh   460800 baud
		  0fh   921600 baud

Fossil Only:
		  80h, 28800 baud
		  81h, 38400 baud
		  82h, 57600 baud
		  83h, reserved (future 76800 baud)
		  84h, 115200 baud
		  0FFh, re-enable function 0 to set baud

|----------------------------------------------------------------------------*/
void Bios04(Caller far *caller)
{
  RCKTPort_t *Port = caller->Port;

//  static char *deb_rates[] =
//    { "11", "15", "30", "60", "120", "240", "480", "960", "1920",
//    "3840",  "5760",  "7680","11520",  "23040",  "46080", "960", "20", "180"};


 static WORD BT[16]
			// 0      1      2     3      4       5    6    7 
	   =    { 11,    15,    30,    60,   120,    240,   480, 960,
			1920,  3840,  5760,  7680, 11520,  23040, 46080, 9216};

					   //  no odd ev stick
  static BYTE Rckt_PT[4]= { 0, 1, 2, 0};
  static BYTE Rckt_ST[2]= { 0, 1 };   // where 0=1, 1=2
  static BYTE DT[4]= { 1, 1, 0, 1 };  // where 1=8, 0=7
  BYTE cl;
  BYTE times_100 = 0;

  //DEB_outs("BIOS4 ",1);

  DEB_twidle(TW_FUNC04);

  cl = caller->cx & 0xff;
  // handle fossil special needs
  if ((driver_mode & BIOS_MODE) == 0)  // fossil
  {
	if (cl == 0xff)  // then allow bios00 access to baudrate
	{
	  Port->baud_no_access = 0;
	  return;
	}
	Port->baud_no_access = 1;
  }
  Port->parityMode= Rckt_PT[(caller->bx >> 8) & 0x3];
  Port->stopBits= Rckt_ST[caller->bx & 0x1];
  Port->dataBits= DT[(caller->cx >> 8) & 3];

  // handle fossil special high 80H bit cases
  if (cl == 0x80)    Port->bitRate = 2880;
  else
  {
	if (cl==0x81) cl =  9;  // 38400
	if (cl==0x82) cl = 10;  // 57600
	if (cl==0x83) cl = 11;  // 76800
	if (cl==0x84) cl = 12;  // 11520
	if (cl==0x86) cl = 13;  // 23040
	if (cl==0x87) cl = 14;  // 46080
	if (cl==0x88) cl = 15;  // 9216
	if (cl > 17) cl = 9;  // def: 38400
	Port->bitRate = BT[cl];
	if (cl==15)  // 921600 is a big big number
	  times_100 = 1;
//  DEB_outs(deb_rates[cl],0);
  }

  Port->bitRate *= 10;
  if (times_100)
	Port->bitRate *= 10;  // 92160 * 10 = 921600

  RPortConfigure(Port);

  if (caller->ax & 1)
	   set_break(Port, 1);
  else set_break(Port, 0);

  RPortRxFlush(Port);
  RPortTxFlush(Port);

  Port->allow_isr_access = 1;  // ISR Tx Access

  caller->ax = Bios03Status(caller);
}

/*----------------------------------------------------------------------------
| Bios00 - Original Port Initialization.  Same as Int14 Bios func 0
|----------------------------------------------------------------------------*/
void Bios00(Caller far *caller)
{
	// replace 75 & 110 with 19200 and 38400 for fossil use
	// too bad for sorry suckers using 110 baud

  static WORD BT[8]= { 1920, 3840, 30, 60, 120, 240, 480, 960};
  static BYTE PT[4]= { 0, 1, 0, 2 };
  static BYTE ST[2]= { 0, 1 };  // 0=1, 1=2
  static BYTE DT[4]= { 1, 1, 0, 1 };  // 1=8, 0=7
  RCKTPort_t *Port = caller->Port;
  BYTE b;
  int baud_access = 1;

  if ((driver_mode & BIOS_MODE) == 0)  // fossil
  {
	if (Port->baud_no_access)
	  baud_access = 0;
  }

  if (baud_access)
  {
	Port->bitRate00= caller->ax & 0xe0;
	b= caller->ax >> 5;
	Port->bitRate = BT[b];
	Port->bitRate *= 10;
  }

  b= (caller->ax & 0x18) >> 3;
  Port->parityMode= PT[b];

  b= (caller->ax & 4) >> 2;
  Port->stopBits= ST[b];

  b= (caller->ax & 3);
  Port->dataBits= DT[b];

  RPortConfigure(Port);

  set_break(Port, 0);

  RPortRxFlush(Port);
  RPortTxFlush(Port);

  Port->allow_isr_access = 1;  // ISR Tx Access

  caller->ax = Bios03Status(caller);
}

/*----------------------------------------------------------------------------
| Bios01 - Transmit character and wait, standard int14 bios

  AL  xx   Character to be transmitted.
Return
  AL  xx   Character transmitted.
  AH  xx   Line status:
   2 second timeout.
|----------------------------------------------------------------------------*/
void Bios01(Caller far *caller)
{
  RCKTPort_t *Port = caller->Port;
  BYTE CharToSend;
  CHANNEL_T *cp;
  WORD TxFree;
  BYTE c;
  WORD time;

  //DEB_outs("BIOS1 ",1);
  time = getdosticks();

#ifdef SOFT_BUFS
  TxFree = QOutSize - Port->QOutCount; // more data than space
  while (TxFree < 1)  // wait for space
  {
	if (kill_time(time)) // timeout
	{
	  caller->ax = Bios03Status(caller) | 0x8000;  // timeout
	  return;
	}
	TxFree = QOutSize - Port->QOutCount; // more data than space
  }
  c = (unsigned char) caller->ax;
  QOUT[Port->QOutPut] = c;
  ++Port->QOutPut;
  if (Port->QOutPut >= QOutSize)
	Port->QOutPut = 0;

  sDisable();
  ++Port->QOutCount;
  sEnable();

#else
  cp= &Port->ch;
  while (sGetTxCnt(cp) >= TXFIFO_SIZE)  // wait for space
  {
	if (kill_time(time)) // timeout
	{
	  caller->ax = Bios03Status(caller) | 0x8000;  // timeout
	  return;
	}
  }
  CharToSend= caller->ax;

  sOutB(cp->TxRxData, CharToSend);
#endif

  caller->ax &= 0x00ff;
  caller->ax |= (Bios03Status(caller) & 0xff00);
}

/*----------------------------------------------------------------------------
| Bios02 - Get received character with wait
  DX  xxxx Channel number.
  AH  02h  Port number.
Return
  AL  xx   Character received.
  AH  xx   Line status:
|----------------------------------------------------------------------------*/
void Bios02(Caller far *caller)
{
  RCKTPort_t *Port = caller->Port;
  BYTE CharToRecv;
  CHANNEL_T *cp;
  WORD time;

  //DEB_outs("BIOS2 ",1);
  cp= &Port->ch;

  time = getdosticks();

  while (sGetRxCnt(cp) == 0)    // wait for a character
  {
	if (kill_time(time)) // timeout
	{
	  caller->ax = Bios03Status(caller) | 0x8000;  // timeout
	  return;
	}
  }
 
  CharToRecv= sInB(cp->TxRxData);
  caller->ax = (Bios03Status(caller) & 0xff00) | CharToRecv;
}

/*----------------------------------------------------------------------------
| Bios03Status - Basic bios Int14 status return(func 3)

 Line status flags.  Except for bit STAT_TIMEOUT, these mirror the
   16450 line status register.
#define LNSTAT_DATARDY    0x0100
#define LNSTAT_OVRRUN     0x0200
#define LNSTAT_PARITY     0x0400
#define LNSTAT_FRAMING    0x0800
#define LNSTAT_BREAK      0x1000
#define LNSTAT_THRE       0x2000  at least 1 empty space in Tx FIFO
#define LNSTAT_TE         0x4000
#define LNSTAT_TIMEOUT    0x8000

 Modem status flags.  These mirror the 16450 modem status register.
#define MDMSTAT_DCTS       0x0001
#define MDMSTAT_DDSR       0x0002
#define MDMSTAT_DRI        0x0004
#define MDMSTAT_DCD        0x0008
#define MDMSTAT_CTS        0x0010
#define MDMSTAT_DSR        0x0020
#define MDMSTAT_RI         0x0040
#define MDMSTAT_CD         0x0080

|----------------------------------------------------------------------------*/
WORD Bios03Status(Caller far *caller)
{
  RCKTPort_t *Port = caller->Port;
  CHANNEL_T *cp;
  WORD stat, TxCnt, ChanStatus;
  cp= &Port->ch;

  //DEB_outs("BIOS3 ",1);

  TxCnt= sGetTxCnt(cp);
  ChanStatus= sInW(cp->ChanStat);

  stat = 0;

  // 8000H set on 3 second timeout.(for fun01,02)
#ifdef SOFT_BUFS
  if ((TxCnt == 0) && (Port->QOutCount == 0))
	stat |= 0x4000;    // output buffer empty
  if (Port->QOutCount+100 < QOutSize)
	stat |= 0x2000;  // output buf not full, LNSTAT_THRE
#else
  if (TxCnt == 0) stat |= 0x4000;    // output buffer empty, LNSTAT_TE
  if (TxCnt < TXFIFO_SIZE) stat |= 0x2000;  // output buffer not full
#endif
  if (ChanStatus & RXBREAK)
  {
	DEB_outs("BREAK!",1);
	stat |= 0x1000; // break error
  }
  if (ChanStatus & RXFRAME)
  {
	DEB_outs("FRAME!",1);
	stat |= 0x800;  // rx framing error
  }
  if (ChanStatus & RXPARITY)
  {
	DEB_outs("PARITY!",1);
	stat |= 0x400;  // parity error
  }

  if (test_flags & 1)
  {
	if (ChanStatus & RXFOVERFL)
	{
	  stat |= 0x0200;  // input buffer overrun
	  DEB_outs("IN OVRFLOW ",1);
	}
  }

  if (ChanStatus & RDA)
  {
	if (sGetRxCnt(&caller->Port->ch) != 0)  // no chars waiting, so return 0xffff
	{
	  stat |= 0x0100;    // chars in input buffer
	}
	else
	{
	  if (test_flags & 1)
	  {
		DEB_outs("RDA WRONG! ",1);
	  }
	}
  }

  if (ChanStatus & CD_ACT)   stat |= 0x80;  // carrier detect signal

  if (cp->CtlP->PortsPerAiop == 4)  // rplus
  {
	if (sGetRPlusModemRI(cp))
	  stat |= 0x40; // Ring Indicator
  }

  if (ChanStatus & DSR_ACT)  stat |= 0x20; // DSR actual
  if (ChanStatus & CTS_ACT)  stat |= 0x10; // CTS actual

  if ((driver_mode & BIOS_MODE) == 0)  // fossil mode
	stat |= 8;  // Delta-CD always on for fossil(??)

#ifdef DELTA_PROCESSING
  // 8H = Delta CD
  // 4H = Trailing Edge Ring Indicator
  // 2H = Delta Data Set Ready 
  // 1H = Delta Clear To Send
  ChanStatus = sGetChanIntID(cp);
  if (ChanStatus & (DELTA_CD | DELTA_CTS | DELTA_DSR))
  {
	if ((driver_mode & BIOS_MODE) != 0)  // fossil mode
	{
	  if (ChanStatus & DELTA_CD)  stat |= MDMSTAT_DCD;
	}
	if (ChanStatus & DELTA_CTS) stat |= MDMSTAT_DCTS;
	if (ChanStatus & DELTA_DSR) stat |= MDMSTAT_DDSR;
	if (cp->CtlP->PortsPerAiop == 4)  // rplus
	{
	  if (sGetRPlusRingIntID(cp)) stat |= MDMSTAT_DRI;
	}
  }
#endif

  // special fossil driver check for char match on ctrl-c or ctrl-k
  // flag it if found
  if (Port->Status & S_CTL_CK)
  {
	if (ChanStatus & (RX1MATCH|RX2MATCH))
	  Port->Status |= S_CTL_CK_GOT;
  }

#ifdef COMMENT_OUT
  if (test_flags & 1)
  {
	hextoa(tmpstr, stat, 4);
	tmpstr[4] = 0;
	putsxy(tmpstr,    0, 5, 0x70);
  }
#endif

 return stat;
}

/*----------------------------------------------------------------------------
| 05 - Extended Communications Port Control
  AH  05h  Foss number.
  AL  xx   00h = Read Modem Control Register.
	   01h = Write Modem Control Register.
  BL  xx   If AL = 01h, and hardware handshaking is disabled:
		   setting bit 0 of BL to 1 will raise DTR;
		   setting bit 0 of BL to 0 will drop DTR;
		   setting bit 1 of BL to 1 will raise RTS;
		   setting bit 1 of BL to 0 will drop RTS;
Return
  AH  xx   Line status (see Foss 03h).
  AL  xx   Modem status (see Foss 03h).
  BL  xx   Modem Control Register:
		   Bit 0 = DTR (active if set to 1).
		   Bit 1 = RTS (active if set to 1).
|----------------------------------------------------------------------------*/
void Bios05(Caller far *caller)
{
  RCKTPort_t *Port = caller->Port;
  CHANNEL_T *cp;
  BYTE b;
  cp= &Port->ch;

//                         BL = Modem control register, where:
//                              Bits 7-5 = 0 (Reserved)
//                              Bit 4    = 1 Loopback mode
//                              Bit 3    = 1 OUT2 (interrupts) enabled
//                              Bit 2    = 1 OUT1 active
//                              Bit 1    = 1 Request to send active
//                              Bit 0    = 1 Data terminal ready (DTR) active

  if ((caller->ax & 0xff) == 1)  // write MCR
  {
	Port->MCRVal= caller->bx & 0xff;

	if (Port->MCRVal & 0x10)  // loopback
		 {sEnLocalLoopback(cp);}
	else {sDisLocalLoopback(cp);}

	if (caller->bx & 1)
	  {sSetDTR(cp);}
	else
	  {sClrDTR(cp);}

	if ((Port->Status & S_RTS_FLOW) == 0)  // not RTS automatic
	{
	  if (caller->bx & 2)
		{sSetRTS(cp);}
	  else
		{sClrRTS(cp);}
	}
  }
  caller->ax = Bios03Status(caller);
  caller->bx &= 0xff00;
  caller->bx = Port->MCRVal | 0x8;
}

#ifdef WIERD_BIOS_MODE
/*-----------------------------------------------*/
/*-------------- BIOS    ------------------------*/
/*-----------------------------------------------*/

/*----------------------------------------------------------------------------
| 00 - Set communications parameters
|----------------------------------------------------------------------------*/
void Rckt00(Caller far *caller)
{
  DEB_twidle(TW_FUNC00);
  Bios00(caller);
}

/*----------------------------------------------------------------------------
| 01 - Transmit character and wait
  DX  xxxx Channel number.
  AH  01h  Foss number.
  AL  xx   Character to be transmitted.
Return
  AL  xx   Character transmitted.
  AH  xx   Line status:
   2 second timeout.
|----------------------------------------------------------------------------*/
void Rckt01(Caller far *caller)
{
  DEB_twidle(TW_FUNC01);
  Bios01(caller);
}

/*----------------------------------------------------------------------------
| 02 - Get received character with wait
  DX  xxxx Channel number.
  AH  02h  Foss number.
Return
  AL  xx   Character received.
  AH  xx   Line status:
|----------------------------------------------------------------------------*/
void Rckt02(Caller far *caller)
{
  DEB_twidle(TW_FUNC02);
  Bios02(caller);
}

/*----------------------------------------------------------------------------
| 03 - Return Serial Port Status
  DX  xxxx Channel number.
  AH  03h  Foss number.
Return
  AH  FFh  Error.  Otherwise:
  AH  xx   Line status:
|----------------------------------------------------------------------------*/
void Rckt03(Caller far *caller)
{
  DEB_twidle(TW_FUNC03);
  caller->ax = Bios03Status(caller);
}

/*----------------------------------------------------------------------------
| 04 - Extended Port Initialization

  AL  xx   Break setting:
		   0 = No break
		   1 = Set Break
  BH  xx   Parity:
		   0 = No parity
		   1 = Odd parity
		   2 = Even parity
		   3 = Stick parity odd (not supported)
		   4 = Stick parity even (not supported)
  BL  xx   Stop bits:
		   0 = 1 stop bit NOTE - 1.5 stop bits can only be
		   1 = 2 stop bitsset by the configuration utility.
  CH  xx   Character length:
		   0 = 5 bits
		   1 = 6 bits
		   2 = 7 bits
		   3 = 8 bits
  CL  xx   Baud rate:
		   00h   110 baud      09h   38400 baud
		   01h   150 baud      0Ah   57600 baud
		   02h   300 baud      0Bh   76800 baud
		   03h   600 baud      0Ch   115200 baud
		   04h   1200 baud     0Dh   50 baud
		   05h   2400 baud     0Eh   75 baud
		   06h   4800 baud     0Fh   134 baud
		   07h   9600 baud     10h   200 baud
		   08h   19200 baud    11h   1800 baud

|----------------------------------------------------------------------------*/
void Rckt04(Caller far *caller)
{
  DEB_twidle(TW_FUNC04);
  Bios04(caller);
}

/*----------------------------------------------------------------------------
| 06 - Get Driver/Board Information
  AH  06h  Foss number.

Sub-Foss 00h: Get port name
  AL  00h  Sub-Foss number.

Sub-Foss 01h: Get driver information
  AL  01h  Sub-Foss number.

Sub-Foss 02h: Get board information
  AL  02h  Sub-Foss number.
  BX  xxxx Board number (0-3).

Sub-Foss FFh: Get driver name
  AL  FFh  Sub-Foss number.

Return
Sub-Foss 00h: Port name:
  AH  FFh  Error indication - otherwise:
  AL  xx   Number of highest Interrupt 14h Foss supported.
  ES:BX    Pointer to 8 byte ASCII string which contains  the
		   name of the port (e.g. "COM5    ") if DOS support  is
		   enabled  (done through the "Options" menu  in  driver
		   set-up  program).   If DOS support is  disabled  (Int
		   14h   support   only),   the  string   will   contain
		   "NoDriver".

Sub-Foss 01h: Driver information:
  AH  FFh  Error indication - otherwise:
  AX  xxxx Total number of channels supported.
  BX  xxxx Driver version number (e.g. V 2.0.1: BX = 0201h).
  CX  xxxx Total number of boards supported.
  DX  xxxx Lowest channel number supported by this driver - to  get
		   highest  supported channel number,  use  the  formula
		   AX+DX-1.

Sub-Foss 02h: Board information:
  AH  FFH  Error indication.  Otherwise:
  AH  xx   IRQ number.
  AL  xx   Board type:
			   01h  COM/Xi
			   02h  MC/Xi
			   03h  PC/Xe
			   04h  PC/Xi
			   05h  PC/Xm
  BX  xxxx Board's dual-ported memory segment.
  CX  xxxx Number of channels on the board.
  DX  xxxx Board's I/O port address.
  SI  xxxx Channel number of first channel on the board.

Sub-Foss FFh: Driver name:
  AX  xxxx Driver version number.
  CX  xxxx Number of channels supported.
  ES:BX    Pointer  to 8-byte string containing the  driver's
		   name (e.g. "RcktFEP5").

|----------------------------------------------------------------------------*/
void Rckt06(Caller far *caller)
{
 char *id_str = {"RcktFEP5"};
 char *id_port = {"NoDriver"};
// RCKTPort_t *Port = caller->Port;

  DEB_twidle(TW_FUNC06);

  switch(caller->ax & 0xff)
  {
	case 0:
	  caller->ax = 0x1f;  // highest func num
	  caller->es = FP_SEG(id_port);
	  caller->bx = FP_OFF(id_port);
	break;
	case 1:
	  caller->ax = nPorts;
	  caller->bx = 0x201;  // vers
	  caller->cx = 1;
	  caller->dx = comFirst;
	break;
	case 2:
	  caller->ax = (irq << 8) | 4;   // 4??
	  caller->bx = 0xB800;  // video mem.
	  caller->cx = nPorts;
	  caller->dx = 0;
	  caller->si = 0;
	break;
	case 0xff:
	  caller->ax = 0x201;
	  caller->cx = nPorts;
	  caller->es = FP_SEG(id_str);
	  caller->bx = FP_OFF(id_str);
	break;
  }

}

/*----------------------------------------------------------------------------
| 07 - Send Break -
  DX  xxxx Channel number.
  AH  07h  Foss number.
  AL  00h  Use default BREAK time.  Otherwise:
  AL  01h  BREAK time is value in BX times 10 milliseconds.
  BX  xxxx Factor for determining length of BREAK
		   (BX times 10 milliseconds).
Return
  AH  00h  No errors.
|----------------------------------------------------------------------------*/
void Rckt07(Caller far *caller)
{
 int t;
  RCKTPort_t *Port = caller->Port;

  DEB_twidle(TW_FUNC07);

  if (caller->ax & 1)
	t = caller->bx / 8;  // 10ms * bx
  else t = 5;

  set_break(Port,1);
  tick_delay(t);
  set_break(Port,0);

  caller->ax = 0;


}

/*----------------------------------------------------------------------------
| 08 -  Alternate Status Check
  This performs a non-destructive read from  the  receive
  buffer.
Return
  ZF  0    Next character in input buffer will appear in AL.
  AL  xx   Next character to be read by a Read Single Character
		   or Read String Foss.
  AH  xx   Line status associated with next character to be read
		   (see Foss 02h).
  ZF  1    No character in buffer.
  AH  FFh  No character in buffer.
 |----------------------------------------------------------------------------*/
void Rckt08(Caller far *caller)
{
 WORD cstat;
 RCKTPort_t *Port = caller->Port;
 CHANNEL_T *cp;
 cp= &Port->ch;

  DEB_twidle(TW_FUNC08);
  caller->ax = (Bios03Status(caller) & 0xff00);
  cstat = peek_char(cp);
  if (cstat != 0xffff)
  {
	caller->ax |= cstat;
	caller->flags &= ~0x0040; // ZF=0, no error
  }
  else
  {
	caller->ax |= 0xff00;
	caller->flags |= 0x0040; // ZF=1, error
  }
}

/*----------------------------------------------------------------------------
| 09 - Clear Buffers
  This Foss clears (flushes) both the transmit and receive buffers
  for the designated channel.
|----------------------------------------------------------------------------*/
void Rckt09(Caller far *caller)
{
 RCKTPort_t *Port = caller->Port;

  DEB_twidle(TW_FUNC09);
  RPortRxFlush(Port);
  RPortTxFlush(Port);
  caller->ax = 0;
}

/*----------------------------------------------------------------------------
| 0A - Input Queue Check - Foss 0Ah
Return
  AX  xxxx Number of characters in input buffer.
|----------------------------------------------------------------------------*/
void Rckt0A(Caller far *caller)
{
 RCKTPort_t *Port = caller->Port;
 CHANNEL_T *cp;
 cp= &Port->ch;

  DEB_twidle(TW_FUNC0A);
  caller->ax = sGetRxCnt(cp);
}

/*----------------------------------------------------------------------------
| 0B - Drop Handshake Lines
Return
|----------------------------------------------------------------------------*/
void Rckt0B(Caller far *caller)
{
 RCKTPort_t *Port = caller->Port;

  DEB_twidle(TW_FUNC0B);
  set_rts(Port,0);
  set_dtr(Port,0);  // need to work on this!!!!! see doc
  caller->ax = 0;
}

/*----------------------------------------------------------------------------
| 0C - Get Channel Parameters
Return
  AH  FFh  Error indication.  Otherwise:
  AH  xx   Software flow control (XON/XOFF):
		 Bit 0 = Transmit flow control-If set to 1, XON/XOFF flow 
		   control is enabled for transmitted data.  Inactive if 
		   set to 0.
		 Bit 1 = Receive flow control-If set to 1, XON/XOFF flow 
		   control is enabled for received data. Inactive if set to 0.
  AL  xx   Hardware flow control:
		 Bit 0 = DTR receive flow control-If set to 1, DTR is used 
		   to control incoming data.
		 Bit 1 = RTS receive flow control-If set to 1, RTS is used 
		   to control incoming data.
		 Bit 4 = CTS transmit flow control-If set to 1, data 
		   transmission will be paused any time CTS (Clear To Send) 
		   is dropped by the receiving device.  If set to 0, CTS 
		   has no effect.
		 Bit 5 = DSR transmit flow control-If set to 1, data 
		   transmission will be paused any time DSR (Data Set Ready) 
		   is dropped by the receiving device.  If set to 0, DSR has 
		   no effect.
		 Bit 7 = DCD transmit flow control-If set to 1, data 
		   transmission will be paused any time DCD (Data Carrier 
		   Detect) is dropped by the receiving device.  If set to 0, 
		   DCD has no effect.
  BH  xx   Parity
		 0 = No parity              
		 1 = Odd parity     
		 2 = Even parity
  BL  xx   Stop bits
		 0 = 1 stop bit
		 1 = 2 stop bits (or 1.5, as set with configuration utility)
  CH  xx   Character length
		 0 = 5 bits
		 1 = 6 bits
		 2 = 7 bits
		 3 = 8 bits
								  19
  CL  xx   Baud rate
		 00h   110 baud      09h   38400 baud
		 01h   150 baud      0Ah   57600 baud
		 02h   300 baud      0Bh   76800 baud
		 03h   600 baud      0Ch   115200 baud
		 04h   1200 baud     0Dh   50 baud
		 05h   2400 baud     0Eh   75 baud
		 06h   4800 baud     0Fh   134 baud
		 07h   9600 baud     10h   200 baud
		 08h   19200 baud    11h   1800 baud

|----------------------------------------------------------------------------*/
void Rckt0C(Caller far *caller)
{
 RCKTPort_t *Port = caller->Port;
 CHANNEL_T *cp;
 cp= &Port->ch;

  DEB_twidle(TW_FUNC0C);

  caller->ax = 0;
  if (Port->Status & S_CTS_FLOW) caller->ax |= 0x0010;
  if (Port->Status & S_RTS_FLOW) caller->ax |= 0x0002;
  if (Port->Status & S_TX_SOFT_CHK) caller->ax |= 0x0100;
  if (Port->Status & S_RX_SOFT_CHK) caller->ax |= 0x0200;

  caller->bx = 0;  // parity, stop-bits, char-len
  if (cp->TxControl[2] & PARITY_EN)
  {
	if (cp->TxControl[2] & EVEN_PAR)
		 caller->bx = 0x200;
	else caller->bx = 0x100;
  }
  // else no parity(0)

  if (cp->TxControl[2] & STOP2) caller->bx |= 1;  // 2 stops

  if (cp->TxControl[2] & DATA8BIT) caller->cx = 0x300;  // 8
  else                             caller->cx = 0x200;  // 7

  caller->cx |= Port->bitRate00;  // bit-rate
}

/*----------------------------------------------------------------------------
| 0D - Get Pointer to Character Ready Flag
  ES:BX    Pointer to the flag.  If the byte pointed to by ES:BX is 00h, 
	   the receive buffer is empty.  If the byte contains FFh, the 
	   buffer is not empty.

Return
|----------------------------------------------------------------------------*/
void Rckt0D(Caller far *caller)
{
// RCKTPort_t *Port = caller->Port;
  caller = caller; // kill compiler warning

  DEB_twidle(TW_FUNC0D);
  //caller->es = FP_SEG((BYTE far *) &Port->chars_ready_flag);
  //caller->bx = FP_OFF((BYTE far *) &Port->chars_ready_flag);
  caller->ax = 0;
}

/*----------------------------------------------------------------------------
| 0E - Write String
  CX  xxxx Number of characters in string.
  ES:BX    Pointer to the string.  Load ES and BX with the segment 
	   and offset, respectively, of the first character of the 
	   string you wish to send.
Return
  AX  xxxx Number of characters transmitted.
  DH  FFh  Error indication.
  ZF  1    Error indication.
|----------------------------------------------------------------------------*/
void Rckt0E(Caller far *caller)
{
  DEB_twidle(TW_FUNC0E);
  rp_block_write(caller);
}

/*----------------------------------------------------------------------------
| 0F - Read String
  CX  xxxx Number of characters to read.
  ES:BX    Pointer to callers buffer.  Load ES and BX with the segment 
	   and offset, respectively, of a character array large enough 
	   to hold the number of characters specified in CX.
Return
  AX  xxxx Number of characters read.
  DH  FFh  Error indication.
  ZF  1    Error indication.
|----------------------------------------------------------------------------*/
void Rckt0F(Caller far *caller)
{
  DEB_twidle(TW_FUNC0F);
  rp_block_read(caller);
}

/*----------------------------------------------------------------------------
| 10 - Clear Receive Buffer
Return
|----------------------------------------------------------------------------*/
void Rckt10(Caller far *caller)
{
  DEB_twidle(TW_FUNC10);
  RPortRxFlush(caller->Port);
  caller->ax = 0;
}

/*----------------------------------------------------------------------------
| 11 - Clear Transmit Buffer
Return
|----------------------------------------------------------------------------*/
void Rckt11(Caller far *caller)
{
  DEB_twidle(TW_FUNC11);
  RPortTxFlush(caller->Port);
  caller->ax = 0;
}

/*----------------------------------------------------------------------------
| 12 - Get Transmit Buffer Free Space
Return
  AX  xxxx Number of free bytes in transmit buffer.
|----------------------------------------------------------------------------*/
void Rckt12(Caller far *caller)
{
 RCKTPort_t *Port = caller->Port;
 CHANNEL_T *cp;

 DEB_twidle(TW_FUNC12);
#ifdef SOFT_BUFS
  caller->ax = (QOutSize - Port->QOutCount);
#else
  cp= &Port->ch;
  caller->ax = TXFIFO_SIZE-sGetTxCnt(cp);
#endif
}

/*----------------------------------------------------------------------------
| 13 - Set Handshake Lines, turns on rts and dtr, Autoflow-Control retains
	priority and control.
Return
|----------------------------------------------------------------------------*/
void Rckt13(Caller far *caller)
{
 RCKTPort_t *Port = caller->Port;

  DEB_twidle(TW_FUNC13);
  set_rts(Port, 1);  
  set_dtr(Port, 1);  
  caller->ax = 0;
}

/*----------------------------------------------------------------------------
| 14 - Non-Destructive Character Read
Return
|----------------------------------------------------------------------------*/
void Rckt14(Caller far *caller)
{
 RCKTPort_t *Port = caller->Port;
 CHANNEL_T *cp;
 WORD cstat;

  cp= &Port->ch;

  DEB_twidle(TW_FUNC14);

  caller->ax = (Bios03Status(caller) & 0xff00);
  cstat = peek_char(cp);
  if (cstat != 0xffff) caller->ax |= cstat;
}

/*----------------------------------------------------------------------------
| 15 - Input Byte Count
Return
|----------------------------------------------------------------------------*/
void Rckt15(Caller far *caller)
{
 RCKTPort_t *Port = caller->Port;
 CHANNEL_T *cp;
 cp= &Port->ch;

  DEB_twidle(TW_FUNC15);
  caller->ax = RXFIFO_SIZE-sGetRxCnt(cp);
}

/*----------------------------------------------------------------------------
| 16 - No Function Assigned.
|----------------------------------------------------------------------------*/
void Rckt16(Caller far *caller)
{
  DEB_twidle(TW_FUNC16);
  caller->ax = caller->ax;  // get rid of compiler warning
}

/*----------------------------------------------------------------------------
| 17 - Pause/Resume Transmit

 This Foss overrides XON/XOFF output flow control.  It can  be
 used   to   immediately  pause  transmission,   or   to   restart
 transmission  which has previously been paused  (either  by  this
 command, or by a received XOFF character).

 NOTE - This  function will not override hardware flow control.

  AL  xx   If AL = 0, pause transmission. If software (XON/XOFF)  
		   flow  control is active,  transmission will resume if 
		   an XON character is received.
	   If AL = 1, resume previously paused transmission.

Return
|----------------------------------------------------------------------------*/
void Rckt17(Caller far *caller)
{
 RCKTPort_t *Port = caller->Port;
 CHANNEL_T *cp;
 cp= &Port->ch;

  DEB_twidle(TW_FUNC17);

  if (caller->ax & 1)
	sClrTxXOFF(cp);
  // else stop flow(can't do this with RocketPort)

   caller->ax = 0;
}

/*----------------------------------------------------------------------------
| 18 - Pause/Resume Receive
 This immediately activates or  releases  receive  flow
 control.

NOTE - If the number of characters in the input buffer is less than
the  Receive  Low Water Mark (see Function 1Bh - Get Water  Marks
and  Buffer  Size),  any  attempts to  pause  reception  will  be
ignored.  Similarly,  if the number of characters  in  the  input
buffer  is greater than the Receive High Water Mark, any attempts
to resume reception will be ignored.
|----------------------------------------------------------------------------*/
void Rckt18(Caller far *caller)
{
  DEB_twidle(TW_FUNC18);
  caller->ax = 0;
}

/*----------------------------------------------------------------------------
| 19 - No Function
|----------------------------------------------------------------------------*/
void Rckt19(Caller far *caller)
{
  DEB_twidle(TW_FUNC19);
  caller->ax = 0;
}

/*----------------------------------------------------------------------------
| 1A - No Function
|----------------------------------------------------------------------------*/
void Rckt1A(Caller far *caller)
{
  DEB_twidle(TW_FUNC1A);
  caller->ax = 0;
}

/*----------------------------------------------------------------------------
| 1B - Get Water Marks and Buffer Size
  This Foss returns the size of the selected buffer (receive or
  transmit),  and the "water marks" for that buffer  (see  Foss
  1Ch -Set Water Marks).
  AL  xx   0 = Get Transmit Low Water Mark.
	   1 = Get Receive Low Water Mark.
	   2 = Get Receive High Water Mark.
Return
  AX  xxxx The requested water mark.
  BX  xxxx Size  of  the  buffer  for which the  water  mark  was
		requested (Transmit if AL was 0; Receive if AL was  1
		or 2).
  DH  FFh  Error indication.
  ZF  1    Error indication.
|----------------------------------------------------------------------------*/
void Rckt1B(Caller far *caller)
{
// RCKTPort_t *Port = caller->Port;
// CHANNEL_T *cp;
// cp= &Port->ch;

  DEB_twidle(TW_FUNC1B);

  switch(caller->ax & 0xff)
  {
	case 0: // tx lo wm
	  caller->ax = 5;
#ifdef SOFT_BUFS
	  caller->bx = QOutSize;
#else
	  caller->bx = TXFIFO_SIZE;
#endif
	break;
	case 1: // rx lo wm
	  caller->ax = 5;
	  caller->bx = RXFIFO_SIZE;
	break;
	case 2: // rx hi wm
	  caller->ax = RXFIFO_SIZE - 50;
	  caller->bx = RXFIFO_SIZE;
	break;
  }
  caller->dx = 0;  // no err
  caller->flags &= ~0x0040; // ZF=0, no error
}

/*----------------------------------------------------------------------------
| 1C - Set Water Marks
Return
|----------------------------------------------------------------------------*/
void Rckt1C(Caller far *caller)
{
  DEB_twidle(TW_FUNC1A);
  caller->ax = 0;
}

/*----------------------------------------------------------------------------
| 1D - No Function
|----------------------------------------------------------------------------*/
void Rckt1D(Caller far *caller)
{
  DEB_twidle(TW_FUNC1D);
  caller->ax = 0;
}

/*----------------------------------------------------------------------------
| 1E - Set Flow Control
  BH  xx   Software flow control (XON/XOFF):
		   Bit  0  =  Transmit  flow  control - If  set  to  1,
		 XON/XOFF  flow  control will be used  for  trans
		 mitted data.  Inactive if set to 0.
		   Bit 1 = Receive flow control - If set to 1, XON/XOFF
		 flow  control  will be used for  received  data.
		 Inactive if set to 0.
		   Bit  2 = Set new XON/XOFF characters - If set to  1,
		 CH  contains  the  new  XOFF  character  and  CL
		 contains the new XON character.
  BL  xx   Hardware flow control:
		   Bit  0 = DTR receive flow control - If set to 1, DTR
		 will be used to control incoming data.
		   Bit  1 = RTS receive flow control - If set to 1, RTS
		 will be used to control incoming data.
		   Bit  4  =  CTS transmit flow control - If set to  1,
		 data  transmission will be paused any  time  CTS
		 (Clear  To  Send)  is dropped by  the  receiving
		 device.   If set to 0, CTS will have  no  effect
		 on transmission.
		   Bit  5  =  DSR transmit flow control - If set to  1,
		 data  transmission will be paused any  time  DSR
		 (Data  Set  Ready) is dropped by  the  receiving
		 device.   If set to 0, DSR will have  no  effect
		 on transmission.
		   Bit  7  =  DCD transmit flow control - If set to  1,
		 data  transmission will be paused any  time  DCD
		 (Data   Carrier  Detect)  is  dropped   by   the
		 receiving  device.  If set to 0, DCD  will  have
		 no effect on transmission.
  CH  xx   New XOFF character, if bit 2 of BH is set to 1.
  CL  xx   New XON character, if bit 2 of BH is set to 1.

Return
|----------------------------------------------------------------------------*/
void Rckt1E(Caller far *caller)
{
 RCKTPort_t *Port = caller->Port;
// CHANNEL_T *cp;
// cp= &Port->ch;

  DEB_twidle(TW_FUNC1E);
  caller->ax = 0;

  if (caller->bx & 0x100) // want tx soft flow ctrl
  {
  }

  if (caller->bx & 0x200) // want rx soft flow ctrl
  {
  }

  if (caller->bx & 0x400) // new xon/xoff flow ctrl
  {
	// ch=xoff, cl=xon
  }

//  if (caller->bx & 0x01) // want dtr rx flow ctrl

  if (caller->bx & 0x02) // want rts rx flow ctrl
	   set_rts_auto(Port, 1);
  else set_rts_auto(Port, 0);

  if (caller->bx & 0x10) // want cts tx flow ctrl
	   set_cts_auto(Port, 1);
  else set_cts_auto(Port, 0);


//  if (caller->bx & 0x20) // want dsr tx flow ctrl
//  if (caller->bx & 0x80) // want dcd tx flow ctrl

}

/*----------------------------------------------------------------------------
| 1F - Enable/Disable Read/Write Timeouts
  AL  xx   0 = Disable read/write timeout.
	   1 = Enable read/write timeout.
Return
|----------------------------------------------------------------------------*/
void Rckt1F(Caller far *caller)
{
// RCKTPort_t *Port = caller->Port;
// CHANNEL_T *cp;
// cp= &Port->ch;

  DEB_twidle(TW_FUNC1F);
  caller->ax = 0;
}

/*----------------------------------------------------------------------------
| F4 - Presence Test 
Return
  AX  0000 EBIOS Fosss present.  Otherwise:
  AH  FFh  EBIOS Fosss not present.
|----------------------------------------------------------------------------*/
void RcktF4(Caller far *caller)
{
  DEB_twidle(TW_FUNCF4);
  caller->ax = 0;
}

/*----------------------------------------------------------------------------
| FB - Specify Modem Leads
  AL  xx   Modem leads:
		   Bit 0 = 1: Raise DTR (drop if 0).
		   Bit 1 = 1: Raise RTS (drop if 0).
		   Bit 4 = 1: Use CTS for output flow control.
		   Bit 5 = 1: Use DSR for output flow control.
		   Bit 6 = 1: Use RI for output flow control  (not
			  supported).
		   Bit 7 = 1: Use DCD for output flow control.
Return
|----------------------------------------------------------------------------*/
void RcktFB(Caller far *caller)
{
// RCKTPort_t *Port = caller->Port;
// CHANNEL_T *cp;
// cp= &Port->ch;

  DEB_twidle(TW_FUNCFB);
  caller->ax = 0;
}

/*----------------------------------------------------------------------------
| FC - Receive with No Wait - EBIOS Foss FCh
Return
  AL  xx   Character read.
  AH  xx   Line  status (see Foss 02h - Read  Single  Character).
|----------------------------------------------------------------------------*/
void RcktFC(Caller far *caller)
{
// RCKTPort_t *Port = caller->Port;
// CHANNEL_T *cp;
// cp= &Port->ch;

  DEB_twidle(TW_FUNCFC);
  caller->ax = 0;
}

/*----------------------------------------------------------------------------
| FD - Retrieve Buffer Count - EBIOS Foss FDh
  AL  xx   Buffer to check:
		   1 = Get transmit buffer count.
		   2 = Get receive buffer count.
Return
  CX  xxxx Character count.
|----------------------------------------------------------------------------*/
void RcktFD(Caller far *caller)
{
// RCKTPort_t *Port = caller->Port;
// CHANNEL_T *cp;
// cp= &Port->ch;

  DEB_twidle(TW_FUNCFD);
  caller->ax = 0;
}

/*----------------------------------------------------------------------------
| FE - Set Pacing Mode - EBIOS Foss FEh
Return
|----------------------------------------------------------------------------*/
void RcktFE(Caller far *caller)
{
// RCKTPort_t *Port = caller->Port;
// CHANNEL_T *cp;
// cp= &Port->ch;

  DEB_twidle(TW_FUNCFE);
  caller->ax = 0;
}

/*----------------------------------------------------------------------------
| FF -Set Buffered Mode - EBIOS Foss FFh
  AL  xx   Buffer to affect:
		   1 = Transmit buffer.
		   2 = Receive buffer.
  CX  xxxx Buffer size; to terminate buffered mode, set CX to 0.
  ES:BX    Pointer to buffer.

Return
|----------------------------------------------------------------------------*/
void RcktFF(Caller far *caller)
{
// RCKTPort_t *Port = caller->Port;
// CHANNEL_T *cp;
// cp= &Port->ch;

  DEB_twidle(TW_FUNCFF);
  caller->ax = 0;
}
#endif // WIERD_BIOS_MODE

/*-----------------------------------------------*/
/*-------------- FOSSIL -------------------------*/
/*-----------------------------------------------*/

/*----------------------------------------------------------------------------
| 00 - Set communications parameters
|----------------------------------------------------------------------------*/
void Foss00(Caller far *caller)
{
  DEB_twidle(TW_FUNC00);
  Bios00(caller);
}

/*----------------------------------------------------------------------------
| 01 - Transmit character and wait
|----------------------------------------------------------------------------*/
//void Foss01(Caller far *caller)

/*----------------------------------------------------------------------------
| 02 - Get received character with wait
|----------------------------------------------------------------------------*/
//void Foss02(Caller far *caller)

/*----------------------------------------------------------------------------
| 03 - Return Serial Port Status
|----------------------------------------------------------------------------*/
//void Foss03(Caller far *caller)

/*----------------------------------------------------------------------------
| 04 - Activate Port
|----------------------------------------------------------------------------*/
void Foss04(Caller far *caller)
{
  RCKTPort_t *Port = caller->Port;
  CHANNEL_T *cp;
  WORD old_ax;

  cp= &Port->ch;
  DEB_twidle(TW_FUNC04);

  Port->Status = S_OPENED;

#ifdef SOFT_BUFS
  Port->QOutGet = 0;
  Port->QOutPut = 0;
  Port->QOutCount = 0;
#endif
  Port->baud_no_access = 0;

  RPortConfigure(Port);
  set_break(Port, 0);

  switch (default_flow_ctrl)
  {
	case 0:  // None
	break;
	case 1:  // RTS & CTS auto flow control
	  set_rts_auto(Port, 1);
	  set_cts_auto(Port, 1);
	break;
	case 2:  // XON & XOFF software flow control
	  sEnTxSoftFlowCtl(cp);
	  Port->Status |=  (S_TX_SOFT_CHK | S_RX_SOFT_CHK);
	break;
	case 3:
	  set_rts_auto(Port, 2);  // rs485
	break;
  }

  // set_dtr(Port, 1);  // turn on DTR

  RPortRxFlush(Port);
  RPortTxFlush(Port);

/*************/
	/* Enable receiver, transmitter, channel interrupts */
	sEnRxFIFO(cp);                     // enable Rx FIFO
	sSetRxTrigger(cp,TRIG_1);          
	sStartRxProcessor(cp);
	sEnTransmit(cp);                   // enable transmit
	//sGetChanIntID(cp);         // clear old interrupt conditions

	// the following interrupts enables are not really generating interrupts
	// in PERIODIC ONLY INTERRUPT MODE.
	sEnInterrupts(cp,
				 TXINT_EN   |  // tx interrupt enable
				 RXINT_EN   |  // recieve interrupt enable
				 CHANINT_EN |  // channel interrupt enable
				 MCINT_EN   |  // modem control interrupt
				 SRCINT_EN);   // special RX inter. enable
/*************/


  Port->allow_isr_access = 1;  // ISR Tx Access

  caller->ax = 0x1954;   // special return signature.
  caller->bx = 0x0522;   // 05 = rev. of fossil, 22 = max Fosss
						 // set same as X00.
}

/*----------------------------------------------------------------------------
| 05 - Deactivate Port
|----------------------------------------------------------------------------*/
void Foss05(Caller far *caller)
{
  RCKTPort_t *Port = caller->Port;
  WORD ax;
  CHANNEL_T *cp;

  cp= &Port->ch;

  //DEB_outs("FOSS5 ",1);
  DEB_twidle(TW_FUNC05);

  set_break(Port, 0);

  /**************************/
  sFlushTxFIFO(cp);
  sFlushRxFIFO(cp);
  sDisInterrupts(cp,TXINT_EN | RXINT_EN | SRCINT_EN | MCINT_EN | CHANINT_EN);
  sDisTransmit(cp);
  sClrBreak(cp);
  sDisRxStatusMode(cp);
  sDisTxSoftFlowCtl(cp);
  sDisIXANY(cp);
  sClrTxXOFF(cp);
  sDisRxFIFO(cp);
  /**************************/

  Port->allow_isr_access = 0;  // ISR Tx Access
  Port->Status = 0;
#ifdef SOFT_BUFS
  Port->QOutGet = 0;
  Port->QOutPut = 0;
  Port->QOutCount = 0;
#endif

  Port->Status &= ~S_OPENED;
  Port->Status &= ~S_FUNC1E_DONE;
}

/*----------------------------------------------------------------------------
| 06 - Raise/lower DTR
|----------------------------------------------------------------------------*/
void Foss06(Caller far *caller)
{
  RCKTPort_t *Port = caller->Port;
  CHANNEL_T *cp;

  DEB_twidle(TW_FUNC06);
  cp= &Port->ch;

  if ((caller->ax & 0xff) == 1)
  {
	Port->MCRVal |= 0x1;  // DTR
	sSetDTR(cp);
  }

  if ((caller->ax & 0xff) == 0)
  {
	Port->MCRVal &= ~0x01;  // DTR
	sClrDTR(cp);
  }
}

/*----------------------------------------------------------------------------
| 07 - Return timer tick information
 Output:  AL = Timer tick interrupt number
		  AH = Approximate interrupts per second (18 on IBM)
		  DX = Approximate number of milliseconds per tick
|----------------------------------------------------------------------------*/
void Foss07(Caller far *caller)
{
  DEB_outs("NO FUNC07 ",1);
  DEB_twidle(TW_FUNC07);
  caller->ax = caller->ax;  // get rid of compiler warning
}

/*----------------------------------------------------------------------------
| 08 - Flush output buffer
|----------------------------------------------------------------------------*/
void Foss08(Caller far *caller)
{
  RCKTPort_t *Port = caller->Port;
  CHANNEL_T *cp;

  cp= &Port->ch;

  DEB_twidle(TW_FUNC08);
#ifdef SOFT_BUFS
  while (Port->QOutCount > 0)  // wait for space
  {
	kill_time(1);

	Port->allow_isr_access = 0;  // ISR Tx Access Off
				 // notice that PutTxChars() is called at app time here
				 // as well as ISR time.  This complicates life.
	PutTxChars(Port);  //*** start sending chars from our buf to rocket immediately
	Port->allow_isr_access = 1;  // ISR Tx Access On
  }
#endif

  while (sGetTxCnt(cp) != 0)
	kill_time(1);

}

/*----------------------------------------------------------------------------
| 09 - Purge output buffer
|----------------------------------------------------------------------------*/
void Foss09(Caller far *caller)
{
  RCKTPort_t *Port = caller->Port;

  DEB_twidle(TW_FUNC09);
  RPortTxFlush(Port);
}

/*----------------------------------------------------------------------------
| 0A - Purge input buffer
|----------------------------------------------------------------------------*/
void Foss0A(Caller far *caller)
{
  RCKTPort_t *Port = caller->Port;

  DEB_twidle(TW_FUNC0A);
  RPortRxFlush(Port);
}

/*----------------------------------------------------------------------------
| 0B - Transmit no wait
|----------------------------------------------------------------------------*/
void Foss0B(Caller far *caller)
{
  RCKTPort_t *Port = caller->Port;
  BYTE CharToSend;
  CHANNEL_T *cp;
  WORD TxFree;
  WORD time;

  DEB_twidle(TW_FUNC0B);
#ifdef SOFT_BUFS

  TxFree = QOutSize - Port->QOutCount; // more data than space
  if (wait_option)
  {
	time = getdosticks();
	while (TxFree < 1)  // wait for space
	{
	  if (kill_time(time)) // timeout
		break;
	  TxFree = QOutSize - Port->QOutCount; // more data than space
	}
  }

  if (TxFree <= 0)
  {
	caller->ax = 0;
	return;
  }

  QOUT[Port->QOutPut] = (unsigned char) caller->ax;
  ++Port->QOutPut;
  if (Port->QOutPut >= QOutSize)
	Port->QOutPut = 0;

//  Port->allow_isr_access = 0;  // ISR Tx Access
  sDisable();
  ++Port->QOutCount;
  sEnable();
//  Port->allow_isr_access = 1;  // ISR Tx Access

#else
  cp= &Port->ch;
  if (sGetTxCnt(cp) >= TXFIFO_SIZE)
  {
	caller->ax = 0;  // char not sent
	return;
  }

  CharToSend= (BYTE)caller->ax;

  sOutB(cp->TxRxData, CharToSend);
#endif

  caller->ax = 1;  // character was sent
}

/*----------------------------------------------------------------------------
| 0C - Non-destructive read-ahead (Peek) of comm buffer.
|----------------------------------------------------------------------------*/
void Foss0C(Caller far *caller)
{
  RCKTPort_t *Port = caller->Port;
  CHANNEL_T *cp;
  WORD RxFIFOOut;
  BYTE RxByte;

  cp= &Port->ch;

  DEB_twidle(TW_FUNC0C);

  if (sGetRxCnt(cp) == 0)    // no chars waiting, so return 0xffff
  {
	caller->ax = 0xffff;
  }
  else
  {  // dork directly with the ram on a port to peek at
	// the top of the RxFIFO

	RxFIFOOut = sChanInWI(cp, _RXF_OUTP) * 2;
	RxByte = sChanInBI(cp, _RX_FIFO + RxFIFOOut);

	caller->ax = RxByte;
  }
}

/*----------------------------------------------------------------------------
| 0D - Keyboard read without wait
  This  Foss  returns  the  keyboard  scan  code  of  the  next
  character from  the keyboard.   This is a  non-destructive (peek)
  read of  the keyboard.  If no keyboard data is available,  0FFFFh
  is returned in AX.

			   Input:    AH = 0Dh
			   Output:   AX = IBM-style scan code if available
							= 0FFFFh if no character not available
|----------------------------------------------------------------------------*/
void Foss0D(Caller far *caller)
{
 WORD RegAX;

  DEB_twidle(TW_FUNC0D);
  ASM {
//    mov ah,11H  // 11H - bios get keyboard status
	mov ah,1H   // 1H - bios get keyboard status
	int 16H     // returns AL=ascii code, AH=scan code
//    MOV AH,00 // why this here???
	jnz  chr_aval
	mov RegAX, 0ffffH  // no char ready
	jmp done0d
chr_aval:
	mov RegAX, ax
done0d:
  }
  caller->ax = RegAX;
}

/*----------------------------------------------------------------------------
| 0E - Keyboard read with wait

  This  Foss  return  the  next character  available  from  the
  keyboard.  If no keyboard input is available, this Foss waits
  until  a keyboard  character  is available.   Returned  character
  codes are the same as Foss 0Dh (IBM style scan codes). 
|----------------------------------------------------------------------------*/
void Foss0E(Caller far *caller)
{
  WORD RegAX;
  DEB_twidle(TW_FUNC0E);
  ASM {
	mov ah,0H  // 0H - bios get keyboard status, with wait
//    mov ah,10H  // 10H - bios get keyboard status, with wait
	int 16H     // returns AL=ascii code, AH=scan code
	mov RegAX, ax
  }
  caller->ax = RegAX;
}

/*----------------------------------------------------------------------------
| 0F - Flow Control for serial I/O
|----------------------------------------------------------------------------*/
void Foss0F(Caller far *caller)
{
  RCKTPort_t *Port = caller->Port;
  BYTE FlowCtrl;
  CHANNEL_T *cp;
#define TX_SW_FLOW 1
#define RX_SW_FLOW 8
#define RTSCTS_FLOW 2

  DEB_twidle(TW_FUNC0F);
  FlowCtrl= caller->ax & (RX_SW_FLOW | TX_SW_FLOW | RTSCTS_FLOW);

  cp= &Port->ch;

  set_break(Port, 0);
  // set_dtr(Port, 1);  // turn on DTR

  if (FlowCtrl & RTSCTS_FLOW)
  {
	set_rts_auto(Port, 1);  // turn on auto rts/cts flow control
	set_cts_auto(Port, 1);  // turn on auto rts/cts flow control
  }
  else
  {
	set_rts_auto(Port, 0);  // turn off auto rts/cts flow control
	set_cts_auto(Port, 0);  // turn off auto rts/cts flow control
  }

  if (FlowCtrl & TX_SW_FLOW)
  {
	Port->Status |=  S_TX_SOFT_CHK;
	sEnTxSoftFlowCtl(cp);
  }
  else
  {
	Port->Status &= ~S_TX_SOFT_CHK;
	sDisTxSoftFlowCtl(cp);
  }

  if (FlowCtrl & RX_SW_FLOW)
	   Port->Status |=  S_RX_SOFT_CHK;
  else Port->Status &= ~S_RX_SOFT_CHK;
}

/*----------------------------------------------------------------------------
| 10 - Control-C / Control-K checking
|----------------------------------------------------------------------------*/
void Foss10(Caller far *caller)
{
  RCKTPort_t *Port = caller->Port;
  CHANNEL_T *cp;

  cp= &Port->ch;

  DEB_twidle(TW_FUNC10);
  if (caller->ax & 2)      // changing transmit state ?
  {
	DEB_outs("TX DIS ",0);
	sDisTransmit(cp);
  }
  else
  {
	DEB_outs("TX ENA ",0);
	sEnTransmit(cp);
  }

  if (caller->ax & 1)      // how about CTLC/CTLK checking ?
  {
	if ((Port->Status & S_CTL_CK) == 0)
	{
	  DEB_outs("CK-USED ",0);
	  Port->Status |= S_CTL_CK;
	  Port->Status &= ~S_CTL_CK_GOT;
	}

	if (Port->Status & S_CTL_CK_GOT)
	{
	  DEB_outs("CTL CK ",0);
	  caller->ax = 1;
	}
	else caller->ax = 0;
  }
  else
  {
	Port->Status &= ~S_CTL_CK;
  }

  Port->Status &= ~S_CTL_CK_GOT;
}

/*----------------------------------------------------------------------------
| 11 - Set current cursor location
|----------------------------------------------------------------------------*/
void Foss11(Caller far *caller)
{
 WORD RegDX = caller->dx;

  DEB_twidle(TW_FUNC11);
  ASM {
	mov ah,2H  // 2H - bios set cursor loc
	mov bh,0
	mov dx,RegDX
	int 10H    // bh=dis_page, dh=row, dl=col
  }
}

/*----------------------------------------------------------------------------
| 12 - Read current cursor location
|----------------------------------------------------------------------------*/
void Foss12(Caller far *caller)
{
 WORD RegDX;

  DEB_twidle(TW_FUNC12);
  ASM {
	xor bx,bx
	mov ah,3H  // 3H - bios get cursor loc
	int 10H    // bh=dis_page, dh=row, dl=col
	mov RegDX,dx
  }
  caller->dx = RegDX;
}

/*----------------------------------------------------------------------------
| 13 - Single character ANSI write to screen
|----------------------------------------------------------------------------*/
void Foss13(Caller far *caller)
{
 WORD RegAX = caller->ax;

  DEB_twidle(TW_FUNC13);
  ASM {
	mov ax,RegAX
	mov ah,02H    // 02H - write char to screen, dos
	mov dl,al     // char to screen in dl
	int 21H       // dos
  }
}

/*----------------------------------------------------------------------------
| 14 - Enable or disable the DCD watchdog
|----------------------------------------------------------------------------*/
void Foss14(Caller far *caller)
{
  RCKTPort_t *Port = caller->Port;
  DEB_twidle(TW_FUNC14);
  if (caller->ax & 1)
  {
	Port->Status |= S_WATCHDOG;
  }
  else Port->Status &= ~S_WATCHDOG;
}

/*----------------------------------------------------------------------------
| 15 - Write character to screen using BIOS
|----------------------------------------------------------------------------*/
void Foss15(Caller far *caller)
{
 WORD RegAX = caller->ax;

  DEB_twidle(TW_FUNC15);
  ASM {
	mov ax,RegAX
	mov ah,0EH    // 0EH - write char to screen, bios tty mode
	mov bx,7H  // color(bl) & display page(bh)
	int 10H    // bh=dis_page, cx=cnt of chars, al=char
  }
}

/*----------------------------------------------------------------------------
| 16 - Add  or delete a routine from  the timer
  Input:    AH = 16h
			AL = 01h - Add a Foss
			   = 00h - Delete a Foss
			ES = Segment of Foss
			DX = Offset of Foss

  Output:   AX =  0000h or  ticks per  second if  operation is
				 successful
			   = 0FFFFh - Operation unsuccessful
|----------------------------------------------------------------------------*/
void Foss16(Caller far *caller)
{
 char far *str;
 int i;

  // overload this function to include our command line passing.
  if ((caller->ax == 0x16ff) && (caller->bx == 0x5555))
  {
	str = MK_FP(caller->si, caller->cx);

	i=0;
	while (*str)   // copy to local string
	  cmd_line[i++] = *str++;
	cmd_line[i] = 0;  // null terminate

	do_cmd_line(cmd_line,1);
	
	caller->ax = 0;
	return;
  }

  if ((caller->ax & 0xff) == 0)  // insert timer hook es:dx
  {
	DEB_outs("HOOK ",1);
	AppHook = (((unsigned long)(caller->es)) << 16) | ((unsigned long) caller->dx);
	caller->ax = 0;
	return;
  }
  else if ((caller->ax & 0xff) == 1)  // remove timer hook es:dx
  {
	DEB_outs("UNHOOK ",1);
	AppHook = 0L;
	caller->ax = 0;
	return;
  }
  
  DEB_twidle(TW_FUNC16);
}

/*----------------------------------------------------------------------------
| 17 - Reboot system
|----------------------------------------------------------------------------*/
void Foss17(Caller far *caller)
{
  DEB_twidle(TW_FUNC17);
  caller->ax = caller->ax;  // get rid of compiler warning
  Reboot();
}

/*----------------------------------------------------------------------------
| 18 - Block Read
|----------------------------------------------------------------------------*/
// void Foss18(Caller far *caller)

/*----------------------------------------------------------------------------
| 19 - Block Write
|----------------------------------------------------------------------------*/
// void Foss19(Caller far *caller)

#ifdef SOFT_BUFS
/*---------------------------------------------------------------------------
| PutTxChars -  Write from Software Queue to Rocket Port Hardware Tx FIFO.
|  This routine is called at Interrupt Time and also at application time
|  inorder to start transmition immediately.  This requires some flags in
|  the software to limit access(contention handling).
|--------------------------------------------------------------------------*/
void PutTxChars(RCKTPort_t *Port)
{
  CHANNEL_T *cp;

 int Send1,Send2;               // data counts
 int ToSend, room;
 unsigned char *qout;
 WORD io_addr;

  cp= &Port->ch;

  ToSend =  Port->QOutCount;

  room = TXFIFO_SIZE - sGetTxCnt(cp);

  if (room > 0)                        // space in Tx FIFO
  {
	if (ToSend > room)                     // don't move more than available
	  ToSend = room;
											   // two parts incase queue wraps
	Send1 = QOutSize - Port->QOutGet; // num bytes to end of buffer

	if (Send1 < ToSend)            // Q wraps
	  Send2 = ToSend - Send1;
	else                             // no Q wrap
	{
	  Send2 = 0;
	  Send1 = ToSend;
	}
  
	qout = (unsigned char *) &QOUT[Port->QOutGet];
	io_addr = cp->TxRxData;
	if (Send1)
	{
	  ASM {
//        push  si --- do we need to save this????? no checked asm listing
		cld               // direction up
		mov   dx,io_addr // I/O address to write to
		mov   cx,Send1   // number bytes to write
		mov   si,qout    // mem address to write from
		test  si,1       // odd boundary?
		jz    isrtx_2     //   no - do double word xfer
		outsb             //   yes - write one byte
		dec   cx         // decr count
isrtx_2:                             // dword xfer on even boundaries
		push  cx       // save count
		shr   cx,1     // number words to write
		rep   outsw     // write words to end of buf
		pop   cx       // get Cnt1
		test  cx,1     // odd number bytes?
		jz    isrtx_3   //    no - done with xfer
		outsb           //    yes - write the last byte
isrtx_3:
//        pop   si
	  } // end ASM
	}
									 // and then send the second chunk
									 // as words and any remaining odd byte
	qout = (unsigned char *) &QOUT[0];
	if (Send2)
	{
	  ASM {
//        push  si --- do we need to save this????? no checked asm listing.
		cld               // direction up
		mov   dx,io_addr // I/O address to write to
		mov   cx,Send2   // number bytes to write
		mov   si,qout    // mem address to write from
		test  si,1       // odd boundary?
		jz    isrtx_4     //   no - do double word xfer
		outsb             //   yes - write one byte
		dec   cx         // decr count
isrtx_4:                             // dword xfer on even boundaries
		push  cx       // save count
		shr   cx,1     // number words to write
		rep   outsw     // write words to end of buf
		pop   cx       // get Cnt1
		test  cx,1     // odd number bytes?
		jz    isrtx_5   //    no - done with xfer
		outsb           //    yes - write the last byte
isrtx_5:
//        pop   si
	  } // end ASM
	}
							   // update the queue Get pointer
							   // and the total count remaining in the queue
	Port->QOutGet += ToSend;
	if (Port->QOutGet >= QOutSize)
	  Port->QOutGet -= QOutSize;
	Port->QOutCount -= ToSend;
  }  // data to send and room in hardware fifo
}
#endif

/*----------------------------------------------------------------------------
| 1A - Break begin or end
|----------------------------------------------------------------------------*/
void Foss1A(Caller far *caller)
{
  RCKTPort_t *Port = caller->Port;

  DEB_twidle(TW_FUNC1A);

  if (caller->ax & 1)        // send break ?
	   set_break(Port, 1);
  else set_break(Port, 0);
}

/*----------------------------------------------------------------------------
| 1B - Return information about
|----------------------------------------------------------------------------*/
void Foss1B(Caller far *caller)
{
  RCKTPort_t *Port = caller->Port;
  WORD RegES = caller->es;
  WORD RegDI = caller->di;
  WORD our_cs;
  WORD RegSI;

  static struct {
	WORD  STRSIZ;
	BYTE  MAJVER;
	BYTE  MINVER;
	char far *IDENT;
	WORD  IBUFR;
	WORD  IFREE;
	WORD  OBUFR;
	WORD  OFREE;
	BYTE  SWIDTH;
	BYTE  SHEIGHT;
	BYTE  BAUD;
  } INFO;

  CHANNEL_T *cp;
  int ToMove;

  cp= &Port->ch;

  DEB_twidle(TW_FUNC1B);

  sDisable();  // lock out others due to shared static INFO structure

  INFO.STRSIZ= sizeof(INFO);
  INFO.MAJVER= 5;   // want fossil spec version as per x00
  INFO.MINVER= 0;
  ASM {mov our_cs, cs}
  //INFO.IDENT= MK_FP(FP_SEG(driver_str), FP_OFF(driver_str));
  INFO.IDENT= MK_FP(our_cs, FP_OFF(driver_str));

  INFO.IBUFR= RXFIFO_SIZE;
#ifdef SOFT_BUFS
  INFO.OBUFR = QOutSize;
  INFO.OFREE = (QOutSize - Port->QOutCount);
#else
  INFO.OBUFR= TXFIFO_SIZE;
  INFO.OFREE = TXFIFO_SIZE-sGetTxCnt(cp);
#endif
  INFO.IFREE= RXFIFO_SIZE-sGetRxCnt(cp);
  INFO.BAUD = Port->bitRate00;

  INFO.SWIDTH = 80;
  INFO.SHEIGHT = 25;

  ToMove= sizeof(INFO);
  if (ToMove > caller->cx)
	ToMove= caller->cx;

   RegSI = (unsigned)&INFO;
   ASM {
	 push si
	 mov si, RegSI
	 mov es,RegES
	 mov di,RegDI
	 mov cx,ToMove
	 rep movsb
	 pop si
   }
  sEnable();

  caller->ax= ToMove;
}

/*----------------------------------------------------------------------------
| 1E - Extended line control initialization.
	Same as PS/2 Bios Int14 func04
|----------------------------------------------------------------------------*/
void Foss1E(Caller far *caller)
{
  DEB_twidle(TW_FUNC1E);
  Bios04(caller);
}

/*----------------------------------------------------------------------------
| 1F - Extended serial port status/control
|----------------------------------------------------------------------------*/
void Foss1F(Caller far *caller)
{
  DEB_twidle(TW_FUNC1F);
  Bios05(caller);

#ifdef COMMENT_OUT
  RCKTPort_t *Port = caller->Port;
  CHANNEL_T *cp;
  BYTE RegAL;

  cp= &Port->ch;


  RegAL= caller->ax & 0xff;

  if (RegAL == 0)    // subFoss 0, Read MCR
  {
	caller->bx = Port->MCRVal | 0x8;
  }

//                         BL = Modem control register, where:
//                              Bits 7-5 = 0 (Reserved)
//                              Bit 4    = 1 Loopback mode
//                              Bit 3    = 1 OUT2 (interrupts) enabled
//                              Bit 2    = 1 OUT1 active
//                              Bit 1    = 1 Request to send active
//                              Bit 0    = 1 Data terminal ready (DTR) active


  if (RegAL == 1)    // subFoss 1, Write MCR
  {
	Port->MCRVal= caller->bx & 0xff;
	if (Port->MCRVal & 0x10)  // loopback
		 {sEnLocalLoopback(cp);}
	else {sDisLocalLoopback(cp);}

	if (Port->MCRVal & 0x1)  // DTR
		 set_dtr(Port, 1);
	else set_dtr(Port, 0);

	if (Port->MCRVal & 0x2)  // RTS
		 set_rts(Port, 1);
	else set_rts(Port, 0);
  }

  caller->ax = Bios03Status(caller);
#endif

}

/*----------------------------------------------------------------------------
| 20 - Read with no wait (destructive)
|----------------------------------------------------------------------------*/
void Foss20(Caller far *caller)
{
  RCKTPort_t *Port = caller->Port;
  BYTE CharToRecv;
  CHANNEL_T *cp;

  DEB_twidle(TW_FUNC20);
  cp= &Port->ch;

  if (sGetRxCnt(cp) == 0)
	caller->ax= 0xffff;
  else
  {
	CharToRecv= sInB(cp->TxRxData);
	caller->ax= CharToRecv;
  }
}

/*----------------------------------------------------------------------------
| 21 - Stuff/Poke the receive buffer
|----------------------------------------------------------------------------*/
void Foss21(Caller far *caller)
{
//  RCKTPort_t *Port = caller->Port;
//  BYTE CharToRecv;
//  CHANNEL_T *cp;
//  cp= &Port->ch;

  DEB_outs("NO FUNC21 ",1);
  DEB_twidle(TW_FUNC21);
  caller->ax = caller->ax;  // get rid of compiler warning
}

/*----------------------------------------------------------------------------
| RCKTPollISR - interrupt routine.
|----------------------------------------------------------------------------*/
void RCKTInt(void)
{

#ifdef EVENT_PROCESS
 static CHANNEL_T *cp;
 static BYTE int_id;
 static RCKTPort_t *Port;
 static WORD ChanStatus;

  if (events_option)
  {
//  for (pn=0; pn<nPorts; pn++)
	Port= &RcktPort[0];  // asume runnning on 1 port only

	cp= &Port->ch;
//    ChanStatus= sInW(cp->ChanStat);

#ifdef COMMENT_OUT
	if (ctlp->BusType == BUS_PCI)
		 { int_id = sPCIGetControllerIntStatus(ctlp); }
	else { int_id = sGetControllerIntStatus(ctlp); }
	if ((int_id & 1) == 0)
	  debv_outs("NoCID ", 1);

	int_id = sGetAiopIntStatus(ctlp,0);  // grab for first aiop only
	if ((int_id & 1) == 0)
	  debv_outs("NoAio ", 1);

#endif

	int_id = sGetChanIntID(cp);
	ChanStatus = sInW(cp->ChanStat);

//    if (int_id & SRC_INT)
//    {
//      debv_outs("SRC_INT!", 1);
//      if ((sInW(cp->ChanStat) & RXFRAME) == 0)
//        debv_outs("No Fr-Stat!", 1);
//    }
	if (int_id & TXFIFO_MT)
	  debv_outs("TX_INT ", 1);
	if (int_id & SRC_INT)
	  debv_outs("SRC_INT ", 1);
	if (int_id & DELTA_CD)
	  debv_outs("CD_INT ", 1);
	if (int_id & DELTA_CTS)
	  debv_outs("CTS_INT ", 1);
	if (int_id & DELTA_DSR)
	  debv_outs("DSR_INT ", 1);
	if (int_id == 0)
	  debv_outs("???_INT ", 1);

	if (ChanStatus & RXFRAME)
	  debv_outs("FRAMING ", 1);
	if (ChanStatus & RXFOVERFL)
	  debv_outs("RXOVERF ", 1);
	if (ChanStatus & RX2MATCH)
	  debv_outs("RX2MAT ", 1);
	if (ChanStatus & RX1MATCH)
	  debv_outs("RX1MAT ", 1);
	if (ChanStatus & RXBREAK)
	  debv_outs("RXBRK ", 1);
	if (ChanStatus & RXPARITY)
	  debv_outs("RXPAR ", 1);
  }
#endif

  poll_rocket();

  sDisable();  // IRET will turn interrupts back on
		  // EOI to PICs
  if (irq > 7)      // slave 8259 ??
	sOutB(0xa0,0x20);
  sOutB(0x20,0x20);    // master 8259

  if (ctlp->BusType == BUS_PCI)
	   sPCIControllerEOI(ctlp);    // EOI to the board
  else    sControllerEOI(ctlp);    // EOI to the board
}

/*----------------------------------------------------------------------------
| RCKTInt8 - Timer Tick Routine.
|----------------------------------------------------------------------------*/
void RCKTInt8()
{
  poll_rocket();
}

/*----------------------------------------------------------------------------
| poll_rocket - periodic service routine.
|----------------------------------------------------------------------------*/
void poll_rocket(void)
{
  int pn, RxCnt;
  RCKTPort_t *Port;
  CHANNEL_T *cp;
  WORD ChanStatus;

  ++poll_ticks;
  for (pn=0; pn<nPorts; pn++)
  {
	Port= &RcktPort[pn];

	if (!Port->allow_isr_access)
	  continue;

#ifdef WIERD_BIOS_MODE
	if (driver_mode & BIOS_MODE)
	{
	  if (sGetRxCnt(cp) == 0)    // no chars waiting, so return 0xffff
		   Port->chars_ready_flag = 0;
	  else Port->chars_ready_flag = 1;
	}
#endif

#ifdef SOFT_BUFS
	if (Port->QOutCount > 0)
		PutTxChars(Port);
#endif

//    if (Port->Status == 0)
//      continue;

	if (Port->Status & (S_WATCHDOG | S_RX_SOFT_CHK | S_CTL_CK))
	{
	  cp= &Port->ch;
	  ChanStatus= sInW(cp->ChanStat);

	  if (Port->Status & S_RX_SOFT_CHK)
	  {
		RxCnt= sGetRxCnt(cp);
		if (RxCnt > ((RXFIFO_SIZE / 10) * 8) &&
			((Port->Status & S_RX_SOFT_OFF) == 0))
		{
		  sWriteTxPrioByte(cp, ASCII_XOFF);
		  Port->Status |= S_RX_SOFT_OFF;
		  DEB_outs("SndXOFF",1);
		}

		if (RxCnt < ((RXFIFO_SIZE / 10) * 2) &&
			(Port->Status & S_RX_SOFT_OFF))
		{
		  sWriteTxPrioByte(cp, ASCII_XON);
		  Port->Status &= ~S_RX_SOFT_OFF;
		  DEB_outs("SndXON",1);
		}
	  }

	  if (Port->Status & S_CTL_CK)
	  {
		if (ChanStatus & (RX1MATCH|RX2MATCH))
		  Port->Status |= S_CTL_CK_GOT;
	  }

	  if (Port->Status & S_WATCHDOG)
	  {
		if (ChanStatus & CD_ACT)
		  Port->Status |= S_CD_ACTIVE;
		else
		  if (Port->Status & S_CD_ACTIVE)
			Reboot();
	  }
	}
  }
}

/*----------------------------------------------------------------------------
| RCKTInt14 - Process FOSSIL interrupt 14 calls.
|----------------------------------------------------------------------------*/
void RCKTInt14(Caller far *caller)
{
  caller->Port = &RcktPort[caller->dx - comFirst];

#ifdef WIERD_BIOS_MODE
  if (driver_mode & BIOS_MODE)
  {
	switch(caller->ax >> 8)
	{
	  case 0x00: Rckt00(caller); break;
	  case 0x01: Rckt01(caller); break;
	  case 0x02: Rckt02(caller); break;
	  case 0x03: Rckt03(caller); break;
	  case 0x04: Rckt04(caller); break;
	  case 0x05: Rckt05(caller); break;
	  case 0x06: Rckt06(caller); break;
	  case 0x07: Rckt07(caller); break;
	  case 0x08: Rckt08(caller); break;
	  case 0x09: Rckt09(caller); break;
	  case 0x0a: Rckt0A(caller); break;
	  case 0x0b: Rckt0B(caller); break;
	  case 0x0c: Rckt0C(caller); break;
	  case 0x0d: Rckt0D(caller); break;
	  case 0x0e: Rckt0E(caller); break;
	  case 0x0f: Rckt0F(caller); break;
	  case 0x10: Rckt10(caller); break;
	  case 0x11: Rckt11(caller); break;
	  case 0x12: Rckt12(caller); break;
	  case 0x13: Rckt13(caller); break;
	  case 0x14: Rckt14(caller); break;
	  case 0x15: Rckt15(caller); break;
	  case 0x16: Rckt16(caller); break;
	  case 0x17: Rckt17(caller); break;
	  case 0x18: Rckt18(caller); break;
	  case 0x19: Rckt19(caller); break;
	  case 0x1a: Rckt1A(caller); break;
	  case 0x1b: Rckt1B(caller); break;
	  case 0x1c: Rckt04(caller); break;
	  case 0x1d: Rckt05(caller); break;
	  case 0x1e: Rckt1E(caller); break;
	  case 0x1f: Rckt1F(caller); break;
	}

	if (sGetRxCnt(&caller->Port->ch) == 0)  // no chars waiting, so return 0xffff
	   caller->Port->chars_ready_flag = 0;
	else caller->Port->chars_ready_flag = 1;
  }
  else  // fossil
#endif // WIERD_BIOS_MODE
  {
#ifdef COMMENT_OUT
	if (debug_level & 4)
	{
	  if (((caller->ax >> 8) == 0x11) && ((caller->ax >> 8) == 0x12)) // status
	  {
		hextoa(&tmpstr[4], caller->ax >> 8, 2);
		tmpstr[3] = 'F';
		tmpstr[6] = 0;
		DEBV_outs(&tmpstr[3],1);
	  }
	}
#endif
	switch(caller->ax >> 8)
	{
	  case 0x00: DEB_twidle(TW_FUNC00); Bios00(caller); break;
	  case 0x01: DEB_twidle(TW_FUNC01); Bios01(caller); break;
	  case 0x02: DEB_twidle(TW_FUNC02); Bios02(caller); break;
	  case 0x03: DEB_twidle(TW_FUNC03); caller->ax = Bios03Status(caller);
	  break;
	  case 0x04: DEB_twidle(TW_FUNC04);
		if (driver_mode & BIOS_MODE)
		  Bios04(caller);
		else
		  Foss04(caller);
	  break;
	  case 0x05: 
		DEB_twidle(TW_FUNC05);
		if (driver_mode & BIOS_MODE)
			 Bios05(caller);
		else Foss05(caller);
	  break;
	  case 0x06: Foss06(caller); break;
	  case 0x07: Foss07(caller); break;  // timer tick, not done
	  case 0x08: Foss08(caller); break;
	  case 0x09: Foss09(caller); break;
	  case 0x0a: Foss0A(caller); break;
	  case 0x0b: Foss0B(caller); break;
	  case 0x0c: Foss0C(caller); break;
	  case 0x0d: Foss0D(caller); break;
	  case 0x0e: Foss0E(caller); break;
	  case 0x0f: Foss0F(caller); break;
	  case 0x10: Foss10(caller); break;
	  case 0x11: Foss11(caller); break;
	  case 0x12: Foss12(caller); break;
	  case 0x13: Foss13(caller); break;
	  case 0x14: Foss14(caller); break;
	  case 0x15: Foss15(caller); break;
	  case 0x16: Foss16(caller); break;  // timer tick, not done
	  case 0x17: Foss17(caller); break;
	  case 0x18: DEB_twidle(TW_FUNC18); rp_block_read(caller); break;
	  case 0x19: DEB_twidle(TW_FUNC19); rp_block_write(caller); break;
	  case 0x1a: Foss1A(caller); break;
	  case 0x1b: Foss1B(caller); break;
	  case 0x1c: Foss04(caller); break;
	  case 0x1d: Foss05(caller); break;
	  case 0x1e: Foss1E(caller); break;
	  case 0x1f: Foss1F(caller); break;
	  case 0x20: Foss20(caller); break;
	  case 0x21: Foss21(caller); break;
	}
  }
}

/*----------------------------------------------------------------------------
| itoa -
|----------------------------------------------------------------------------*/
static void intoa(char *str, WORD v, int places)
{
  while (places > 0)
  {
	--places;
	str[places] = (v % 10) + '0';
	v /= 10;
  }
}

/*----------------------------------------------------------------------------
| hextoa -
|----------------------------------------------------------------------------*/
static void hextoa(char *str, WORD v, int places)
{
  while (places > 0)
  {
	--places;
	if ((v & 0xf) < 9)
		 str[places] = '0' + (v & 0xf);
	else str[places] = 'A' + (v & 0xf) - 0xa;
	v >>= 4;
  }
}

/*----------------------------------------------------------------------------
| debv_outs -
|----------------------------------------------------------------------------*/
static void debv_outs(char *str, int level)
{
#define AL_MSG_ATTR 0x47
#define MSG_ATTR 0x37
static int msg_y = 4;
static int msg_x = 0;

  if (level == 0)
	   putsxy(str,  msg_x, msg_y, MSG_ATTR);
  else putsxy(str,  msg_x, msg_y, AL_MSG_ATTR);
  ++msg_y;
  if (msg_y > 22)
  {
	msg_y = 4;
	msg_x += 10;
	if (msg_x > 70) msg_x = 0;
  }
}

/*----------------------------------------------------------------------------
| deb_outs -
|----------------------------------------------------------------------------*/
static void deb_outs(char *str, int level)
{
#define AL_MSG_ATTR 0x47
#define MSG_ATTR 0x37
static int msg_y = 10;

  if (level == 0)
	   putsxy(str,  0, msg_y, MSG_ATTR);
  else putsxy(str,  0, msg_y, AL_MSG_ATTR);
  ++msg_y;
  if (msg_y > 20)
	msg_y = 10;
}

/*----------------------------------------------------------------------------
| twidle - debug.
|----------------------------------------------------------------------------*/
static void twidle(int func)
{
#define RUL_Y 0
#define NORM_ATTR 0x70
#define HI_ATTR 0x73
 WORD old_xy;
//static char smstr[3];
//static char dstr[0x33];

static char *smstr = {"123"};
static char *dstr = {"01234567890abcdef01234567890abcdef01234567890abcdef01234"};
static char *ruler = {"0123456789ABCDEF "};
static int init_flag = 0;
 int i,c;
 int place;

  if (!init_flag)
  {
	init_flag = 1;
	for (i=0; i<0x30; i++)
	{
	  c = '.';
	  dstr[i] = c;
	}
	dstr[0x31] = 0;
  }
  dstr[0x31] = 0;

  place = func;     // adjust for the space between 16H spaces
  if (place >= 0x10)
  {
	if (place >= 0x20)
	  ++place;
	++place;
  }
  putsxy(ruler,    0, RUL_Y, NORM_ATTR);
  putsxy(ruler, 0x11, RUL_Y, NORM_ATTR);
  putsxy(ruler, 0x22, RUL_Y, NORM_ATTR);

  dstr[place] += 1;

  if (dstr[place] < '1') dstr[place] = '1';
  if (dstr[place] > '9') dstr[place] = '1';
  putsxy(dstr, 0, RUL_Y+1, NORM_ATTR);
  smstr[0] = dstr[place];
  smstr[1] = 0;
  putsxy(smstr, place, RUL_Y+1, HI_ATTR);

  smstr[0] = '0' + (poll_ticks & 0xf);
  putsxy(smstr, TW_POLL+2, RUL_Y+1, NORM_ATTR);
}

/*----------------------------------------------------------------------------
| putsxy -
|----------------------------------------------------------------------------*/
static void putsxy(char *str, int x, int y, char attr)
{
 unsigned char far *scr;
 int offset = (160 * y) + (2 * x);
 int j;

  scr = (char far *) 0xB8000000L;

  scr += offset;
  while (*str)
  {
	*scr++ = *str++;
	*scr++ = attr;
  }
}

/*-----------------------------------------------------------------
 upcase - to upcase case
|------------------------------------------------------------------*/
static int upcase(int c)
{
  if ((c >= 'a') && (c <= 'z'))
	return ((c-'a') + 'A');
  else return c;
}

/*------------------------------------------------------------------------
| getlint -
|-----------------------------------------------------------------------*/
long getlint(char *textptr, int *countptr)
{
  long number;
  int tempcount;
  int negate = 0;

  tempcount = 0;
  number = 0;
  while (*textptr == 0x20)
  {
	++textptr;
	++tempcount;
  }

  if (*textptr == '-')
  {
	++textptr;
	++tempcount;
	negate = 1;
  }

  if ( ((*textptr >= 0x30) && (*textptr <= 0x39)) )
   {
	 while ( ((*textptr >= 0x30) && (*textptr <= 0x39)) )
	 {
	   number = (number * 10) + ( *textptr & 0x0f);
	   ++textptr;
	   ++tempcount;
	 }
   }
  else
   {
	 if (countptr != NULL)
	   *countptr = 0;
	 return 0;
   }   
  if (countptr != NULL)
	*countptr = tempcount;
  if (negate)
	return (-number);
  return number;
} /* getlint */

/*------------------------------------------------------------------------
| getint -
|-----------------------------------------------------------------------*/
int getint(char *textptr, int *countptr)
{
  int number;
  int tempcount;
  int negate = 0;

  tempcount = 0;
  number = 0;
  while (*textptr == 0x20)
  {
	++textptr;
	++tempcount;
  }

  if (*textptr == '-')
  {
	++textptr;
	++tempcount;
	negate = 1;
  }

  if ( ((*textptr >= 0x30) && (*textptr <= 0x39)) )
   {
	 while ( ((*textptr >= 0x30) && (*textptr <= 0x39)) )
	 {
	   number = (number * 10) + ( *textptr & 0x0f);
	   ++textptr;
	   ++tempcount;
	 }
   }
  else
  {
	 if (countptr != NULL)
	   *countptr = 0;
	 return 0;
  }   
  if (countptr != NULL)
	*countptr = tempcount;

  if (negate)
	return (-number);
  return number;
} /* getint */

/*------------------------------------------------------------------------
| gethint - for finding hex words.
|-----------------------------------------------------------------------*/
unsigned int gethint(char *bufptr, int *countptr)
{
  unsigned int count;
  unsigned char temphex;
  unsigned int number;

  number = 0;
  count = 0;

  while (*bufptr == 0x20)
	 { ++bufptr;
	   ++count;
	 }

  while ( ((*bufptr >= 0x30) && (*bufptr <= 0x39))
								  ||
		  ((toupper(*bufptr) >= 0x41) && (toupper(*bufptr) <= 0x46)) )
  {
	if (*bufptr > 0x39) temphex = (toupper(*bufptr) & 0x0f) + 9;
	else temphex = *bufptr & 0x0f;
	number = (number * 16) + temphex;
	++bufptr;
	++count;
  }

  if (countptr != NULL)
	*countptr = count;
  return number;
} /* gethint */

/*----------------------------------------------------------------------------
| get_vect -
|----------------------------------------------------------------------------*/
DWORD get_vect(int vector)
{
 DWORD far *ptr;
  ptr = (DWORD far *)(vector << 2);
  return *ptr;
}

/*----------------------------------------------------------------------------
| set_vect -
|----------------------------------------------------------------------------*/
void set_vect(int vector, DWORD proc)
{
 DWORD far *ptr;
 //WORD off, seg;
  //off = FP_OFF(proc);
  //seg = FP_SEG(proc);

  vector = (vector * 4);
  //Dprintf("setvect %x RCKTIsr14[%x:%x]\n", vector, seg, off);

  ptr = (DWORD far *) vector;
  *ptr = proc;
}


/*-----------------------------------------------------------------
 toupper - to my_toupper case
|------------------------------------------------------------------*/
int toupper(int c)
{
  if ((c >= 'a') && (c <= 'z'))
	return ((c-'a') + 'A');
  else return c;
}

/*----------------------------------------------------------------------------
| memcpy -
|----------------------------------------------------------------------------*/
void memcpy(void far *dest, void far *src, unsigned int size)
{
  unsigned int i;

  for (i=0; i<size; i++)
	*((unsigned char far *)(dest))++ = *((unsigned char far *)(src))++;
}

/*----------------------------------------------------------------------------
| memset -
|----------------------------------------------------------------------------*/
void memset(void *ptr, unsigned char v, int size)
{
  int i;
  unsigned char *str = ptr;

  for (i=0; i<size; i++)
	*str++ = v;
}

/*----------------------------------------------------------------------------
| strlen -
|----------------------------------------------------------------------------*/
int strlen(char *str)
{
  int i = 0;

  while (*str++ != 0)
	++i;
  return i;
}


/*----------------------------------------------------------------------------
| puts -
|----------------------------------------------------------------------------*/
void puts(char *str)
{
 int ln = strlen(str);
 int i;
 char c;

 for (i=0; i<ln; i++)
 {
   c = *str++;
   ASM
   {
	 mov al,c
	 mov ah,02H    // 02H - write char to screen, dos
	 mov dl,al     // char to screen in dl
	 int 21H       // dos
   }
 }
   c = 0x0d;
   ASM
   {
	 mov al,c
	 mov ah,02H    // 02H - write char to screen, dos
	 mov dl,al     // char to screen in dl
	 int 21H       // dos
   }

   c = 0x0a;
   ASM
   {
	 mov al,c
	 mov ah,02H    // 02H - write char to screen, dos
	 mov dl,al     // char to screen in dl
	 int 21H       // dos
   }
}
#ifdef COMMENT_OUT
#endif
