/* rcktfoss.h */


#define DEBUG 1

#ifdef DEBUG
static void deb_outs(char *str, int alarm);
#define DEBV_outs(s,l) {if (debug_level > 2) debv_outs(s,l);}
#define DEB_outs(s,l) {if (debug_level > 1) deb_outs(s,l);}
#define DEB_twidle(x) {if (debug_level > 0) twidle(x);}
#else
#define DEB_outs(s,l)
#define DEB_twidle(x)
#endif

// following routines in assembly module.
extern void interrupt RCKTIsr14();
extern void interrupt RCKTIsr8();
extern void interrupt RCKTIsr();
extern RCKT_Strat();
extern RCKT_Intr();
extern char named_dev;

// reduce heap and stack length for TSR
extern unsigned _heaplen= 1;
extern unsigned _stklen= 1024;


#define ASCII_XON 0x11
#define ASCII_XOFF 0x13
#define ASCII_CTRLC 3
#define ASCII_CTRLK 0x0b

#define S_WATCHDOG    0x0001
#define S_CTL_CK      0x0002
#define S_CTL_CK_GOT  0x0004
#define S_RX_SOFT_OFF 0x0008
#define S_RX_SOFT_CHK 0x0010
#define S_CD_ACTIVE   0x0020
#define S_BREAK_ON    0x0040
#define S_FUNC1E_DONE 0x0080
#define S_OPENED      0x0100
#define S_RTS_FLOW    0x0200
#define S_CTS_FLOW    0x0400
#define S_TX_SOFT_CHK 0x0800
#define S_RS485_FLOW  0x1000

#define TW_FUNC00  0x00
#define TW_FUNC01  0x01
#define TW_FUNC02  0x02
#define TW_FUNC03  0x03
#define TW_FUNC04  0x04
#define TW_FUNC05  0x05
#define TW_FUNC06  0x06
#define TW_FUNC07  0x07
#define TW_FUNC08  0x08
#define TW_FUNC09  0x09
#define TW_FUNC0A  0x0A
#define TW_FUNC0B  0x0B
#define TW_FUNC0C  0x0C
#define TW_FUNC0D  0x0D
#define TW_FUNC0E  0x0E
#define TW_FUNC0F  0x0F
#define TW_FUNC10  0x10
#define TW_FUNC11  0x11
#define TW_FUNC12  0x12
#define TW_FUNC13  0x13
#define TW_FUNC14  0x14
#define TW_FUNC15  0x15
#define TW_FUNC16  0x16
#define TW_FUNC17  0x17
#define TW_FUNC18  0x18
#define TW_FUNC19  0x19
#define TW_FUNC1A  0x1A
#define TW_FUNC1B  0x1B
#define TW_FUNC1C  0x1C
#define TW_FUNC1D  0x1D
#define TW_FUNC1E  0x1E
#define TW_FUNC1F  0x1F
#define TW_FUNC20  0x20
#define TW_FUNC21  0x21

#define TW_FUNCF4  0x24
#define TW_FUNCFB  0x25
#define TW_FUNCFC  0x26
#define TW_FUNCFD  0x27
#define TW_FUNCFE  0x28
#define TW_FUNCFF  0x29

#define TW_KILL_TIME 0x22
#define TW_POLL    0x23

// flag for driver_mode
#define BIOS_MODE 1

typedef struct
{
  CHANNEL_T  ch;

  DWORD bitRate;
  BYTE  bitRate00;  // the bitrate index func 00 would use
  BYTE  dataBits   : 1;  // 0=7, 1=8
  BYTE  stopBits   : 1;  // 0=1, 1=2
  BYTE  parityMode : 2;  // 0=none, 1=ODD, 2=EVEN
  BYTE  baud_no_access: 1; // used as flag to indicate access to baud by bios00
  BYTE  spare_bits : 3;

  BYTE MCRVal;
  BYTE allow_isr_access;  // ISR Tx Access
  WORD Status;
  BYTE chars_ready_flag;
  BYTE dev;  // device index(com-port number)

#ifdef SOFT_BUFS
  WORD QOutGet;
  WORD QOutPut;
  WORD QOutCount;
  unsigned char *QOutBase;
#endif

} RCKTPort_t; 

// following structure must match what assembly routine passes.  It forms
// this structure on the stack by doing a pusha.
typedef struct
{
  WORD ss;   // reg from push ss(needed in call)
  RCKTPort_t *Port;  // Port handle(from push ax)
  WORD flags; // flags(digi-bios needs to return flags)
  WORD es;    // reg from push es
  WORD di;
  WORD si;
  WORD bp;

  WORD sp;
  WORD bx;
  WORD dx;
  WORD cx;
  WORD ax;   // regs from pusha, in and returned
} Caller;


#define getdosticks() (*((unsigned int far *)(0x46cL)))
void tick_delay(WORD ticks);


static void set_break(RCKTPort_t *Port, int state);
static void set_rts(RCKTPort_t *Port, WORD state);
static void set_rts_auto(RCKTPort_t *Port, WORD state);
static void set_cts_auto(RCKTPort_t *Port, WORD state);
static void set_dtr(RCKTPort_t *Port, WORD state);

static void putsxy(char *str, int x, int y, char attr);
static void twidle(int func);
static void init_port_software(RCKTPort_t *Port);
static void init_port_hardware(RCKTPort_t *Port);
void poll_rocket(void);
void RCKTInt14(Caller far *caller);
void RCKTInt8();
int fossil_loaded(void);
int kill_time(WORD base_time);
void Reboot(void);
void interrupt RCKTPollISR();
void RPortTxFlush(RCKTPort_t *Port);
void RPortRxFlush(RCKTPort_t *Port);
void RPortConfigure(RCKTPort_t *Port);

void Bios04(Caller far *caller);
void Bios00(Caller far *caller);
void Bios01(Caller far *caller);
void Bios02(Caller far *caller);
WORD Bios03Status(Caller far *caller);
void Bios05(Caller far *caller);

void Foss00(Caller far *caller);
void Foss01(Caller far *caller);
void Foss02(Caller far *caller);
void Foss03(Caller far *caller);
void Foss04(Caller far *caller);
void Foss05(Caller far *caller);
void Foss06(Caller far *caller);
void Foss07(Caller far *caller);
void Foss08(Caller far *caller);
void Foss09(Caller far *caller);
void Foss0A(Caller far *caller);
void Foss0B(Caller far *caller);
void Foss0C(Caller far *caller);
void Foss0D(Caller far *caller);
void Foss0E(Caller far *caller);
void Foss0F(Caller far *caller);
void Foss10(Caller far *caller);
void Foss11(Caller far *caller);
void Foss12(Caller far *caller);
void Foss13(Caller far *caller);
void Foss14(Caller far *caller);
void Foss15(Caller far *caller);
void Foss16(Caller far *caller);
void Foss17(Caller far *caller);
void Foss18(Caller far *caller);
void Foss19(Caller far *caller);
void Foss1A(Caller far *caller);
void Foss1B(Caller far *caller);
void Foss1C(Caller far *caller);
void Foss1D(Caller far *caller);
void Foss1E(Caller far *caller);
void Foss1F(Caller far *caller);
void Foss20(Caller far *caller);
void Foss21(Caller far *caller);


void Rckt00(Caller far *caller);
void Rckt01(Caller far *caller);
void Rckt02(Caller far *caller);
void Rckt03(Caller far *caller);
void Rckt04(Caller far *caller);
// void Rckt05(Caller far *caller);
void Rckt06(Caller far *caller);
void Rckt07(Caller far *caller);
void Rckt08(Caller far *caller);
void Rckt09(Caller far *caller);
void Rckt0A(Caller far *caller);
void Rckt0B(Caller far *caller);
void Rckt0C(Caller far *caller);
void Rckt0D(Caller far *caller);
void Rckt0E(Caller far *caller);
void Rckt0F(Caller far *caller);
void Rckt10(Caller far *caller);
void Rckt11(Caller far *caller);
void Rckt12(Caller far *caller);
void Rckt13(Caller far *caller);
void Rckt14(Caller far *caller);
void Rckt15(Caller far *caller);
void Rckt16(Caller far *caller);
void Rckt17(Caller far *caller);
void Rckt18(Caller far *caller);
void Rckt19(Caller far *caller);
void Rckt1A(Caller far *caller);
void Rckt1B(Caller far *caller);
void Rckt1C(Caller far *caller);
void Rckt1D(Caller far *caller);
void Rckt1E(Caller far *caller);
void Rckt1F(Caller far *caller);


int fossil_loaded(void);
static int upcase(int c);
long getlint(char *textptr, int *countptr);
int getint(char *textptr, int *countptr);
unsigned int gethint(char *bufptr, int *countptr);
// void puts(char *str);
// void memset(unsigned char *str, unsigned char v, int size);
// int strlen(char *str);




/* Baud rate flags (unsigned char) */
#define COM_BAUD_50     0
#define COM_BAUD_75     0x1
#define COM_BAUD_110    0x2
#define COM_BAUD_134    0x3
#define COM_BAUD_150    0x4
#define COM_BAUD_300    0x5
#define COM_BAUD_600    0x6
#define COM_BAUD_1200   0x7
#define COM_BAUD_1800   0x8
#define COM_BAUD_2000   0x9
#define COM_BAUD_2400   0xa
#define COM_BAUD_4800   0xb
#define COM_BAUD_7200   0xc
#define COM_BAUD_9600   0xd
#define COM_BAUD_19200  0xe
#define COM_BAUD_38400  0xf
#define COM_BAUD_57600  0x10
#define COM_BAUD_76800  0x11
#define COM_BAUD_115200 0x12
#define COM_BAUD_230400 0x13

// following are values for standard int14 bios get status function(3)
#define LNSTAT_DATARDY    0x0100
#define LNSTAT_OVRRUN     0x0200
#define LNSTAT_PARITY     0x0400
#define LNSTAT_FRAMING    0x0800
#define LNSTAT_BREAK      0x1000
#define LNSTAT_THRE       0x2000
#define LNSTAT_TE         0x4000
#define LNSTAT_TIMEOUT    0x8000

// Modem status flags.  These mirror the 16450 modem status register.
#define MDMSTAT_DCTS       0x0001
#define MDMSTAT_DDSR       0x0002
#define MDMSTAT_DRI        0x0004
#define MDMSTAT_DCD        0x0008
#define MDMSTAT_CTS        0x0010
#define MDMSTAT_DSR        0x0020
#define MDMSTAT_RI         0x0040
#define MDMSTAT_CD         0x0080

//---- dos structure  chain for each named device: AUX, CON, etc.
struct dos_name_dev {
  struct dos_name_dev far *next;
  unsigned int attr;
  unsigned int strat;
  unsigned int inter;
  char name[8];
};

//---- dos structure passed to strat. routine(ES:BX) pts to this.
//---- first part fixed length, rest varies with call.
struct dos_req_header {
  BYTE len_rec;
  BYTE unit_code;  // not used with char devices
  BYTE command;
  WORD status;
  BYTE reserved[8];
};
